/*
 * Copyright (C) 2013 by I.Kato @ BS21 Lab
 * Color Kanji LCD library for Arduino
 *    ver.1.00 (15/aug/2013)
 *
 * 対象     : カラー漢字 LCD シールド (BS21LAB-006)
 * 接続方法 : TWI(i2c) 通信方式
 *
 * 詳細は http://www.bs21-lab.com/products/p006-ColorColorKanjiLCD/ を御覧下さい。
 *
 * This file is free software; you can redistribute it and/or modify it
 * under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1,
 * both as published by the Free Software Foundation.
 */

#include "Arduino.h"
#include <stdio.h>
#include <string.h>

#include "Wire.h" // 必須

#include "ColorKanjiLCD.h"


///////////////////////////////////////////////////////////////
// 定数定義


#define CKLCD_CMD_TIMEOUT  100  // タイムアウト時間:100ms (全領域 flip に約 30ms かかるので、それよりも大きい値にしてください)


// カラー漢字 LCD シールドの現在のステータスの内のスクリーンモードのビットマスク
#define CKLCD_DISP_SCREEN_BITMASK  B111111

// フォントモードのビットマスク
#define CKLCD_FONT_TYPE_BITMASK    B11111

// フォント属性のビットマスク
#define CKLCD_FONT_ATTR_BITMASK    (CKLCD_FONT_ATTR_UNDERLINE | CKLCD_FONT_ATTR_STRIKE | CKLCD_FONT_ATTR_INVERTED)

// LoadXxx 系関数の effect バイトのビットマスク
#define CKLCD_LOAD_IMG_EFFECT_BITMASK  (CKLCD_LOAD_IMG_DISP_ERR_MSG)


///////////////////////////////////////////////////////////////
// マクロ定義

#define MAKE_UINT16(l,h) (((uint16_t)(h) << 8) + (l))

#define lowWord(w)  ((uint16_t) ((w) & 0xFFFF))
#define highWord(w) ((uint16_t) ((w) >> 16))


///////////////////////////////////////////////////////////////
// Arduino Due 対応

// 2013/8/14 現在、Arduino IDE 1.5.2 の Wire ライブラリのバグにより、Arduino Due ではまともに動作しません。


///////////////////////////////////////////////////////////////
// カラー漢字 LCD コマンド・バイト定義

#define CKLCD_CMD_NUL                     0x00  // 何もしない: Busy 状態を確認する為のコマンド
#define CKLCD_CMD_INIT                    0x01  // 初期化: 引数 無し

#define CKLCD_CMD_EXT_CMD                 0x0F  // HX8353-Dのコマンド: 引数 cmd, dat, ... (Arduino の場合、Wireライブラリの関係で dat は最大 30byte まで)

#define CKLCD_CMD_GET_STATUS              0x10  // ステータス取得:       引数 無し  : 戻り stat[2]
#define CKLCD_CMD_GET_SCREEN_MODE         0x12  // スクリーンモード取得: 引数 無し  : 戻り mode

#define CKLCD_CMD_DISP_OFF                0x20  // 画面表示 OFF:         引数 無し
#define CKLCD_CMD_DISP_ON                 0x21  // 画面表示 ON :         引数 無し
#define CKLCD_CMD_SET_SCREEN_MODE         0x22  // スクリーンモード設定: 引数 mode
#define CKLCD_CMD_SET_FLIP_MODE           0x23  // フリップモード設定  : 引数 mode

#define CKLCD_CMD_FLIP                    0x30  // 仮想画面#1/#2 を合成しながら LCD 表示面に転送:  引数 x0, y0, x1, y1

#define CKLCD_CMD_LOAD_IMG_XLAT           0x38  // 要求された画像データを転送:  引数 data, ...
#define CKLCD_CMD_LOAD_IMG_CANCEL         0x39  // 転送中断: 引数 無し

#define CKLCD_CMD_GET_FONT_INFO           0x41  // 仮想画面#n のフォント情報取得:  引数 vp  : 戻り stat[3]

#define CKLCD_CMD_GET_LOAD_IMG_SIZE       0x44  // 仮想画面#n の LoadImg した画像サイズ取得: 引数 vp  : 戻り w16, h16

#define CKLCD_CMD_SET_ENCODING            0x50  // 仮想画面#n のエンコード設定:      引数 vp, enc
#define CKLCD_CMD_SET_FONT_TYPE           0x51  // 仮想画面#n のフォントタイプ設定:  引数 vp, type
#define CKLCD_CMD_SET_FONT_ATTR           0x52  // 仮想画面#n のフォント属性設定:    引数 vp, attr
#define CKLCD_CMD_SET_COLOR               0x53  // 仮想画面#n の描画/背景色の設定:   引数 vp/effect, r, g, b, a

#define CKLCD_CMD_COPY                    0x60  // 仮想画面#n へ等倍コピー     :  引数 vp/effect, x, y
#define CKLCD_CMD_COPY_W_ROT              0x61  // 仮想画面#n へ回転/拡大コピー:  引数 vp/effect, x0, y0, x1, y1, ratio, bam10

#define CKLCD_CMD_DILATE                  0x68  // 仮想画面#n へ膨張処理  :  引数 vp/effect, x0, y0, x1, y1
#define CKLCD_CMD_ERODE                   0x69  // 仮想画面#n へ収縮処理  :  引数 vp/effect, x0, y0, x1, y1
#define CKLCD_CMD_BLUR                    0x6A  // 仮想画面#n へブラー処理:  引数 vp/effect, x0, y0, x1, y1

#define CKLCD_CMD_PSET                    0x70  // 仮想画面#n へ点描画  : 引数 vp/effect, x, y
#define CKLCD_CMD_LINE                    0x71  // 仮想画面#n へ線描画  : 引数 vp/effect, x0, y0, x1, y1
#define CKLCD_CMD_RECT                    0x72  // 仮想画面#n へ矩形描画: 引数 vp/effect, x0, y0, x1, y1
#define CKLCD_CMD_CIRCLE                  0x73  // 仮想画面#n へ円描画  : 引数 vp/effect, x, y, rad

#define CKLCD_CMD_LOAD_JPG                0x78  // 仮想画面#n へ JPG 展開: 引数 vp/effect, x16, y16, size24
#define CKLCD_CMD_LOAD_PNG                0x79  // 仮想画面#n へ PNG 展開: 引数 vp/effect, x16, y16, size24
#define CKLCD_CMD_LOAD_GIF                0x7A  // 仮想画面#n へ GIF 展開: 引数 vp/effect, x16, y16, size24
#define CKLCD_CMD_LOAD_BMP                0x7B  // 仮想画面#n へ BMP 展開: 引数 vp/effect, x16, y16, size24

#define CKLCD_CMD_VP1_GET_STATUS          0x80  // 仮想画面#1 の現在のステータスを取得:  引数 無し  : 戻り stat
#define CKLCD_CMD_VP2_GET_STATUS          0x81  // 仮想画面#2 の現在のステータスを取得:  引数 無し  : 戻り stat
#define CKLCD_CMD_VP1_GET_PIXEL           0x84  // 仮想画面#1 の仮想画面から直接取得:    引数 x, y  : 戻り col16
#define CKLCD_CMD_VP2_GET_PIXEL           0x85  // 仮想画面#2 の仮想画面から直接取得:    引数 x, y  : 戻り col16
#define CKLCD_CMD_VP1_GET_DISP_START_POS  0x88  // 仮想画面#1 の LCD 表示開始座標の取得: 引数 無し  : 戻り x, y
#define CKLCD_CMD_VP2_GET_DISP_START_POS  0x89  // 仮想画面#2 の LCD 表示開始座標の取得: 引数 無し  : 戻り x, y
#define CKLCD_CMD_VP1_GET_CURSOR_POS      0x8C  // 仮想画面#1 の カーソル位置取得:       引数 無し  : 戻り x,y
#define CKLCD_CMD_VP2_GET_CURSOR_POS      0x8D  // 仮想画面#2 の カーソル位置取得:       引数 無し  : 戻り x,y

#define CKLCD_CMD_VP1_SET_PIXEL           0xA4  // 仮想画面#1 の仮想画面に直接書き込み:  引数 x, y, col16
#define CKLCD_CMD_VP2_SET_PIXEL           0xA5  // 仮想画面#2 の仮想画面に直接書き込み:  引数 x, y, col16
#define CKLCD_CMD_VP1_SET_DISP_START_POS  0xA8  // 仮想画面#1 の LCD 表示開始座標の指定: 引数 x, y
#define CKLCD_CMD_VP2_SET_DISP_START_POS  0xA9  // 仮想画面#2 の LCD 表示開始座標の指定: 引数 x, y
#define CKLCD_CMD_VP1_SET_CURSOR_POS      0xAC  // 仮想画面#1 のカーソル位置設定:        引数 x,y
#define CKLCD_CMD_VP2_SET_CURSOR_POS      0xAD  // 仮想画面#2 のカーソル位置設定:        引数 x,y

#define CKLCD_CMD_VP1_SET_SRC_RECT        0xB0  // 仮想画面#1 のコピー元の座標設定:  引数 x0, y0, x1, y1
#define CKLCD_CMD_VP2_SET_SRC_RECT        0xB1  // 仮想画面#2 のコピー元の座標設定:  引数 x0, y0, x1, y1

#define CKLCD_CMD_VP1_CLR_SCRN            0xC0  // 仮想画面#1 の画面クリア: 引数 無し
#define CKLCD_CMD_VP2_CLR_SCRN            0xC1  // 仮想画面#2 の画面クリア: 引数 無し

#define CKLCD_CMD_VP1_TEXT_OPAQ_OVER      0xD0  // 仮想画面#1 へテキスト転送: 不透明/上書き:  引数 byte, ...
#define CKLCD_CMD_VP2_TEXT_OPAQ_OVER      0xD1  // 仮想画面#2 へテキスト転送: 不透明/上書き:  引数 byte, ...
#define CKLCD_CMD_VP1_TEXT_TRANS_OVER     0xD4  // 仮想画面#1 へテキスト転送: 透明  /上書き:  引数 byte, ...
#define CKLCD_CMD_VP2_TEXT_TRANS_OVER     0xD5  // 仮想画面#2 へテキスト転送: 透明  /上書き:  引数 byte, ...
#define CKLCD_CMD_VP1_TEXT_OPAQ_BLEND     0xD8  // 仮想画面#1 へテキスト転送: 不透明/合成  :  引数 byte, ...
#define CKLCD_CMD_VP2_TEXT_OPAQ_BLEND     0xD9  // 仮想画面#2 へテキスト転送: 不透明/合成  :  引数 byte, ...
#define CKLCD_CMD_VP1_TEXT_TRANS_BLEND    0xDC  // 仮想画面#1 へテキスト転送: 透明  /合成  :  引数 byte, ...
#define CKLCD_CMD_VP2_TEXT_TRANS_BLEND    0xDD  // 仮想画面#2 へテキスト転送: 透明  /合成  :  引数 byte, ...



///////////////////////////////////////////////////////////////
// ColorKanjiLCDCore : カラー漢字 LCD コアクラス

// コンストラクタ：カラー漢字 LCD シールドの TWI(i2c) アドレス値を指定する
ColorKanjiLCDCore::ColorKanjiLCDCore (uint8_t twiAddr) : _twiAddr(twiAddr)
{
  res    = CKLCD_ERR_NOT_INIT; // 未初期化
  _count = 0;

  _phyWidth  = 0;
  _phyHeight = 0;

  _currScreenMode = 0; // 現在のスクリーンモード

  // カラー漢字 LCD シールドの現在のステータスも 0 に初期化
  _currStatus[0] = _currStatus[1] = 0;
}


// デストラクタ：特に何もしません
ColorKanjiLCDCore::~ColorKanjiLCDCore ()
{
}


// 送信開始：実際は TWI アドレスをセットするだけ
errCode_t ColorKanjiLCDCore::sendBegin (void)
{
  CKLCD_Wire.beginTransmission(_twiAddr); // カラー漢字 LCD シールドの TWIアドレスをセット

  return CKLCD_ERR_NONE;
}


// データ送信：実際は内部バッファにデータをセットするだけです。セットできた時は非0 が返されます。
// 内部バッファは最大32バイトなので、それ以上の時は、0 が返されます。
size_t ColorKanjiLCDCore::sendData (uint8_t dat)
{
  return CKLCD_Wire.write(dat); // Wire ライブラリの内部バッファ末尾に追加するだけ
}


// 送信終了：ここで初めて TWI 通信する。
// sendData() の時点で内部バッファ一杯になってしまったら、必ず CKLCD_ERR_TWI_BUFF_FULL エラーになります。
errCode_t ColorKanjiLCDCore::sendEnd (void)
{
  errCode_t res = CKLCD_Wire.endTransmission(); // ここで初めて TWI 通信する

#ifdef __arm__
    // Arduino Due 対応：でも Arduino IDE 1.5.2 では上手くいかない
    delayMicroseconds(100); // 100μ秒待つ
    if (res == 1) {
      res = 0;
    }
#endif

  return res;
}


// データ受信：siz バイト分を受信し、dat 配列に入れる。dat配列の大きさは siz 以上ある事。
// ※ カラー漢字 LCD シールドがコマンド処理中は CKLCD_ERR_BUSY エラーを返しますので、呼び出し側は必ず返り値をチェックして下さい。
errCode_t ColorKanjiLCDCore::recvData (uint8_t * dat, uint8_t siz)
{
  errCode_t res = waitWhileBusy(); // カラー漢字 LCD シールドが次のコマンド受付け可能になるまで待つ

  if (res != CKLCD_ERR_NONE) { // タイムアウトした場合
    return res;
  }

  CKLCD_Wire.requestFrom (_twiAddr, siz, (uint8_t)true); // カラー漢字 LCD シールドへ siz バイト要求する

  if (CKLCD_Wire.available() < siz) { // siz バイト分、受信できなかった
    return CKLCD_ERR_BUSY; // カラー漢字 LCD シールドはコマンド処理中なので、データが取得できなかった
  }

  while (siz) {
    *dat++ = CKLCD_Wire.read(); // １バイトずつ dat 配列に格納していく
    --siz;
  }

  return CKLCD_ERR_NONE; // データが取得できた
}


// カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ
// 戻り値が CKLCD_ERR_NONE なら、コマンド受付け可能状態
errCode_t ColorKanjiLCDCore::waitWhileBusy (void)
{
  unsigned long t = millis(); // プログラム実行開始からの経過時間(ミリ秒)

  while (1) { // 永久ループ

    // カラー漢字 LCD シールドがコマンド受付け可能かチェック
    sendBegin();              // カラー漢字 LCD シールドの TWI アドレスをセット
    sendData (CKLCD_CMD_NUL); // 何もしないコマンドバイトを送ってみる：実際はこの段階では Wire ライブラリの送信バッファ(txBuffer)に入れるだけ
    res = sendEnd();          // この段階で、始めて TWI 通信される

    if (res == CKLCD_ERR_NONE) {
      // 問題無く NUL コマンドが処理された
      return CKLCD_ERR_NONE; // コマンド受付け可能状態
    }

    // 「まだだよ」返答：カラー漢字 LCD シールドは以前のコマンドを処理中(SLA+W に対して NACK が返ってきた)か、そもそも、TWI 通信自体ができていない
    if ((millis() - t) > CKLCD_CMD_TIMEOUT) { // 所定の時間が経過した (おそらく TWI 通信できていない)
      break; // 強制ループアウトさせる
    }

    delayMicroseconds(10);  // 10μ秒待つ
  }

  return CKLCD_ERR_TIME_OUT; // タイムアウトした
}


// コマンド(１バイト)転送
errCode_t ColorKanjiLCDCore::command (uint8_t cmd)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(２バイト)転送
errCode_t ColorKanjiLCDCore::command2 (uint8_t cmd, uint8_t dat1)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(３バイト)転送
errCode_t ColorKanjiLCDCore::command3 (uint8_t cmd, uint8_t dat1, uint8_t dat2)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    sendData (dat2);  // dat2 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(４バイト)転送
errCode_t ColorKanjiLCDCore::command4 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    sendData (dat2);  // dat2 を送信バッファに入れる
    sendData (dat3);  // dat3 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(５バイト)転送
errCode_t ColorKanjiLCDCore::command5 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    sendData (dat2);  // dat2 を送信バッファに入れる
    sendData (dat3);  // dat3 を送信バッファに入れる
    sendData (dat4);  // dat4 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(６バイト)転送
errCode_t ColorKanjiLCDCore::command6 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    sendData (dat2);  // dat2 を送信バッファに入れる
    sendData (dat3);  // dat3 を送信バッファに入れる
    sendData (dat4);  // dat4 を送信バッファに入れる
    sendData (dat5);  // dat5 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(７バイト)転送
errCode_t ColorKanjiLCDCore::command7 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5, uint8_t dat6)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    sendData (dat2);  // dat2 を送信バッファに入れる
    sendData (dat3);  // dat3 を送信バッファに入れる
    sendData (dat4);  // dat4 を送信バッファに入れる
    sendData (dat5);  // dat5 を送信バッファに入れる
    sendData (dat6);  // dat6 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(８バイト)転送
errCode_t ColorKanjiLCDCore::command8 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5, uint8_t dat6, uint8_t dat7)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    sendData (dat2);  // dat2 を送信バッファに入れる
    sendData (dat3);  // dat3 を送信バッファに入れる
    sendData (dat4);  // dat4 を送信バッファに入れる
    sendData (dat5);  // dat5 を送信バッファに入れる
    sendData (dat6);  // dat6 を送信バッファに入れる
    sendData (dat7);  // dat7 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(９バイト)転送
errCode_t ColorKanjiLCDCore::command9 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5, uint8_t dat6, uint8_t dat7, uint8_t dat8)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    sendData (dat2);  // dat2 を送信バッファに入れる
    sendData (dat3);  // dat3 を送信バッファに入れる
    sendData (dat4);  // dat4 を送信バッファに入れる
    sendData (dat5);  // dat5 を送信バッファに入れる
    sendData (dat6);  // dat6 を送信バッファに入れる
    sendData (dat7);  // dat7 を送信バッファに入れる
    sendData (dat8);  // dat8 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}

// コマンド(１０バイト)転送
errCode_t ColorKanjiLCDCore::command10 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5, uint8_t dat6, uint8_t dat7, uint8_t dat8, uint8_t dat9)
{
  res = waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ

  if (res == CKLCD_ERR_NONE) { // コマンド受付け可能
    sendBegin();      // カラー漢字 LCD シールドの TWIアドレスをセット
    sendData (cmd);   // cmd  を送信バッファに入れる
    sendData (dat1);  // dat1 を送信バッファに入れる
    sendData (dat2);  // dat2 を送信バッファに入れる
    sendData (dat3);  // dat3 を送信バッファに入れる
    sendData (dat4);  // dat4 を送信バッファに入れる
    sendData (dat5);  // dat5 を送信バッファに入れる
    sendData (dat6);  // dat6 を送信バッファに入れる
    sendData (dat7);  // dat7 を送信バッファに入れる
    sendData (dat8);  // dat8 を送信バッファに入れる
    sendData (dat9);  // dat9 を送信バッファに入れる
    res = sendEnd();  // ここで、始めて TWI 通信される。※waitWhileBusy 関数を無事に通過しているので、まず間違いなく 0 が返ってくると思うが、一応、res に格納する
  }
  return res;
}


// 現在のスクリーンモード
errCode_t ColorKanjiLCDCore::getScreenMode (void)
{
  // カラー漢字 LCD シールドのスクリーンモードを１バイト要求します。
  errCode_t res = command (CKLCD_CMD_GET_SCREEN_MODE); // スクリーンモード取得コマンド

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = recvData(&_currScreenMode, 1); // １バイト受信する
  }
  return res;
}


// LCD を表示させない
// ただし、バックライトは点きっぱなしなので、真っ白の画面になります。
errCode_t ColorKanjiLCDCore::noDisplay (void)
{
  return command (CKLCD_CMD_DISP_OFF); // 画面表示 OFF コマンド
}

// LCD を表示させる
errCode_t ColorKanjiLCDCore::display (void)
{
  return command (CKLCD_CMD_DISP_ON); // 画面表示 ON コマンド
}


// フリップモードの設定
// CKLCD_FLIP_MODE_xxx のいずれかを指定します
errCode_t ColorKanjiLCDCore::setFlipMode (uint8_t flipMode)
{
  errCode_t res = command2 (CKLCD_CMD_SET_FLIP_MODE, flipMode); // フリップモード設定コマンド: 引数 mode

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = getStatus(); // 現在のステータスを更新する：getFlipMode() が現状を反映する様にさせる為
  }
  return res;
}


// カラー漢字 LCD シールドの現在のステータスを取得する
// ※ カラー漢字 LCD シールドがコマンド処理中は CKLCD_ERR_BUSY エラーを返しますので、呼び出し側は必ず返り値をチェックして下さい。
errCode_t ColorKanjiLCDCore::getStatus (void)
{
  // カラー漢字 LCD シールドのステータスを２バイト要求します。
  errCode_t res = command (CKLCD_CMD_GET_STATUS); // ステータス取得コマンド

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = recvData(_currStatus, 2); // ２バイト受信する
  }
  return res;
}


// 以下の関数は、直前に getStatus 関数を呼び出し、CKLCD_ERR_NONE が返ってきたのを確認してから、呼び出す事

// 各 status の MSB からのビット割り当て
// stat[0]  1固定: Err6: Err5: Err4: Err3: Err2: Err1: Err0
// stat[1]  0固定: DispON/OFF: AutoFlipON/OFF: 0: SwR: SwD: SwU: SwL

// 現在のカラー漢字 LCD シールドのエラーコード : CKLCD_ERR_xxx のいずれかの値
uint8_t ColorKanjiLCDCore::getError (void)
{
  if (_currStatus[0] == 0x80) {
    // 0x80 の時はエラーではなく、正常をあらわします。
    return CKLCD_ERR_NONE;
  }
  return _currStatus[0];
}

// 現在の LCD 表示の状態 : 0=OFF, 1=ON
uint8_t ColorKanjiLCDCore::getDisplay (void)
{
  return (_currStatus[1] >> 6) & B00000001; // stat1 の bit6
}

// 現在のフリップモード
uint8_t ColorKanjiLCDCore::getFlipMode (void)
{
  return _currStatus[1] & CKLCD_FLIP_MODE_MANU; // stat1 の bit5
}

// 現在の左スイッチの状態 : 0=OFF, 1=ON
uint8_t ColorKanjiLCDCore::getLeftSW (void)
{
  return (_currStatus[1] >> 0) & B00000001; // stat1 の bit0
}

// 現在の上スイッチの状態 : 0=OFF, 1=ON
uint8_t ColorKanjiLCDCore::getUpSW (void)
{
  return (_currStatus[1] >> 1) & B00000001; // stat1 の bit1
}

// 現在の下スイッチの状態 : 0=OFF, 1=ON
uint8_t ColorKanjiLCDCore::getDownSW (void)
{
  return (_currStatus[1] >> 2) & B00000001; // stat1 の bit2
}

// 現在の右スイッチの状態 : 0=OFF, 1=ON
uint8_t ColorKanjiLCDCore::getRightSW (void)
{
  return (_currStatus[1] >> 3) & B00000001; // stat1 の bit3
}


// グラフィック描画

// 仮想画面#1/#2 をアルファ合成しながら LCD へ転送する事で、初めて LCD に表示されます
// 注意：全画面領域 (0,0,159,127) または (0,0,127,159) を指定した時が、もっとも転送時間が掛かります。指定範囲が狭ければ狭いほど、早く転送が終わります。

// 全表示領域
errCode_t ColorKanjiLCDCore::flip (void)
{
  return command5 (CKLCD_CMD_FLIP, 0, 0, _phyWidth - 1, _phyHeight - 1); // 仮想画面転送コマンド:  引数 x0, y0, x1, y1
}

// 指定領域:(x0,y0)-(x1,y1) のみ
errCode_t ColorKanjiLCDCore::flip (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  return command5 (CKLCD_CMD_FLIP, x0, y0, x1, y1); // 仮想画面転送コマンド:  引数 x0, y0, x1, y1
}



///////////////////////////////////////////////////////////////////////////////////////////////
// ColorKanjiLCDVPlane : カラー漢字 LCD 仮想画面クラスの宣言 ※Printクラスの派生クラス

// コンストラクタ：カラー漢字 LCD シールドの TWI(i2c) アドレス値を指定する
ColorKanjiLCDVPlane::ColorKanjiLCDVPlane (uint8_t vplane_no, ColorKanjiLCDCore& core)
  : _vplane_no(vplane_no), _core(core)
{
  _cond   = CKLCD_VP_COND_OPAQUE;        // 仮想画面の状態設定：CKLCD_VP_COND_xxx のいずれか
  _blend  = CKLCD_BLEND_OVER;            // アルファブレンドの設定：CKLCD_BLEND_xxx のいずれか
  _merge  = _vplane_no + _cond + _blend; // 上記設定をORしたもの

  _orient = CKLCD_ORIENT_LANDSCAPE;   // 仮想画面のオリエンテーション設定：CKLCD_ORIENT_xxx のいずれか

  _currFontInfo[CKLCD_CURR_FONT_INFO_TYPE]   =  0;  // 現在のフォントタイプ：CKLCD_FONT_TYPE_xxx のいずれか
  _currFontInfo[CKLCD_CURR_FONT_INFO_WIDTH]  =  8;  // 現在のフォントの幅
  _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT] = 20;  // 現在のフォントの高さ

  _currVPStatus = 0; // カラー漢字 LCD シールドの仮想画面の現在の状態

  _width  = 0; // 仮想画面の最大幅  ：ランドスケープモード時は 160、ポートレートモード時は 128 か 256
  _height = 0; // 仮想画面の最大高さ：ランドスケープモード時は 128 か 256、ポートレートモード時は 160
}


// デストラクタ：特に何もしません
ColorKanjiLCDVPlane::~ColorKanjiLCDVPlane ()
{
}


// 仮想画面の状態設定：CKLCD_VP_COND_xxx のいずれかを設定します
// ※ ColorKanjiLCD::setScreenMode() から呼び出されます。ユーザーが変更する事も可能ですが、間違うとおかしな色で描画される事になるので注意して下さい。
errCode_t ColorKanjiLCDVPlane::setCondition (uint8_t cond)
{
  if (cond == CKLCD_VP_COND_OPAQUE) {
    _cond = CKLCD_VP_COND_OPAQUE;  // 非透過モード
  } else {
    _cond = CKLCD_VP_COND_TRANS;   // 透過モード
  }
  _merge = _vplane_no + _cond + _blend; // 設定をORしたもの

  return CKLCD_ERR_NONE;
}


// 仮想画面の状態取得
uint8_t ColorKanjiLCDVPlane::getCondition (void)
{
  return _cond;
}


// 仮想画面のオリエンテーションモード設定：CKLCD_ORIENT_xxx のいずれかを設定します
// ※ ColorKanjiLCD::setScreenMode() から呼び出されます。通常、ユーザーが設定する必要はありません。
errCode_t ColorKanjiLCDVPlane::setOrientMode (uint8_t orient)
{
  if (orient == CKLCD_ORIENT_LANDSCAPE) {
    _orient = CKLCD_ORIENT_LANDSCAPE;   // 仮想画面のオリエンテーション設定：ランドスケープモード：液晶画面を横長にした状態
    _width  = (_vplane_no == CKLCD_VPLANE1) ? CKLCD_VPLANE1_MAX_L_WIDTH  : CKLCD_VPLANE2_MAX_L_WIDTH;  // 仮想画面の最大幅  ：160
    _height = (_vplane_no == CKLCD_VPLANE1) ? CKLCD_VPLANE1_MAX_L_HEIGHT : CKLCD_VPLANE2_MAX_L_HEIGHT; // 仮想画面の最大高さ：仮想画面#1は 256、仮想画面#2は 128
  } else {
    _orient = CKLCD_ORIENT_PORTRAIT;   // 仮想画面のオリエンテーション設定：ポートレートモード：液晶画面を縦長にした状態
    _width  = (_vplane_no == CKLCD_VPLANE1) ? CKLCD_VPLANE1_MAX_P_WIDTH  : CKLCD_VPLANE2_MAX_P_WIDTH;  // 仮想画面の最大幅  ：仮想画面#1は 256、仮想画面#2は 128
    _height = (_vplane_no == CKLCD_VPLANE1) ? CKLCD_VPLANE1_MAX_P_HEIGHT : CKLCD_VPLANE2_MAX_P_HEIGHT; // 仮想画面の最大高さ：160
  }
  return CKLCD_ERR_NONE;
}


// カラー漢字 LCD シールドの符号化方式を変更する: CKLCD_ENC_xxx のいずれかを指定します
// 範囲外の値を指定した時は CKLCD_ERR_INVALID_PARAM エラーを返します
// いつでも変更できます。当然ですが、変更した後は、その符号化方式にて、文字コードを漢字 LCD シールドに送らなければなりません。
// さらに UTF-16 方式 (BOM 自動判別) にセットした時は、最初に BOM を送信して、エンディアンを自動判別させる必要があります。
errCode_t ColorKanjiLCDVPlane::setEncoding (uint8_t enc)
{
  if (enc >= CKLCD_MAX_ENCODING) {
    return CKLCD_ERR_INVALID_PARAM; // サポート範囲外
  }
  return _core.command3 (CKLCD_CMD_SET_ENCODING, _vplane_no, enc); // エンコード設定コマンド: 引数 vp, enc
}


// フォントタイプの設定
// モノスペース(固定幅)か、プロポーショナル(可変幅)か、を指定します。
// CKLCD_FONT_TYPE_xxx のいずれかを設定します
errCode_t ColorKanjiLCDVPlane::setFontType (uint8_t type)
{
  type &= CKLCD_FONT_TYPE_BITMASK;
  errCode_t res = _core.command3 (CKLCD_CMD_SET_FONT_TYPE, _vplane_no, type); // フォントタイプ設定コマンド: 引数 vp, type

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = getFontInfo ();  // 現在のフォント情報を取得する
  }
  return res;
}

// フォントの属性を設定します
// CKLCD_FONT_ATTR_xxx の組み合わせを設定します
errCode_t ColorKanjiLCDVPlane::setFontAttr (uint8_t attr)
{
  attr &= CKLCD_FONT_ATTR_BITMASK;
  return _core.command3 (CKLCD_CMD_SET_FONT_ATTR, _vplane_no, attr); // フォント属性設定コマンド: 引数 vp, attr
}


// 現在のフォント情報を取得する
errCode_t ColorKanjiLCDVPlane::getFontInfo (void)
{
  // カラー漢字 LCD シールドの仮想画面のォント情報を３バイト要求します。
  errCode_t res = _core.command2 (CKLCD_CMD_GET_FONT_INFO, _vplane_no); // 仮想画面#n のフォント情報取得:  引数 vp  : 戻り stat[3]

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = _core.recvData(_currFontInfo, 3); // ３バイト受信する
  }
  return res;
}


// カラー漢字 LCD シールドの仮想画面の現在のステータスを取得する
// ※ カラー漢字 LCD シールドがコマンド処理中は CKLCD_ERR_BUSY エラーを返しますので、呼び出し側は必ず返り値をチェックして下さい。
errCode_t ColorKanjiLCDVPlane::getStatus (void)
{
  // カラー漢字 LCD シールドの仮想画面のステータスを１バイト要求します。
  errCode_t res = _core.command (CKLCD_CMD_VP1_GET_STATUS + _vplane_no); // 仮想画面ステータス取得コマンド: 引数 無し  : 戻り stat

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = _core.recvData(&_currVPStatus, 1); // １バイト受信する
  }
  return res;
}


// 以下の関数は、直前に getStatus 関数を呼び出し、CKLCD_ERR_NONE が返ってきたのを確認してから、呼び出す事

// VPstat の MSB からのビット割り当て
// VPstat   0固定: 反転: 取消し線: 下線: 未処理データ有: Enc2: Enc1: Enc0

// 現在の符号化方式の状態 : CKLCD_ENC_xxx のいずれかの値
uint8_t ColorKanjiLCDVPlane::getEncoding (void)
{
  return _currVPStatus & B00000111; // VPstat の bit2～0
}

// 文字コード欠乏状態：0=充足中, 1=欠乏中
// 例えば UTF-8 方式で、３バイト必要な文字コードで、まだ２バイトしか転送していない時に 1 になります。
// その後、３バイト目が転送されて、１文字を表示する事ができたら、0 になります。
uint8_t ColorKanjiLCDVPlane::getLackedCode (void)
{
  return (_currVPStatus >> 3) & B00000001; // VPstat の bit3
}

// 現在のフォント属性の設定：アンダーライン ON/OFF
uint8_t ColorKanjiLCDVPlane::getFontUnderline (void)
{
  return (_currVPStatus >> 4) & B00000001; // VPstat の bit4
}

// 現在のフォント属性の設定：取り消し線 ON/OFF
uint8_t ColorKanjiLCDVPlane::getFontStrike (void)
{
  return (_currVPStatus >> 5) & B00000001; // VPstat の bit5
}

// 現在のフォント属性の設定：文字色反転 ON/OFF
uint8_t ColorKanjiLCDVPlane::getFontInverted (void)
{
  return (_currVPStatus >> 6) & B00000001; // VPstat の bit6
}


// LCD 表示開始座標：この座標が LCD 画面の左上になります。
// LCD 表示開始座標の指定：あくまで表示位置の設定だけであり、現在のカーソル位置には影響しません。
errCode_t ColorKanjiLCDVPlane::setDispStartPos (uint8_t x, uint8_t y)
{
  return _core.command3 (CKLCD_CMD_VP1_SET_DISP_START_POS + _vplane_no, x, y); // 仮想画面#n の LCD 表示開始座標の指定:  引数 x, y
}

// 現在の LCD 表示開始座標の取得
errCode_t ColorKanjiLCDVPlane::getDispStartPos (uint8_t * x, uint8_t * y)
{
  // カラー漢字 LCD シールドの仮想画面のLCD 表示開始座標を２バイト要求します。
  errCode_t res = _core.command (CKLCD_CMD_VP1_GET_DISP_START_POS + _vplane_no); // 仮想画面#n の LCD 表示開始座標取得コマンド: 戻り x, y

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた

    uint8_t pos[2];
    res = _core.recvData(pos, 2); // ２バイト受信する

    if (res == CKLCD_ERR_NONE) { // データが受信できた
      if (x) {
        *x = pos[0];  // x座標
      }
      if (y) {
        *y = pos[1];  // y座標
      }
    }
  }
  return res;
}


// 論理色の設定：
//   描画色と背景色の２つがあります。基本的に、文字本体やグラフィック関係(pset,line,rect,circle)は描画色で描画されます。
//   背景色は、文字や画像の背景に使われます。描画色と違う背景色を指定しておかないと、文字が見づらかったり、四角形が表示されるだけ、になってしまいます。

// 描画色: 基本的にこの色が使用されます。
// 論理色：ARGB8888 の内のR/G/B 指定のみ。Alphaは255になる。Back面用の論理色指定に最適
errCode_t ColorKanjiLCDVPlane::setDrawColor (uint8_t red, uint8_t green, uint8_t blue)
{
  return _core.command6 (CKLCD_CMD_SET_COLOR, _vplane_no + CKLCD_DRAW_COLOR, red, green, blue, 255); // 仮想画面#n の描画色の設定: 引数 vp/effect, r, g, b, a
}

// 論理色：ARGB8888 を個別に指定。Front面用の論理色指定に最適
errCode_t ColorKanjiLCDVPlane::setDrawColor (uint8_t red, uint8_t green, uint8_t blue, uint8_t alpha)
{
  return _core.command6 (CKLCD_CMD_SET_COLOR, _vplane_no + CKLCD_DRAW_COLOR, red, green, blue, alpha); // 仮想画面#n の描画色の設定: 引数 vp/effect, r, g, b, a
}

// 論理色：ARGB8888 をダイレクトに指定
errCode_t ColorKanjiLCDVPlane::setDrawColor (cklcdCol32_t col32)
{
  return _core.command6 (CKLCD_CMD_SET_COLOR, _vplane_no + CKLCD_DRAW_COLOR,
    CKLCD_COL32_SPLIT_RED  (col32),
    CKLCD_COL32_SPLIT_GREEN(col32),
    CKLCD_COL32_SPLIT_BLUE (col32),
    CKLCD_COL32_SPLIT_ALPHA(col32)); // 仮想画面#n の描画色の設定: 引数 vp/effect, r, g, b, a
}

// 背景色: write(), print(), roll(), loadBmp(), loadPng(), loadGif() で、背景の色として使用されます。
// 論理色：ARGB8888 の内のR/G/B 指定のみ。Alphaは255になる。Back面用の論理色指定に最適
errCode_t ColorKanjiLCDVPlane::setBackColor (uint8_t red, uint8_t green, uint8_t blue)
{
  return _core.command6 (CKLCD_CMD_SET_COLOR, _vplane_no + CKLCD_BACK_COLOR, red, green, blue, 255); // 仮想画面#n の背景色の設定: 引数 vp/effect, r, g, b, a
}

// 論理色：ARGB8888 を個別に指定。Front面用の論理色指定に最適
errCode_t ColorKanjiLCDVPlane::setBackColor (uint8_t red, uint8_t green, uint8_t blue, uint8_t alpha)
{
  return _core.command6 (CKLCD_CMD_SET_COLOR, _vplane_no + CKLCD_BACK_COLOR, red, green, blue, alpha); // 仮想画面#n の背景色の設定: 引数 vp/effect, r, g, b, a
}

// 論理色：ARGB8888 をダイレクトに指定
errCode_t ColorKanjiLCDVPlane::setBackColor (cklcdCol32_t col32)
{
  return _core.command6 (CKLCD_CMD_SET_COLOR, _vplane_no + CKLCD_BACK_COLOR,
    CKLCD_COL32_SPLIT_RED  (col32),
    CKLCD_COL32_SPLIT_GREEN(col32),
    CKLCD_COL32_SPLIT_BLUE (col32),
    CKLCD_COL32_SPLIT_ALPHA(col32)); // 仮想画面#n の背景色の設定: 引数 vp/effect, r, g, b, a
}


// 仮想画面に描画する時のアルファブレンド方法を指定します。
// アルファブレンドの設定：CKLCD_BLEND_xxx のいずれかを設定
errCode_t ColorKanjiLCDVPlane::setBlendMode (uint8_t blend)
{
  if (blend == CKLCD_VP_COND_OPAQUE) {
    _blend = CKLCD_BLEND_OVER;  // ブレンドしないで上書きします。処理速度は速いです。
  } else {
    _blend = CKLCD_BLEND_ADD;   // アルファ値を元に、ブレンド処理します。論理アルファ値==1.0 の時は上書きした時と同じ結果になります。各画素毎に計算するので、処理速度は遅いです。
  }
  _merge = _vplane_no + _cond + _blend; // 設定をORしたもの

  return CKLCD_ERR_NONE;
}

// 現在のアルファブレンドの設定の取得
uint8_t ColorKanjiLCDVPlane::getBlendMode (void)
{
  return _blend;
}


// 仮想画面のクリア

// 仮想画面の全域を 0x0000 の絶対色でクリアし、LCD 表示開始座標とカーソル位置を (0,0) にします。
errCode_t ColorKanjiLCDVPlane::clear (void)
{
  return clearEx (CKLCD_CLR_FLG_FLIP + CKLCD_CLR_FLG_DISP_START_HOME + CKLCD_CLR_FLG_CURSOR_HOME);
}

// 仮想画面の全域を 0x0000 の絶対色でクリアし、LCD 表示開始座標とカーソル位置を (0,0) にするか・しないかを選べます。CKLCD_CLR_FLG_XXX を組み合わせます。
errCode_t ColorKanjiLCDVPlane::clearEx (uint8_t flags)
{
  errCode_t res = _core.command (CKLCD_CMD_VP1_CLR_SCRN + _vplane_no); // 仮想画面#n の画面クリアコマンド

  if (res == CKLCD_ERR_NONE) {
    if (flags & CKLCD_CLR_FLG_CURSOR_HOME) { // フラグが立っている
      res = setCursorMin (0, 0);  // カーソル位置を (0,0) にします。
    }

    if (res == CKLCD_ERR_NONE) {
      if (flags & CKLCD_CLR_FLG_DISP_START_HOME) { // フラグが立っている
        res = setDispStartPos (0, 0); // LCD 表示開始座標を (0,0) にします。
      }

      if (res == CKLCD_ERR_NONE) {
        if (flags & CKLCD_CLR_FLG_FLIP) { // フラグが立っている
          if (_core.getFlipMode() == CKLCD_FLIP_MODE_AUTO) { // フリップモードが AUTO モードの時のみ
            res = _core.flip (); // 現在の LCD 表示開始座標に従って、全表示領域を flip させます
          }
        }
      }
    }
  }
  return res;
}


// 仮想画面の矩形領域(x0,y0)-(x1,y1)を setBackColor() で設定した論理色でクリアします。
errCode_t ColorKanjiLCDVPlane::clearRect (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  // 指定矩形領域を背景色で塗りつぶします。
  return _core.command6 (CKLCD_CMD_RECT, // 仮想画面#n へ矩形描画
    _vplane_no + _cond + CKLCD_BLEND_OVER + CKLCD_BACK_COLOR + CKLCD_DRAW_FILL, // 現在のブレンド設定は無視して、常に上書きにする
    x0, y0, x1, y1);
}


// 仮想画面をＹ方向に ySizeライン分スクロールUP/DOWN します。空いた領域は setBackColor() で設定した論理色になります。
// 現在のカーソル位置と LCD 表示開始座標は変化しませんので、注意して下さい。
// 実際は setSrcRect() と copy() と clearRect() の３つを呼び出しているので、遅いです。

// 全領域
errCode_t ColorKanjiLCDVPlane::scrollUp (uint8_t ySize)
{
  return scrollUp (0, 255, ySize);  // 終了ライン=255 でも、scrollUp(y0, y1, ySize) の方で Y端に補正されるので問題ない
}

// y0～y1 の範囲内
errCode_t ColorKanjiLCDVPlane::scrollUp (uint8_t y0, uint8_t y1, uint8_t ySize)
{
  errCode_t res = CKLCD_ERR_NONE;
  if (ySize) { // スクロール量が 1ライン以上ある
    uint8_t yEnd = (uint8_t)(getHeight() - 1); // Y端
    if (y0 <= yEnd) { // 開始ラインがY端以内である
      if (y1 > yEnd) { // 終了ラインがY端より大きい
        y1 = yEnd; // 終了ラインをY端にする
      }
      if (y0 <= y1) { // 開始ライン≦終了ラインである
        uint8_t xEnd   = (uint8_t)(getWidth() - 1); // X端
        uint8_t yRange = y1 - y0; // スクロールの範囲
        if (ySize > yRange) { // スクロール量の方が、スクロールの範囲より大きい
          ySize = yRange + 1; // スクロール量をスクロールの範囲に収める
        } else {
          // スクロールの範囲よりスクロール量の方が小さい場合、画像のコピーが発生する
          res = setSrcRect(0, y0 + ySize, xEnd, y1); // コピー元の矩形領域を指定
          if (res == CKLCD_ERR_NONE) {
            res = copy(0, y0);  // 上記で指定したコピー元矩形領域を、コピー先にコピーする
          }
        }
        if (res == CKLCD_ERR_NONE) {
          res = clearRect(0, y1 - ySize + 1, xEnd, y1); // 空いた領域をクリアする
        }
      }
    }
  }
  return res;
}

// 全領域
errCode_t ColorKanjiLCDVPlane::scrollDown (uint8_t ySize)
{
  return scrollDown (0, 255, ySize);  // 終了ライン=255 でも、scrollDown(y0, y1, ySize) の方で Y端に補正されるので問題ない
}

// y0～y1 の範囲内
errCode_t ColorKanjiLCDVPlane::scrollDown (uint8_t y0, uint8_t y1, uint8_t ySize)
{
  errCode_t res = CKLCD_ERR_NONE;
  if (ySize) { // スクロール量が 1ライン以上ある
    uint8_t yEnd = (uint8_t)(getHeight() - 1); // Y端
    if (y0 <= yEnd) { // 開始ラインがY端以内である
      if (y1 > yEnd) { // 終了ラインがY端より大きい
        y1 = yEnd; // 終了ラインをY端にする
      }
      if (y0 <= y1) { // 開始ライン≦終了ラインである
        uint8_t xEnd   = (uint8_t)(getWidth() - 1); // X端
        uint8_t yRange = y1 - y0; // スクロールの範囲
        if (ySize > yRange) { // スクロール量の方が、スクロールの範囲より大きい
          ySize = yRange + 1; // スクロール量をスクロールの範囲に収める
        } else {
          // スクロールの範囲よりスクロール量の方が小さい場合、画像のコピーが発生する
          res = setSrcRect(0, y0, xEnd, y1 - ySize); // コピー元の矩形領域を指定
          if (res == CKLCD_ERR_NONE) {
            res = copy(0, y0 + ySize);  // 上記で指定したコピー元矩形領域を、コピー先にコピーする
          }
        }
        if (res == CKLCD_ERR_NONE) {
          res = clearRect(0, y0, xEnd, y0 + ySize - 1); // 空いた領域をクリアする
        }
      }
    }
  }
  return res;
}

// フォントがモノスペース(固定幅)の時は、文字の幅が固定なので setCursor() で指定できます（もちろん setCursorMin() でもOKです）。
// プロポーショナル(可変幅)の時は、文字の幅が一意ではありませんので、より詳細な位置を指定できる setCursorMin() でカーソル位置を指定して下さい。

// 画面はそのままで、LCD 表示開始座標とカーソル位置の両方を (0,0) にします。
errCode_t ColorKanjiLCDVPlane::home (void)
{
  errCode_t res = setDispStartPos (0, 0); // LCD 表示開始座標を (0,0) にします。
  if (res == CKLCD_ERR_NONE) {
    res = setCursorMin (0, 0);  // カーソル位置を (0,0) にします。
  }
  return res;
}


// カーソル位置(キャラクタ座標)を設定: フォントがモノスペース(固定幅)の時に使用できます。
// ※ 実際は内部で setCursorMin(col*fontWidth, row*fontFeight) を呼び出しているだけです。
errCode_t ColorKanjiLCDVPlane::setCursor (uint8_t col, uint8_t row)
{
  if (col >= CKLCD_VPLANE1_MAX_P_WIDTH / _currFontInfo[CKLCD_CURR_FONT_INFO_WIDTH]) { // 範囲外
    return CKLCD_ERR_INVALID_PARAM; // 引数が不正
  }
  if (row >= CKLCD_VPLANE1_MAX_L_HEIGHT / _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT]) { // 範囲外
    return CKLCD_ERR_INVALID_PARAM; // 引数が不正
  }

  return _core.command3 (CKLCD_CMD_VP1_SET_CURSOR_POS + _vplane_no, // 仮想画面#n の カーソル位置設定:  引数 x,y
    col * _currFontInfo[CKLCD_CURR_FONT_INFO_WIDTH],
    row * _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT]);
}

// 現在のカーソル位置(キャラクタ座標)を取得します。フォントがモノスペース(固定幅)の時に使用できます。
errCode_t ColorKanjiLCDVPlane::getCursor (uint8_t * col, uint8_t * row)
{
  // カラー漢字 LCD シールドの仮想画面のカーソル位置を４バイト要求します。
  errCode_t res = _core.command (CKLCD_CMD_VP1_GET_CURSOR_POS + _vplane_no); // 仮想画面#n の カーソル位置取得コマンド  : 戻り y, x

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた

    uint8_t pos[4];
    res = _core.recvData(pos, 4); // ４バイト受信する

    if (res == CKLCD_ERR_NONE) { // データが受信できた
      if (col) {
        uint16_t x = MAKE_UINT16(pos[0], pos[1]);  // x座標
        *col = (uint8_t)(x / _currFontInfo[CKLCD_CURR_FONT_INFO_WIDTH]);
      }
      if (row) {
        uint16_t y = MAKE_UINT16(pos[2], pos[3]);  // y座標
        *row = (uint8_t)(y / _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT]);
      }
    }
  }
  return res;
}

// カーソル位置(グラフィック座標)を設定: x=0～CKLCD_VPLANEn_MAX_o_WIDTH-1, y=0～CKLCD_VPLANEn_MAX_o_HEIGHT-1 の範囲で指定。nは1か2、oはLかPです。
// フォントがプロポーショナル(可変幅)モードの時は、こちらを使用して下さい。
errCode_t ColorKanjiLCDVPlane::setCursorMin (uint8_t x, uint8_t y)  // Min は "minute" の意味
{
  return _core.command3 (CKLCD_CMD_VP1_SET_CURSOR_POS + _vplane_no, x, y); // 仮想画面#n の カーソル位置設定:  引数 x,y
}

// 現在のカーソル位置(グラフィック座標)を取得します。
// 型が uint16_t * なのは、カーソル移動の結果、仮想画面の範囲外に出ている可能性が有るためです。
errCode_t ColorKanjiLCDVPlane::getCursorMin (uint16_t * x, uint16_t * y)  // Min は "minute" の意味
{
  // カラー漢字 LCD シールドの仮想画面のカーソル位置を４バイト要求します。
  errCode_t res = _core.command (CKLCD_CMD_VP1_GET_CURSOR_POS + _vplane_no); // 仮想画面#n の カーソル位置取得コマンド  : 戻り y, x

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた

    uint8_t pos[4];
    res = _core.recvData(pos, 4); // ４バイト受信する

    if (res == CKLCD_ERR_NONE) { // データが受信できた
      if (x) {
        *x = MAKE_UINT16(pos[0], pos[1]);  // x座標
      }
      if (y) {
        *y = MAKE_UINT16(pos[2], pos[3]);  // y座標
      }
    }
  }
  return res;
}


// 文字コード(UTF-8/SJIS/EUC-JP)転送：
//   フリップモードが CKLCD_FLIP_MODE_AUTO の時は、仮想画面に文字を描画後、自動的に仮想画面から LCD 画面に転送されます。
//   一方、フリップモードが CKLCD_FLIP_MODE_MANU の時は、仮想画面に文字が描画されるだけで、
//   自動的に LCD 画面には転送されませんので、ユーザーが明示的に flip() を呼び出す必要があります。

//   モノスペース(固定幅)の時、カーソル位置は、半角文字表示後に +8、全角文字表示後に +16 されます。
//   プロポーショナル(可変幅)の時は、漢字表示後に +16、それ以外の文字の時は +5～+16 のいずれかの値が加算されます。

//   カーソル位置が仮想画面の右端に来ても自動的に改行はされず、カーソル位置は仮想画面の範囲外のままです。
//   0x0D('\r')の復帰コードで、行頭に移動します。つまり (0,y) のカーソル位置に移動します。
//   また、0x0A('\n')の改行コードで改行されます。つまり (0,y+20) のカーソル位置に移動します。
//   カーソル位置が仮想画面の下端を越えても自動的にスクロールとかはされず、カーソル位置は仮想画面の範囲外のままです。

//   カーソル位置が仮想画面の範囲外の状態で write/print しても、LCD画面上はおろか、仮想画面上にすら文字は描画されません。
//   ユーザーが明示的に、仮想画面上の描画可能域に、カーソル位置を移動させる必要があります。


// 文字コードについて
// 以下の文字コードは、画面に文字が表示されること無く、以下の動作を行います

// 0x0D('\r')：復帰：カーソル位置を同じ行の行頭に移動

// 0x0A('\n')：改行：カーソル位置を次の行の行頭に移動 (単純な LineFeed ではなく、これだけで本来の CR+LF の動作になります)

// 文字コードを１バイト転送する
// ※基底クラスの Print のメンバー関数から呼び出される関数です。ユーザーが直接呼び出す事はありません。
size_t ColorKanjiLCDVPlane::write (uint8_t dat)
{
  // カラー漢字 LCD シールドの方でマルチバイト処理しますので、どんな符号化方式でも、１バイトずつ転送するだけで OK です
  errCode_t res = _core.command2 (CKLCD_CMD_VP1_TEXT_OPAQ_OVER + _merge, dat); // 仮想画面#n へテキスト転送コマンド:  引数 byte, ...

  if (res != CKLCD_ERR_NONE) { // 何らかのエラーが発生した
    return 0; // 転送できなかった
  }
  return 1; // １バイト転送した
}

// 文字コードを数バイトまとめて転送する：基底クラスの Print の関数をオーバーライド
// 数バイトまとめてテキスト転送しても OK です。こちらの方が、TWI 通信のオーバーヘッドが少なくなるので、若干早いです。
// ただし、カラー漢字 LCD シールド側は、コマンドバイト(CKLCD_CMD_TEXT)含めて 256 バイトまで受信可能なのですが、
// Arduino IDE 付属の Wire ライブラリが、コマンドバイト含めて 32バイトまで、という制約 (txBuffer[32]) がある為、
// 第２引数の size に指定された値が 32 以上の場合は、分割してテキスト転送する様にしています。
size_t ColorKanjiLCDVPlane::write (const uint8_t * buffer, size_t size)
{
  size_t total = 0; // 総転送バイト数
  errCode_t res;

  do {

    res = _core.waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ
    if (res != CKLCD_ERR_NONE) { // 何らかのエラーが発生した
      break;
    }

    size_t n = size;
    if (n > (BUFFER_LENGTH - 1)) { // n が 32バイト以上の時は、31バイトだけに制限する ※Wire ライブラリの制約の為
      n = (BUFFER_LENGTH - 1);
    }

    _core.sendBegin(); // カラー漢字 LCD シールドとの TWI 通信の開始
    _core.sendData (CKLCD_CMD_VP1_TEXT_OPAQ_OVER + _merge); // 仮想画面#n へテキスト転送コマンド:  引数 byte, ...

    while (n--) {
      uint8_t dat = *buffer; // １バイト取り出し
      if (_core.sendData(dat) == 0) { // Wire ライブラリの txBuffer[] 末尾に詰め込む
        break; // _core.sendData 関数が 0 を返した時は送信バッファが一杯なので、while 文を抜け出す (32バイト制限しているので、この現象は発生しないと思いますが、念のため)
      }
      buffer++;
      total++; // 総転送バイト数のカウント
      size--;
    }

    res = _core.sendEnd(); // カラー漢字 LCD シールドとの TWI 通信の終了とレスポンスの取得
    if (res != CKLCD_ERR_NONE) { // 何らかのエラーが発生した
      break;
    }
  } while (size); // 指定バイト分を転送し終わるまで do ～ while 文を繰り返す

  if (res != CKLCD_ERR_NONE) { // 何らかのエラーが発生した
    return 0; // 転送できなかった
  }
  return total; // total バイト転送した
}


// プログラムメモリ上の文字列を転送する
size_t ColorKanjiLCDVPlane::print_P (const prog_char * pstr)
{
  return write_P(pstr, strlen_P(pstr)); // strlen_P 関数で文字数をカウントして、write_P 関数に投げる
}

// プログラムメモリ上の文字コードを指定バイト分まとめて転送する
size_t ColorKanjiLCDVPlane::write_P (const prog_char * buffer, size_t size)
{
  size_t total = 0; // 総転送バイト数
  errCode_t res;

  do {

    res = _core.waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ
    if (res != CKLCD_ERR_NONE) { // 何らかのエラーが発生した
      break;
    }

    size_t n = size;
    if (n > (BUFFER_LENGTH - 1)) { // n が 32バイト以上の時は、31バイトだけに制限する
      n = (BUFFER_LENGTH - 1);
    }

    _core.sendBegin(); // カラー漢字 LCD シールドとの TWI 通信の開始
    _core.sendData (CKLCD_CMD_VP1_TEXT_OPAQ_OVER + _merge); // 仮想画面#n へテキスト転送コマンド:  引数 byte, ...

    while (n--) {
      uint8_t dat = pgm_read_byte(buffer); // プログラムメモリから１バイト取り出し
      if (_core.sendData(dat) == 0) { // Wire ライブラリの txBuffer[] 末尾に詰め込む
        break; // _core.sendData 関数が 0 を返した時は送信バッファが一杯なので、while 文を抜け出す (32バイト制限しているので、この現象は発生しないと思いますが、念のため)
      }
      buffer++;
      total++; // 総転送バイト数のカウント
      size--;
    }

    res = _core.sendEnd(); // カラー漢字 LCD シールドとの TWI 通信の終了とレスポンスの取得
    if (res != CKLCD_ERR_NONE) { // 何らかのエラーが発生した
      break;
    }
  } while (size); // 指定バイト分を転送し終わるまで do ～ while 文を繰り返す

  if (res != CKLCD_ERR_NONE) { // 何らかのエラーが発生した
    return 0; // 転送できなかった
  }
  return total; // total バイト転送した
}


// ワイド文字(UTF-16)対応
// ASCII 文字が少なく、和文が多い場合は、こちらの方がメモリ効率が良くなります。
// ただし、Arduino IDE は、ワイド文字に対応した文字列操作関数が存在しないので、文字列の操作に不自由します。
// なお、以下の関数を使用する時は、先に setEncoding 関数で UTF-16 方式に切り替えておいてください。

// UTF-16 文字を１文字転送する
size_t ColorKanjiLCDVPlane::wprint (wchar_t wch)
{
  size_t size = write((const uint8_t *)&wch, sizeof(wchar_t));
  if (size == 0) {
    return 0; // 何らかのエラーが発生した
  }
  return 1; // １文字転送した
}

// UTF-16 文字列を転送する
size_t ColorKanjiLCDVPlane::wprint (const wchar_t * wstr)
{
  size_t len = 0;

  while (1) { // L'\0' になるまでループ
    wchar_t wch = *wstr; // UTF-16 文字を１文字取り出す
    if (wch == L'\0') { // NUL 文字の時はループアウト
      break;
    }
    wstr++;

    size_t size = write((const uint8_t *)&wch, sizeof(wchar_t));
    if (size == 0) {
      return 0; // 何らかのエラーが発生した
    }

    len++; // ２バイト単位で +1 させる
  }
  return len; // len 文字数、転送した
}

// プログラムメモリ上の UTF-16 文字列を転送する
size_t ColorKanjiLCDVPlane::wprint_P (const prog_wchar * pwstr)
{
  size_t len = 0;

  while (1) { // L'\0' になるまでループ
    wchar_t wch = (wchar_t)pgm_read_word(pwstr); // プログラムメモリ上の UTF-16 文字を１文字取り出す
    if (wch == L'\0') { // NUL 文字の時はループアウト
      break;
    }
    pwstr++;

    size_t size = write((const uint8_t *)&wch, sizeof(wchar_t));
    if (size == 0) {
      return 0; // 何らかのエラーが発生した
    }

    len++; // ２バイト単位で +1 させる
  }
  return len; // len 文字数、転送した
}


// 特殊効果付きの文字列の描画
// 一旦、仮想画面上で処理してから、一気に液晶画面に表示するようにします。この為、内部で ColorKanjiLCDCore::setFlipMode() が呼び出されます。
// 注意：現在の setDrawColor() の設定が変更されてしまいますので、再度、setDrawColor() で論理色を指定しなおす必要があります。なお、setBackColor() の方は変更されません。

// 影付き文字列の描画
// X/Y ともに offset分ずれた位置に colShadow 色の影がある colBody 色の文字列を描画します。
// 注意：余りにも離れた位置に影があると、とても変な影付き文字に感じますので、offset値は程々にしておくべきでしょう。
// 内部では、str を colShadow 色で描画、blur 処理、再度 str を colBody 色でアルファブレンド描画しています。
errCode_t ColorKanjiLCDVPlane::shadowPrint_M (uint8_t x, uint8_t y, uint8_t offset, cklcdCol32_t colBody, cklcdCol32_t colShadow, const void * str, uint8_t memType)
{
  _core._count = 0;
  errCode_t res = _core.getStatus(); // カラー漢字 LCD シールドの現在のステータスを取得

  if (res == CKLCD_ERR_NONE) {
    _core._count++;

    uint8_t oldFlipMode = _core.getFlipMode(); // 現在のフリップモードを保存
    res = _core.setFlipMode (CKLCD_FLIP_MODE_MANU); // 自動フリップをOFF

    if (res == CKLCD_ERR_NONE) {
      _core._count++;

      uint8_t oldBlendMode = _blend; // 現在のアルファブレンドの設定を保存
      setBlendMode (CKLCD_BLEND_ADD); // アルファブレンドON

      // 影文字から描画する
      res = setDrawColor (colShadow); // 影文字の論理色

      if (res == CKLCD_ERR_NONE) {
        _core._count++;

        res = setCursorMin (x + offset, y + offset);  // 影文字の座標

        if (res == CKLCD_ERR_NONE) {
          _core._count++;

          // 影文字として描画
          if (memType) {
            print_P ((const prog_char *)str);
          } else {
            print ((const char *)str);
          }

          uint16_t x1, y1;
          res = getCursorMin (&x1, &y1); // 現在のカーソル位置を取得

          if (res == CKLCD_ERR_NONE) {
            _core._count++;

            y1 += _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT] - 1;

            res = blur (x + offset - 1, y + offset, (uint8_t)x1, (uint8_t)y1); // 影文字に対してブラー処理

            if (res == CKLCD_ERR_NONE) {
              _core._count++;

              // 続いて、本体文字を描画する
              res = setDrawColor (colBody); // 本体文字の論理色

              if (res == CKLCD_ERR_NONE) {
                _core._count++;

                res = setCursorMin (x, y);  // 本体文字の座標

                if (res == CKLCD_ERR_NONE) {
                  _core._count++;

                  // 本体文字として描画
                  if (memType) {
                    print_P ((const prog_char *)str);
                  } else {
                    print ((const char *)str);
                  }

                  res = _core.flip (x, y, (uint8_t)x1, (uint8_t)y1); // 影文字も含めて、強制的に flip させて、LCD 画面に表示させる
                }
              }
            }
          }
        }
      }
      setBlendMode (oldBlendMode); // アルファブレンド設定を元に戻す
      _core.setFlipMode (oldFlipMode); // フリップモードを元に戻す
    }
  }
  return res;
}

// 縁取り付き文字列の描画
// colStroke 色の縁取りが付いた colBody 色の文字列を描画します。遅いです。
// 注意：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate 処理がかかり、おかしくなってしまいます。
// 内部では、str を colStroke 色で描画、dilate 処理、再度 str を colBody 色でアルファブレンド描画しています。
errCode_t ColorKanjiLCDVPlane::strokePrint_M (uint8_t x, uint8_t y, cklcdCol32_t colBody, cklcdCol32_t colStroke, const void * str, uint8_t memType)
{
  _core._count = 0;
  errCode_t res = _core.getStatus(); // カラー漢字 LCD シールドの現在のステータスを取得

  if (res == CKLCD_ERR_NONE) {
    _core._count++;

    uint8_t oldFlipMode = _core.getFlipMode(); // 現在のフリップモードを保存
    res = _core.setFlipMode (CKLCD_FLIP_MODE_MANU); // 自動フリップをOFF

    if (res == CKLCD_ERR_NONE) {
      _core._count++;

      uint8_t oldBlendMode = _blend; // 現在のアルファブレンドの設定を保存
      setBlendMode (CKLCD_BLEND_ADD); // アルファブレンドON

      // 縁取り文字から描画する
      res = setDrawColor (colStroke); // 縁取り文字の論理色

      if (res == CKLCD_ERR_NONE) {
        _core._count++;

        res = setCursorMin (x + 1, y);  // 縁取り文字の座標

        if (res == CKLCD_ERR_NONE) {
          _core._count++;

          // 縁取り文字として描画
          if (memType) {
            print_P ((const prog_char *)str);
          } else {
            print ((const char *)str);
          }

          uint16_t x1, y1;
          res = getCursorMin (&x1, &y1); // 現在のカーソル位置を取得

          if (res == CKLCD_ERR_NONE) {
            _core._count++;

            y1 += _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT] - 1;

            res = dilate (x, y, (uint8_t)x1, (uint8_t)y1); // 縁取り文字に対して膨張処理

            if (res == CKLCD_ERR_NONE) {
              _core._count++;

              // 続いて、本体文字を描画する
              res = setDrawColor (colBody); // 本体文字の論理色

              if (res == CKLCD_ERR_NONE) {
                _core._count++;

                res = setCursorMin (x + 1, y);  // 本体文字の座標

                if (res == CKLCD_ERR_NONE) {
                  _core._count++;

                  // 本体文字として描画
                  if (memType) {
                    print_P ((const prog_char *)str);
                  } else {
                    print ((const char *)str);
                  }

                  res = _core.flip (x, y, (uint8_t)x1, (uint8_t)y1); // 縁取り文字も含めて、強制的に flip させて、LCD 画面に表示させる
                }
              }
            }
          }
        }
      }
      setBlendMode (oldBlendMode); // アルファブレンド設定を元に戻す
      _core.setFlipMode (oldFlipMode); // フリップモードを元に戻す
    }
  }
  return res;
}

// グロー効果の文字列の描画
// colGlow 色のグロー効果が付いた colBody 色の文字列を描画します。すごく遅いです。
// 注意：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate & blur 処理がかかり、おかしくなってしまいます。
// 内部では、str を colGlow 色で描画、dilate 処理、blur 処理x2、再度 str を colBody 色でアルファブレンド描画しています。
errCode_t ColorKanjiLCDVPlane::glowPrint_M (uint8_t x, uint8_t y, cklcdCol32_t colBody, cklcdCol32_t colGlow, const void * str, uint8_t memType)
{
  _core._count = 0;
  errCode_t res = _core.getStatus(); // カラー漢字 LCD シールドの現在のステータスを取得

  if (res == CKLCD_ERR_NONE) {
    _core._count++;

    uint8_t oldFlipMode = _core.getFlipMode(); // 現在のフリップモードを保存
    res = _core.setFlipMode (CKLCD_FLIP_MODE_MANU); // 自動フリップをOFF

    if (res == CKLCD_ERR_NONE) {
      _core._count++;

      uint8_t oldBlendMode = _blend; // 現在のアルファブレンドの設定を保存
      setBlendMode (CKLCD_BLEND_ADD); // アルファブレンドON

      // グロー文字から描画する
      res = setDrawColor (colGlow); // グロー文字の論理色

      if (res == CKLCD_ERR_NONE) {
        _core._count++;

        res = setCursorMin (x + 1, y);  // グロー文字の座標

        if (res == CKLCD_ERR_NONE) {
          _core._count++;

          // グロー文字として描画
          if (memType) {
            print_P ((const prog_char *)str);
          } else {
            print ((const char *)str);
          }

          uint16_t x1, y1;
          res = getCursorMin (&x1, &y1); // 現在のカーソル位置を取得

          if (res == CKLCD_ERR_NONE) {
            _core._count++;

            y1 += _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT] - 1;

            res = dilate (x, y, (uint8_t)x1, (uint8_t)y1); // グロー文字に対して膨張処理

            if (res == CKLCD_ERR_NONE) {
              _core._count++;

              res = blur (x, y, (uint8_t)x1, (uint8_t)y1); // グロー文字に対してブラー処理：１回目

              if (res == CKLCD_ERR_NONE) {
                _core._count++;

                res = blur (x, y, (uint8_t)x1, (uint8_t)y1); // グロー文字に対してブラー処理：２回目

                if (res == CKLCD_ERR_NONE) {
                  _core._count++;

                  // 続いて、本体文字を描画する
                  res = setDrawColor (colBody); // 本体文字の論理色

                  if (res == CKLCD_ERR_NONE) {
                    _core._count++;

                    res = setCursorMin (x + 1, y);  // 本体文字の座標

                    if (res == CKLCD_ERR_NONE) {
                      _core._count++;

                      // 本体文字として描画
                      if (memType) {
                        print_P ((const prog_char *)str);
                      } else {
                        print ((const char *)str);
                      }

                      res = _core.flip (x, y, (uint8_t)x1, (uint8_t)y1); // グロー文字も含めて、強制的に flip させて、LCD 画面に表示させる
                    }
                  }
                }
              }
            }
          }
        }
      }
      setBlendMode (oldBlendMode); // アルファブレンド設定を元に戻す
      _core.setFlipMode (oldFlipMode); // フリップモードを元に戻す
    }
  }
  return res;
}

// ネオン効果の文字列の描画
// colGlow2 色のグロー効果に、さらに colGlow1 色のグロー効果が付いた colBody 色の文字列を描画します。激しく遅いです。
// 注意1：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate & blur 処理がかかり、おかしくなってしまいます。
// 注意2：colGlow1 色と colGlow2 色の組み合わせによっては、２回目の dilate 処理の時に破綻して、変なネオン効果になる場合があります。
// 内部では、str を colGlow2 色で描画、dilate 処理、blur 処理、str を colGlow1 色で再描画、再 dilate 処理、再 blur 処理、str を colBody 色で再々描画しています。
errCode_t ColorKanjiLCDVPlane::neonPrint_M (uint8_t x, uint8_t y, cklcdCol32_t colBody, cklcdCol32_t colGlow1, cklcdCol32_t colGlow2, const void * str, uint8_t memType)
{
  _core._count = 0;
  errCode_t res = _core.getStatus(); // カラー漢字 LCD シールドの現在のステータスを取得

  if (res == CKLCD_ERR_NONE) {
    _core._count++;

    uint8_t oldFlipMode = _core.getFlipMode(); // 現在のフリップモードを保存
    res = _core.setFlipMode (CKLCD_FLIP_MODE_MANU); // 自動フリップをOFF

    if (res == CKLCD_ERR_NONE) {
      _core._count++;

      uint8_t oldBlendMode = _blend; // 現在のアルファブレンドの設定を保存
      setBlendMode (CKLCD_BLEND_ADD); // アルファブレンドON

      // グロー2文字から描画する
      res = setDrawColor (colGlow2); // グロー2文字の論理色

      if (res == CKLCD_ERR_NONE) {
        _core._count++;

        res = setCursorMin (x + 2, y);  // グロー2文字の座標

        if (res == CKLCD_ERR_NONE) {
          _core._count++;

          // グロー2文字として描画
          if (memType) {
            print_P ((const prog_char *)str);
          } else {
            print ((const char *)str);
          }

          uint16_t x1, y1;
          res = getCursorMin (&x1, &y1); // 現在のカーソル位置を取得

          if (res == CKLCD_ERR_NONE) {
            _core._count++;

            x1 += 1;
            y1 += _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT] - 1;

            res = dilate (x, y, (uint8_t)x1, (uint8_t)y1); // グロー2文字に対して膨張処理

            if (res == CKLCD_ERR_NONE) {
              _core._count++;

              res = blur (x, y, (uint8_t)x1, (uint8_t)y1); // グロー2文字に対してブラー処理

              if (res == CKLCD_ERR_NONE) {
                _core._count++;

                // グロー1文字から描画する
                res = setDrawColor (colGlow1); // グロー1文字の論理色

                if (res == CKLCD_ERR_NONE) {
                  _core._count++;

                  res = setCursorMin (x + 2, y);  // グロー1文字の座標

                  if (res == CKLCD_ERR_NONE) {
                    _core._count++;

                    // グロー1文字として描画
                    if (memType) {
                      print_P ((const prog_char *)str);
                    } else {
                      print ((const char *)str);
                    }

                    res = dilate (x, y, (uint8_t)x1, (uint8_t)y1); // グロー1文字に対して膨張処理

                    if (res == CKLCD_ERR_NONE) {
                      _core._count++;

                      res = blur (x, y, (uint8_t)x1, (uint8_t)y1); // グロー1文字に対してブラー処理

                      if (res == CKLCD_ERR_NONE) {
                        _core._count++;

                        // 続いて、本体文字を描画する
                        res = setDrawColor (colBody); // 本体文字の論理色

                        if (res == CKLCD_ERR_NONE) {
                          _core._count++;

                          res = setCursorMin (x + 2, y);  // 本体文字の座標

                          if (res == CKLCD_ERR_NONE) {
                            _core._count++;

                            // 本体文字として描画
                            if (memType) {
                              print_P ((const prog_char *)str);
                            } else {
                              print ((const char *)str);
                            }

                            res = _core.flip (x, y, (uint8_t)x1, (uint8_t)y1); // グロー1/2文字も含めて、強制的に flip させて、LCD 画面に表示させる
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
      setBlendMode (oldBlendMode); // アルファブレンド設定を元に戻す
      _core.setFlipMode (oldFlipMode); // フリップモードを元に戻す
    }
  }
  return res;
}


// グラフィック描画
// 一旦、仮想画面#1/#2 に描いておいてから、ColorKanjiLCD::flip 関数で一括して LCD へ転送する事で、初めて表示されます。

// 仮想画面に点(x,y)を描画：setBlendMode()の手法で setDrawColor()の論理色で描画します。
errCode_t ColorKanjiLCDVPlane::pset (uint8_t x, uint8_t y)
{
  // 点を描画色で描画します。
  return _core.command4 (CKLCD_CMD_PSET, // 仮想画面#n へ点描画: 引数 vp/effect, x, y
    _merge + CKLCD_DRAW_COLOR, // 現在の透過モード、及びブレンド設定に従って描画します
    x, y);
}


// 仮想画面に線(x0,y0)-(x1,y1)を描画：setBlendMode()の手法で setDrawColor()の論理色で描画します。
errCode_t ColorKanjiLCDVPlane::line (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  // 線を描画色で描画します。
  return _core.command6 (CKLCD_CMD_LINE, // 仮想画面#n へ線描画: 引数 vp/effect, x0, y0, x1, y1
    _merge + CKLCD_DRAW_COLOR, // 現在の透過モード、及びブレンド設定に従って描画します
    x0, y0, x1, y1);
}


// 仮想画面に矩形(x0,y0)-(x1,y1)を描画：setBlendMode()の手法で setDrawColor()の論理色で描画します。
// 注意：矩形の範囲が広いと、それだけ処理に時間が掛かります。

// 枠線のみ
errCode_t ColorKanjiLCDVPlane::rect (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  // 指定矩形を描画色で描画します。塗りつぶしません。
  return _core.command6 (CKLCD_CMD_RECT, // 仮想画面#n へ矩形描画: 引数 vp/effect, x0, y0, x1, y1
    _merge + CKLCD_DRAW_COLOR + CKLCD_DRAW_NO_FILL, // 現在の透過モード、及びブレンド設定に従って描画します
    x0, y0, x1, y1);
}

// 内部も塗りつぶす
errCode_t ColorKanjiLCDVPlane::rectFill (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  // 指定矩形領域を描画色で塗りつぶします。
  return _core.command6 (CKLCD_CMD_RECT, // 仮想画面#n へ矩形描画: 引数 vp/effect, x0, y0, x1, y1
    _merge + CKLCD_DRAW_COLOR + CKLCD_DRAW_FILL, // 現在の透過モード、及びブレンド設定に従って描画します。塗りつぶし指定有り。
    x0, y0, x1, y1);
}


// 仮想画面に丸(x,y)を描画：setBlendMode()の手法で setDrawColor()の論理色で描画します。
// 注意：円の範囲が広いと、それだけ処理に時間が掛かります。

// 円周線のみ
errCode_t ColorKanjiLCDVPlane::circle (uint8_t x, uint8_t y, uint8_t radius)
{
  // 円周線を描画色で描画します。塗りつぶしません。
  return _core.command5 (CKLCD_CMD_CIRCLE, // 仮想画面#n へ円描画: 引数 vp/effect, x, y, rad
    _merge + CKLCD_DRAW_COLOR + CKLCD_DRAW_NO_FILL, // 現在の透過モード、及びブレンド設定に従って描画します。
    x, y, radius);
}

// 内部も塗りつぶす
errCode_t ColorKanjiLCDVPlane::circleFill (uint8_t x, uint8_t y, uint8_t radius)
{
  // 指定の円領域を描画色で塗りつぶします。
  return _core.command5 (CKLCD_CMD_CIRCLE, // 仮想画面#n へ円描画: 引数 vp/effect, x, y, rad
    _merge + CKLCD_DRAW_COLOR + CKLCD_DRAW_FILL, // 現在の透過モード、及びブレンド設定に従って描画します。塗りつぶし指定有り。
    x, y, radius);
}


// 仮想画面内でのコピー
// 注意：基本的には、同一の仮想画面内でコピーする事が望ましいです。これらの関数は、２種類ある16bit絶対色の変換は行わず、単なる16bitの値として、そのままコピーするだけです。
//       仮想画面#1と#2にまたがってのコピーもできますが、その場合、16bit絶対色が違っている可能性に気をつける必要があります。
//       片方が ARGB4444 で、もう片方が RGB565 の場合、コピー処理そのものは正しく行われますが、コピー後の画像は滅茶苦茶な色になっているでしょう。
// copy() の場合、同一の仮想画面内でコピー元とコピー先の領域が部分的に重なってる場合でも、正しくコピーされます。
// 一方、copyWithRotate() の方は、コピー元とコピー先の領域は部分的にも重なっていてはいけません（暴走こそしませんが、おかしな結果になります）。

// コピー元の矩形領域(x0,y0)-(x1,y1)を指定します。copy()の前に指示します。この関数単独では仮想画面を変化させる事はありません。
// なお、仮想画面#1と#2にまたがってコピー出来るようにした為、ある瞬間に setSrcRect() で指定できるコピー元矩形領域は、仮想画面#1・#2のどちらか片方だけです。
// 仮想画面#1のコピー元矩形領域と、仮想画面#2のコピー元矩形領域を、同時に別々に設定させる事はできませんので注意して下さい。最後に設定した方だけが有効です。
errCode_t ColorKanjiLCDVPlane::setSrcRect (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  return _core.command5 (CKLCD_CMD_VP1_SET_SRC_RECT + _vplane_no, x0, y0, x1, y1);  // 仮想画面#n のコピー元の座標設定:  引数 x0, y0, x1, y1
}

// setSrcRect()で指定されている矩形領域を(x,y)～(x+w-1,y+h-1)にコピーします。w と h はコピー元矩形領域の幅と高さが、そのまま使用されます。
// 1:1 のコピーだけで、拡大/縮小とかはしません。また、仮想画面の領域からはみ出した部分は、単純にクリッピングされます。
// 仮想画面#1と#2にまたがってのコピーもできますが、上記の注意事項に注意して下さい。
errCode_t ColorKanjiLCDVPlane::copy (uint8_t x, uint8_t y)
{
  return _core.command4 (CKLCD_CMD_COPY, _merge, x, y); // 仮想画面#n へ等倍コピー:  引数 vp/effect, x, y
}

// 回転＆拡大/縮小させながらコピー：劇遅!!
// setSrcRect()で指定されている矩形領域を回転("BAM10"単位)させながら、(x0,y0)-(x1,y1)の範囲内にコピーします。
// invRatio10 倍の拡大/縮小もできますが、単なる近傍点補間しかしてないので、結果画像は汚いです。大昔の家庭用ゲーム機みたいな効果しかありません。
// 回転の中心はそれぞれ、コピー元の矩形領域の中心座標と、コピー先の矩形領域の中心座標、になります。
// 回転＆拡大/縮小の時、コピー元の矩形領域に存在し無い部分は setBackColor()で設定している色になります。
// コピー元の矩形領域と、コピー先の矩形領域は、部分的にも重なっていてはいけません。重なってた場合、おかしな結果画像になってしまいます。
errCode_t ColorKanjiLCDVPlane::copyWithRotate (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, cklcdInvRatio10_t invRatio10, cklcdBAM10_t bam10)
{
  return _core.command10 (CKLCD_CMD_COPY_W_ROT, _merge, x0, y0, x1, y1, lowByte(invRatio10), highByte(invRatio10), lowByte(bam10), highByte(bam10));  // 仮想画面#n へ回転/拡大コピー:  引数 vp/effect, x0, y0, x1, y1, ratio, bam10
}


// 画像処理
// ３つしかありません。基本的に特殊効果付きの文字列に使用する処理だけです。

// 膨張処理：(x0,y0)-(x1,y1)の領域内に 3x3 の膨張処理を行います。
// 文字のグロー/ネオン効果の為に追加したので、かなりインチキくさい処理です。一般的な画像に対して膨張処理させると、見るも無残な結果になるだけ、と思われます。
errCode_t ColorKanjiLCDVPlane::dilate (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  // 現在の透過モードに従って膨張処理します。
  return _core.command6 (CKLCD_CMD_DILATE, _merge, x0, y0, x1, y1);  // 仮想画面#n へ膨張処理:  引数 vp/effect, x0, y0, x1, y1
}

// 収縮処理：(x0,y0)-(x1,y1)の領域内に 3x3 の収縮処理を行います。
// こちらも、かなりインチキくさい処理です。一般的な画像に対して収縮処理させると、見るも無残な結果になるだけ、と思われます。
errCode_t ColorKanjiLCDVPlane::erode (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  // 現在の透過モードに従って収縮処理します。
  return _core.command6 (CKLCD_CMD_ERODE, _merge, x0, y0, x1, y1);  // 仮想画面#n へ収縮処理:  引数 vp/effect, x0, y0, x1, y1
}

// ブラー処理：(x0,y0)-(x1,y1)の領域内に 3x3 のガウシアンフィルター処理を行います。とても遅いです。
// 文字列だけでなく、一般的な画像に対しても効果があります。
errCode_t ColorKanjiLCDVPlane::blur (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1)
{
  // 現在の透過モードに従ってブラー処理します。
  return _core.command6 (CKLCD_CMD_BLUR, _merge, x0, y0, x1, y1);  // 仮想画面#n へブラー処理:  引数 vp/effect, x0, y0, x1, y1
}


// bmp/jpg/png/gif 画像の読込み
// 画像データを仮想画面上に展開します。当たり前ですが、バイト数が多い画像データほど、展開に時間が掛かります。
// なお、サポートしているのは読込み機能のみで、Save(保存)機能はありません。

// 注意：各画像フォーマット毎に色々な制限事項があります。制限事項に逸脱した画像データの場合、読み込みエラーになる可能性が有りますので注意して下さい。

// アルファ値有り(PNG/BMP)or透過色有り(GIF)の画像の場合、仮想画面の状態設定と、ブレンド処理の設定から、以下の様に処理されます。
//     非透過モード ＋ 上書き        :  現在の setBackColor()で設定されている色に対してαブレンドしながら、RGB565 として上書きされる。仮想画面に現在描画されている画像は無視される。
//     非透過モード ＋ αブレンドON  :  仮想画面に現在描画されている画像に対してαブレンドされ、RGB565 として書き込まれる。アルファ値は保持されない。
//     透過モード   ＋ 上書き        :  アルファ値を保持したまま、ARGB4444 として上書きされる。仮想画面に現在描画されている画像は無視される。
//     透過モード   ＋ αブレンドON  :  仮想画面に現在描画されている画像に対してαブレンドされ、ARGB4444 として書き込まれる。アルファ値は保持される。
// アルファ値なし(JPG/PNG/BMP)or透過色なし(GIF)の画像の場合は、ブレンド処理の現在の設定は無視されて、以下の様に処理されます。
//     非透過モード  :  完全不透明の  RGB565  として上書きされる。仮想画面に現在描画されている画像は無視される。
//     透過モード    :  完全不透明の ARGB4444 として上書きされる。仮想画面に現在描画されている画像は無視される。

// 搭載メモリー容量の関係上、img_size バイト分の画像データは一気に読込めないため、cklcdLoadImgCallback_t 型のユーザー定義 callback 関数を使って、ちまちまと読込みます。
// カラー漢字LCDシールド側から、buff_size バイト分の画像データが欲しいと要求してきますので、callback 関数内で buff ポインタが示す配列に画像データを転送して下さい。

// 画像読み込みループ関数：内部使用
// ※ LoadImg 時の CKLCD→Host への status
//     0固定: S6:S5:S4:S3:S2:S1:S0   転送バイト(1～127)
//     1固定: E6:E5:E4:E3:E2:E1:E0   転送中断＋エラーコード

errCode_t ColorKanjiLCDVPlane::loopLoadImg (cklcdLoadImgCallback_t funcLoad, cklcdImgFrameCallback_t funcFrame, ColorKanjiLCDLoadImgSize * user_def_data)
{
  uint8_t stat, index, buff[32];
  errCode_t res;

  do {
    res = _core.recvData (&stat, 1); // カラー漢字 LCD シールドから、転送要請ステータスを１バイト受信する

    if (res == CKLCD_ERR_NONE) { // ステータスが受信できた

      if (stat & 0x80) { // bit7==1 の時は、カラー漢字 LCD シールドから中断要請が来た

        if (stat == CKLCD_ERR_LOAD_IMG_END) {  // エラーではなく、画像読込み終了の通知だった
          // 無事に画像転送できた
          return getLoadImgSize (user_def_data);  // 画像のサイズを取得して、ループ終了
        }

        // やっぱりエラーだった
        return stat;  // それ以外の時は CKLCD_ERR_XXX として、呼び出し元への返り値とする
      }

      // bit7==0 の時は、カラー漢字LCD シールドからデータ転送要請が来た
      uint8_t buff_size = stat;  // 転送するバイト数は 1～31 の範囲
      if (buff_size == 0) { // 0が来る事は無いはずだが、もし来た場合は中断コマンド発行して、終了する。
        // カラー漢字LCD シールドへ、読み込み処理を中断するように通知する
        return _core.command (CKLCD_CMD_LOAD_IMG_CANCEL);
      }
      if (buff_size > 31) {
        buff_size = 31;  // Wireライブラリの制限で、コマンドバイト含めて32バイトまでなので、転送できるデータは31バイトまで
      }

      if (funcFrame) { // ユーザー定義 callback 関数(Frame) が NULL でない
        res = _core.recvData (&stat, 1); // カラー漢字 LCD シールドから、エラー状態ステータスを１バイト受信する
        if (res == CKLCD_ERR_NONE) { // ステータスが受信できた
          if (stat == CKLCD_ERR_IMG_LOADING_FRAME_END) {
            // 画像読み込みの最中だが、各フレームの終端に達した：アニメーション GIF の２回目以降の 0x2C インディケータ検出直後に、この状態になります。
            res = getLoadImgSize (user_def_data);  // フレームのサイズを取得する
            if (res == CKLCD_ERR_NONE) { // ステータスが受信できた
              (*funcFrame)(user_def_data); // ユーザー定義 callback 関数(Frame) を呼び出す
            }
          }
        }
      }

      res = (*funcLoad)(buff, buff_size, user_def_data); // ユーザー定義 callback 関数(Load) を呼び出す

      if (res != CKLCD_ERR_NONE) {  // ユーザー定義 callback 関数が、中断する様に言ってきた
        // カラー漢字LCD シールドへ、読み込み処理を中断するように通知する
        res = _core.command (CKLCD_CMD_LOAD_IMG_CANCEL);

      } else {  
        // ユーザー定義 callback 関数が OK を返した

        res = _core.waitWhileBusy(); // カラー漢字 LCD シールドがコマンド受付け可能になるまで待つ
        if (res == CKLCD_ERR_NONE) { // コマンド受付け可能

          _core.sendBegin(); // カラー漢字 LCD シールドとの TWI 通信の開始
          _core.sendData (CKLCD_CMD_LOAD_IMG_XLAT); // 要求された画像データを転送:  引数 data, ...

          for (index = 0; index < buff_size; index++) {
            _core.sendData (buff[index]); // Wire ライブラリの txBuffer[] 末尾に詰め込む
          }

          res = _core.sendEnd(); // カラー漢字 LCD シールドとの TWI 通信の終了とレスポンスの取得
        }
      }
    }
  } while (res == CKLCD_ERR_NONE); // エラーにならない場合は無限ループする

  return res;
}

// 最後に LoadImg した画像のサイズを取得する関数：内部使用
errCode_t ColorKanjiLCDVPlane::getLoadImgSize (ColorKanjiLCDLoadImgSize * user_def_data)
{
  // カラー漢字 LCD シールドから８バイト要求します。
  errCode_t res = _core.command2 (CKLCD_CMD_GET_LOAD_IMG_SIZE, _vplane_no);  // 仮想画面#n の LoadImg した画像サイズ取得: 引数 vp  : 戻り x16, y16, w16, h16

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた

    uint8_t siz[8];
    res = _core.recvData(siz, 8); // ８バイト受信する

    if (res == CKLCD_ERR_NONE) { // データが受信できた
      if (user_def_data) {
        user_def_data->x_offset = MAKE_UINT16(siz[0], siz[1]);  // xオフセット
        user_def_data->y_offset = MAKE_UINT16(siz[2], siz[3]);  // yオフセット
        user_def_data->width    = MAKE_UINT16(siz[4], siz[5]);  // 幅
        user_def_data->height   = MAKE_UINT16(siz[6], siz[7]);  // 高さ
      }
    }
  }
  return res;
}


// jpg 画像の読込み
// Chan 氏作成の TJpgDec を使用しています。
// 読み込める jpg 画像は、一般的な JFIFフォーマットのみです。ガンマ値の JFIFタグや、テキスト等の画像データに関与しない JFIFタグは、在っても無視します。
// また、プログレッシブ(Progressive)タイプの jpg は扱えません。必ず CKLCD_ERR_JPG_UNSUP_PRGRS エラーになりますので、お絵かきソフト等で非プログレッシブに変換しておいてください。
// それと、YCbCr 表色系でない JFIF フォーマットや、一般的には使われないサンプリングファクターが使用された JFIF フォーマットも、エラーになります（TJpgDec が対応してない為）。
// jpg 画像のサイズが(w,h)ならば、仮想画面の(x,y)～(x+w-1,y+h-1)の範囲に描画します。ただし、仮想画面の領域からはみ出した部分はクリッピングされます。
// jpg 画像にアルファ値は有りませんので、setBlendMode() の設定に関わらず、常に上書きされます(アルファブレンドされません)。
// なお、仮想画面が透過モードの時は、論理アルファ値==1.0(不透明)として描画します。

// ファイルサイズが 64kB未満の時
errCode_t ColorKanjiLCDVPlane::loadJpg (int16_t xPos, int16_t yPos, uint8_t effect, uint16_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data)
{
  if (funcLoad == NULL) {  // callback関数が NULL
    return CKLCD_ERR_INVALID_PARAM;
  }

  effect = (effect & CKLCD_LOAD_IMG_EFFECT_BITMASK) + CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31;  // TWI通信時の画像データ転送時のMAXバッファサイズ (Arduino UNO では 31byteのみ指定可能)
  errCode_t res = _core.command9 (CKLCD_CMD_LOAD_JPG, _merge + effect, lowByte(xPos), highByte(xPos), lowByte(yPos), highByte(yPos), 
      lowByte(file_size), highByte(file_size), 0);  // 仮想画面#n へ JPG 展開: 引数 vp/effect, x16, y16, size24

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = loopLoadImg (funcLoad, NULL, user_def_data);  // 画像読み込みループ関数を呼び出す
  }
  return res;
}

// ファイルサイズが 64kB以上の時
errCode_t ColorKanjiLCDVPlane::loadJpgLarge (int16_t xPos, int16_t yPos, uint8_t effect, uint32_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data)
{
  if (funcLoad == NULL) {  // callback関数が NULL
    return CKLCD_ERR_INVALID_PARAM;
  }

  effect = (effect & CKLCD_LOAD_IMG_EFFECT_BITMASK) + CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31;  // TWI通信時の画像データ転送時のMAXバッファサイズ (Arduino UNO では 31byteのみ指定可能)
  errCode_t res = _core.command9 (CKLCD_CMD_LOAD_JPG, _merge + effect, lowByte(xPos), highByte(xPos), lowByte(yPos), highByte(yPos), 
      lowByte(lowWord(file_size)), highByte(lowWord(file_size)), lowByte(highWord(file_size)));  // 仮想画面#n へ JPG 展開: 引数 vp/effect, x16, y16, size24

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = loopLoadImg (funcLoad, NULL, user_def_data);  // 画像読み込みループ関数を呼び出す
  }
  return res;
}


// png 画像の読込み
// 注意!!：読込む仮想画面では無い方の仮想画面を作業エリアとして破壊します。例えば、仮想画面#1 に読み込む時は、仮想画面#2 の内容が破壊されます。逆の時も同じです。
// libpng v1.6.1 と zlib 1.2.8 を使用しています。ただし、かなり改変しています（オリジナル版だとメモリ消費量が膨大な為、必要最小限まで削りました）。
// 読み込める png 画像は、0x89+"PNG" から開始する一般的なフォーマットのみです。
// 処理できるタグは、表示に必要な最低限レベルである IDAT, IEND, IHDR, PLTE, sBIT, tRNS タグのみです。これ以外のタグは、在っても無視されます。
// この為、ガンマ補正や色空間設定などは読み飛ばして無視されます。また、対応しているチャンネル毎のbit深度(depth)は1/2/4/8bitまでです。16bit深度には対応していません(エラーになります)。
// なお、ここで言う16bit深度とは、16bitグレー、16bitグレー＋16bitアルファ、48bitR/G/B、48bitR/G/B＋16bitアルファ、の４種類を指します。
// また、インターレース形式の png は、１ラインごと展開する方式では余りに複雑なので、非サポートです。必ずエラーになりますので注意して下さい。
// 読み込める png 画像は、最大 256×256 画素までです。これよりも大きい場合は、画像サイズエラーになります。
// png 画像のサイズが(w,h)ならば、仮想画面の(x,y)～(x+w-1,y+h-1)の範囲に描画します。ただし、仮想画面の領域からはみ出した部分はクリッピングされます。
// アルファ値を持つ png 画像の場合、CKLCD_BLEND_xxx に従ってブレンド処理されます。アルファ値無しの時は常に上書き、です。

// ファイルサイズが 64kB未満の時
errCode_t ColorKanjiLCDVPlane::loadPng (int16_t xPos, int16_t yPos, uint8_t effect, uint16_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data)
{
  if (funcLoad == NULL) {  // callback関数が NULL
    return CKLCD_ERR_INVALID_PARAM;
  }

  effect = (effect & CKLCD_LOAD_IMG_EFFECT_BITMASK) + CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31;  // TWI通信時の画像データ転送時のMAXバッファサイズ (Arduino UNO では 31byteのみ指定可能)
  errCode_t res = _core.command9 (CKLCD_CMD_LOAD_PNG, _merge + effect, lowByte(xPos), highByte(xPos), lowByte(yPos), highByte(yPos), 
      lowByte(file_size), highByte(file_size), 0);  // 仮想画面#n へ PNG 展開: 引数 vp/effect, x16, y16, size24

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = loopLoadImg (funcLoad, NULL, user_def_data);  // 画像読み込みループ関数を呼び出す
  }
  return res;
}

// ファイルサイズが 64kB以上の時
errCode_t ColorKanjiLCDVPlane::loadPngLarge (int16_t xPos, int16_t yPos, uint8_t effect, uint32_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data)
{
  if (funcLoad == NULL) {  // callback関数が NULL
    return CKLCD_ERR_INVALID_PARAM;
  }

  effect = (effect & CKLCD_LOAD_IMG_EFFECT_BITMASK) + CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31;  // TWI通信時の画像データ転送時のMAXバッファサイズ (Arduino UNO では 31byteのみ指定可能)
  errCode_t res = _core.command9 (CKLCD_CMD_LOAD_PNG, _merge + effect, lowByte(xPos), highByte(xPos), lowByte(yPos), highByte(yPos), 
      lowByte(lowWord(file_size)), highByte(lowWord(file_size)), lowByte(highWord(file_size)));  // 仮想画面#n へ PNG 展開: 引数 vp/effect, x16, y16, size24

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = loopLoadImg (funcLoad, NULL, user_def_data);  // 画像読み込みループ関数を呼び出す
  }
  return res;
}


// gif 画像の読込み
// 注意!!：読込む仮想画面では無い方の仮想画面を作業エリアとして破壊します。例えば、仮想画面#1 に読み込む時は、仮想画面#2 の内容が破壊されます。逆の時も同じです。
// giflib 5.0.4 を使用しています。
// 読み込める gif 画像は、"GIF87A" もしくは "GIF89A" から開始する一般的なフォーマットのみです。
// 読み込める gif 画像は、最大 256×256 画素までです。これよりも大きい場合は、画像サイズエラーになります。
// gif 画像のサイズが(w,h)ならば、仮想画面の(x,y)～(x+w-1,y+h-1)の範囲に描画します。ただし、仮想画面の領域からはみ出した部分はクリッピングされます。
// アニメーション gif は、各フレームごとに、ユーザー定義関数の funcFrame が呼び出されますので、アニメーションさせる事が可能です。
// ただし、大きな画像になると、読み込み＆表示に時間が掛かるので、スムースなアニメーションにはならなくなります。また、呼び出せる関数に制限があります。
// なお、仮想画面が透過モードの時は、透過ピクセル部分は論理アルファ値==0.0(透明)、それ以外の画素は論理アルファ値==1.0(不透明)として描画します。
// 透過ピクセルが有る gif 画像を Back面に上書き(CKLCD_BLEND_OVER)する時は、setBackColor()で設定している色になります。

// ファイルサイズが 64kB未満の時
errCode_t ColorKanjiLCDVPlane::loadGif (int16_t xPos, int16_t yPos, uint8_t effect, uint16_t file_size, cklcdLoadImgCallback_t funcLoad, cklcdImgFrameCallback_t funcFrame, ColorKanjiLCDLoadImgSize * user_def_data)
{
  if (funcLoad == NULL) {  // callback関数が NULL
    return CKLCD_ERR_INVALID_PARAM;
  }

  effect = (effect & CKLCD_LOAD_IMG_EFFECT_BITMASK) + CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31;  // TWI通信時の画像データ転送時のMAXバッファサイズ (Arduino UNO では 31byteのみ指定可能)
  errCode_t res = _core.command9 (CKLCD_CMD_LOAD_GIF, _merge + effect, lowByte(xPos), highByte(xPos), lowByte(yPos), highByte(yPos), 
      lowByte(file_size), highByte(file_size), 0);  // 仮想画面#n へ GIF 展開: 引数 vp/effect, x16, y16, size24

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = loopLoadImg (funcLoad, funcFrame, user_def_data);  // 画像読み込みループ関数を呼び出す
  }
  return res;
}

// ファイルサイズが 64kB以上の時
errCode_t ColorKanjiLCDVPlane::loadGifLarge (int16_t xPos, int16_t yPos, uint8_t effect, uint32_t file_size, cklcdLoadImgCallback_t funcLoad, cklcdImgFrameCallback_t funcFrame, ColorKanjiLCDLoadImgSize * user_def_data)
{
  if (funcLoad == NULL) {  // callback関数が NULL
    return CKLCD_ERR_INVALID_PARAM;
  }

  effect = (effect & CKLCD_LOAD_IMG_EFFECT_BITMASK) + CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31;  // TWI通信時の画像データ転送時のMAXバッファサイズ (Arduino UNO では 31byteのみ指定可能)
  errCode_t res = _core.command9 (CKLCD_CMD_LOAD_GIF, _merge + effect, lowByte(xPos), highByte(xPos), lowByte(yPos), highByte(yPos), 
      lowByte(lowWord(file_size)), highByte(lowWord(file_size)), lowByte(highWord(file_size)));  // 仮想画面#n へ GIF 展開: 引数 vp/effect, x16, y16, size24

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = loopLoadImg (funcLoad, funcFrame, user_def_data);  // 画像読み込みループ関数を呼び出す
  }
  return res;
}


// bmp 画像の読込み
// 読み込める bmp 画像は、"BM" から開始する一般的なフォーマットのみです。
// 自作の展開エンジンにて展開しています。
// bmp 画像のサイズが(w,h)ならば、仮想画面の(x,y)～(x+w-1,y+h-1)の範囲に描画します。ただし、仮想画面の領域からはみ出した部分はクリッピングされます。
// なお、無圧縮(BI_RGBかBI_BITFIELDS)の bmp 画像にのみ対応しており、RLE 圧縮(BI_RLE4/BI_RLE8)の bmp 画像には対応していません。
// また、OS/2フォーマット(BITMAPCOREHEADER: サイズが12byteのヤツ)にも対応していません(今時、そんな bmp 画像ファイルは無いでしょうが)。
// BITMAPINFOHEADER(40byte), BITMAPV4HEADER(108byte), BITMAPV5HEADER(124byte) に対応してます。ただし、色空間、ガンマ値、プロファイルデータは在っても無視します。
// パレット情報(RGBQUAD)を持つ bmp 画像は、そのまま読み込めます。高さ情報(biHeight)が負の bmp 画像も、そのまま読み込めます。
// アルファ値を持つ bmp 画像の場合、CKLCD_BLEND_xxx に従ってブレンド処理されます。アルファ値無しの時は常に上書き、です。

// ファイルサイズが 64kB未満の時
errCode_t ColorKanjiLCDVPlane::loadBmp (int16_t xPos, int16_t yPos, uint8_t effect, uint16_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data)
{
  if (funcLoad == NULL) {  // callback関数が NULL
    return CKLCD_ERR_INVALID_PARAM;
  }

  effect = (effect & CKLCD_LOAD_IMG_EFFECT_BITMASK) + CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31;  // TWI通信時の画像データ転送時のMAXバッファサイズ (Arduino UNO では 31byteのみ指定可能)
  errCode_t res = _core.command9 (CKLCD_CMD_LOAD_BMP, _merge + effect, lowByte(xPos), highByte(xPos), lowByte(yPos), highByte(yPos), 
      lowByte(file_size), highByte(file_size), 0);  // 仮想画面#n へ BMP 展開: 引数 vp/effect, x16, y16, size24

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = loopLoadImg (funcLoad, NULL, user_def_data);  // 画像読み込みループ関数を呼び出す
  }
  return res;
}

// ファイルサイズが 64kB以上の時
errCode_t ColorKanjiLCDVPlane::loadBmpLarge (int16_t xPos, int16_t yPos, uint8_t effect, uint32_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data)
{
  if (funcLoad == NULL) {  // callback関数が NULL
    return CKLCD_ERR_INVALID_PARAM;
  }

  effect = (effect & CKLCD_LOAD_IMG_EFFECT_BITMASK) + CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31;  // TWI通信時の画像データ転送時のMAXバッファサイズ (Arduino UNO では 31byteのみ指定可能)
  errCode_t res = _core.command9 (CKLCD_CMD_LOAD_BMP, _merge + effect, lowByte(xPos), highByte(xPos), lowByte(yPos), highByte(yPos), 
      lowByte(lowWord(file_size)), highByte(lowWord(file_size)), lowByte(highWord(file_size)));  // 仮想画面#n へ BMP 展開: 引数 vp/effect, x16, y16, size24

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた
    res = loopLoadImg (funcLoad, NULL, user_def_data);  // 画像読み込みループ関数を呼び出す
  }
  return res;
}


// 16bit絶対色の操作
// 仮想画面の設定が Front面か Back面により、ARGB4444 か RGB565 のどちらかになります。
// 仮想画面を LCD 表示に使用せず、単なる cklcdCol16_t (uint16_tと同じ) の SRAM 二次元配列として使用する時も、この関数を使用して下さい。

// 仮想画面上の絶対座標(x,y)に、16bit絶対色を上書きします。setBlend()の設定は無視されます。
errCode_t ColorKanjiLCDVPlane::setPixel (uint8_t x, uint8_t y, cklcdCol16_t col16)
{
  return _core.command5 (CKLCD_CMD_VP1_SET_PIXEL + _vplane_no, x, y, lowByte(col16), highByte(col16)); // 仮想画面#n の仮想画面に直接書き込み: 引数 x, y, col16
}

// 仮想画面上の絶対座標(x,y)の、現在の16bit絶対色を取得します。
errCode_t ColorKanjiLCDVPlane::getPixel (uint8_t x, uint8_t y, cklcdCol16_t * col16)
{
  // カラー漢字 LCD シールドの仮想画面から２バイト要求します。
  errCode_t res = _core.command3 (CKLCD_CMD_VP1_GET_PIXEL + _vplane_no, x, y); // 仮想画面#n の仮想画面から直接取得: 引数 x, y  : 戻り col16

  if (res == CKLCD_ERR_NONE) { // コマンドが送信できた

    uint8_t dat[2];
    res = _core.recvData(dat, 2); // ２バイト受信する

    if (res == CKLCD_ERR_NONE) { // データが受信できた
      if (col16) {
        *col16 = MAKE_UINT16(dat[0], dat[1]);
      }
    }
  }
  return res;
}



///////////////////////////////////////////////////////////////
// ColorKanjiLCD : カラー漢字 LCD クラス

// コンストラクタ：カラー漢字 LCD シールドの TWI(i2c) アドレス値を指定する
ColorKanjiLCD::ColorKanjiLCD (uint8_t twiAddr)
  : ColorKanjiLCDCore(twiAddr), v1(CKLCD_VPLANE1, *this), v2(CKLCD_VPLANE2, *this)
{
  // 0 に初期化しておく
  _klcdType = 0;
  _firmVer  = 0;
}


// デストラクタ：特に何もしません
ColorKanjiLCD::~ColorKanjiLCD ()
{
}


// カラー漢字 LCD シールドを初期化します。
// スクリーンモードと符号化方式を指定します。これらは、後で個別に変更することも可能です。
// 正常に実行された場合は、画面クリア、カーソル位置 (0,0)、LCD 表示ON、フリップモード=AUTO、になります
// 注意：カラー漢字 LCD シールドは電源ONした直後、200ms ほど待たないと TWI 通信が開始できません。waitTime にはその待ち時間(ミリ秒)を指定します。
//       デフォルト値は 200 ですが、この関数を呼び出す前に、別の処理で例えば 20ms 待っていれば、その分を差っ引いた 180 にしても構いません。
//    ※ 待ち時間 200ms ですが、カラー液晶モジュールだけで 154ms＋αは確実に掛かっています。Reset 後 10ms の待った後、SleepOut コマンド発行直後に 144ms の wait を入れています。
//       SpecSheet では 120ms なのですが、保険で 1.2 倍した時間(144ms)にしています。さらに SAM4S の初期化、メモリクリアなどの処理を行ってから、TWI 通信の開始をしています。
//       この為、実際は 160～170ms 程度で TWI 通信可能になると思いますが、余裕みて 200ms としています。
errCode_t ColorKanjiLCD::begin (uint8_t screenMode, uint8_t enc, uint8_t waitTime)
{
  if (_klcdType) { // 既に実施済みの時は、直ぐにリターンさせる
    return CKLCD_ERR_NONE;
  }

  // 初期化処理：１回だけ実行すればOK
  _count = 0; // ただのカウンター

  delay (waitTime); // カラー漢字 LCD シールドが、電源ON されてから TWI 通信可能になるまで、waitTime ミリ秒待つ

#ifdef __arm__
      uint8_t dummy[2];
      res = recvData (dummy, 2); // ２バイト受信する
#endif

  res = command (CKLCD_CMD_INIT); // カラー漢字 LCD シールドへ初期化コマンドバイトを送信

  if (res == CKLCD_ERR_NONE) { // 初期化成功
    _count++;

#ifdef __arm__
    delayMicroseconds(500); // 500μ秒待つ
#endif

    res = waitWhileBusy(); // カラー漢字 LCD シールドが次のコマンド受付け可能になるまで待つ

    if (res == CKLCD_ERR_NONE) {
      _count++;

#ifdef __arm__
    delayMicroseconds(100); // 100μ秒待つ
#endif

      uint8_t dat[2];
      res = recvData (dat, 2); // ２バイト受信する

      if (res == CKLCD_ERR_NONE) {
        _count++;

        // 初期化コマンド(CKLCD_CMD_INIT) の直後のみ、以下の２バイトの情報が取得できる。
        // この２バイトは変化しないので、毎回取得する必要性が無い為、初期化コマンド直後に１回だけ取得可能にした
        _klcdType = dat[0]; // １バイト目: カラー漢字 LCD シールドのタイプ (常に "BS21LAB-006" を意味する 0x06 になる)
        _firmVer  = dat[1]; // ２バイト目: カラー漢字 LCD シールドのファームウェアの Version 番号 (例: v1.2 の時は 0x12 になる)

#ifdef __arm__
    delayMicroseconds(100); // 100μ秒待つ
#endif

        res = setScreenMode (screenMode); // スクリーンモードをセット

        if (res == CKLCD_ERR_NONE) {
          _count++;

          res = getStatus();  // カラー漢字 LCD シールドの現在のステータスを取得する

          if (res == CKLCD_ERR_NONE) {
            _count++;

            res = v1.setEncoding(enc); // 仮想画面#1: 符号化方式をセット

            if (res == CKLCD_ERR_NONE) {
              _count++;

              res = v2.setEncoding(enc); // 仮想画面#2: 符号化方式をセット

              if (res == CKLCD_ERR_NONE) {
                _count++;

                res = v1.getFontInfo();  // 仮想画面#1: 現在のフォント情報を取得

                if (res == CKLCD_ERR_NONE) {
                  _count++;

                  res = v2.getFontInfo();  // 仮想画面#2: 現在のフォント情報を取得

                  if (res == CKLCD_ERR_NONE) {
                    _count++;

                    res = v1.clearEx (CKLCD_CLR_FLG_DISP_START_HOME + CKLCD_CLR_FLG_CURSOR_HOME); // 仮想画面#1: 画面クリア

                    if (res == CKLCD_ERR_NONE) {
                      _count++;

                      res = v2.clearEx (CKLCD_CLR_FLG_DISP_START_HOME + CKLCD_CLR_FLG_CURSOR_HOME); // 仮想画面#2: 画面クリア

                      if (res == CKLCD_ERR_NONE) {
                        _count++;

                        res = flip(); // LCD 全域を再描画

                        if (res == CKLCD_ERR_NONE) {
                          _count++;

                          res = v1.getStatus();  // 仮想画面#1: 現在のステータスを取得する

                          if (res == CKLCD_ERR_NONE) {
                            _count++;

                            res = v2.getStatus();  // 仮想画面#2: 現在のステータスを取得する
                          }
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
      }
    }
  }
  return res;
}


// 終了処理：通常、呼び出す事は無いと思いますが…
errCode_t ColorKanjiLCD::end (void)
{
  _klcdType = 0; // 再度、初期化処理できるようにする

  return CKLCD_ERR_NONE;
}


// スクリーンモードの設定
// CKLCD_ORIENT_xxx のいずれかと、CKLCD_DISP_SCREEN_xxx のいずれかを OR 指定します。
// 注意1：CKLCD_ORIENT_xxx を変更した場合、仮想画面#1/#2 それぞれの LCD 表示開始座標とカーソル位置を設定し直してください。
// 注意2：この関数内で ColorKanjiLCDVPlane::setCondition() をセットしています。
errCode_t ColorKanjiLCD::setScreenMode (uint8_t screenMode)
{
  res = command2 (CKLCD_CMD_SET_SCREEN_MODE, screenMode); // スクリーンモード設定コマンド: 引数 mode

  if (res == CKLCD_ERR_NONE) {

    res = getScreenMode();  // 現在のスクリーンモード取得

    if (res == CKLCD_ERR_NONE) {

      // 仮想画面#1/#2 にオリエンテーションモードをセット
      uint8_t orient = _currScreenMode & CKLCD_ORIENT_PORTRAIT;
      if (orient == 0) {
        // ランドスケープモード
        _phyWidth  = CKLCD_MAX_L_WIDTH;  // LCD 物理画面の幅  ：160
        _phyHeight = CKLCD_MAX_L_HEIGHT; // LCD 物理画面の高さ：128
      } else {
        // ポートレートモード
        _phyWidth  = CKLCD_MAX_P_WIDTH;  // LCD 物理画面の幅  ：128
        _phyHeight = CKLCD_MAX_P_HEIGHT; // LCD 物理画面の高さ：160
      }
      v1.setOrientMode (orient);
      v2.setOrientMode (orient);

      // 仮想画面#1/#2 に状態をセット
      switch (_currScreenMode & CKLCD_DISP_SCREEN_BITMASK) {
      case CKLCD_DISP_SCREEN_NONE: // Front面：無し、Back面：無し  ※このモードの時は、仮想画面１と２は LCD に表示されないので作業エリアとして使用可能
      default:  // 不正な CKLCD_DISP_SCREEN_XXX: CKLCD_DISP_SCREEN_NONEと同じ扱いにしておく
        v1.setCondition (CKLCD_VP_COND_OPAQUE);  // 非透過モード: 表示されない作業エリアも、コピー用に OPAQUE にしておく
        v2.setCondition (CKLCD_VP_COND_OPAQUE);  // 非透過モード: 表示されない作業エリアも、コピー用に OPAQUE にしておく
        break;
      case CKLCD_DISP_SCREEN_B1:   // Front面：無し、Back面：仮想画面１(160×256)  ※このモードの時は、仮想画面２は LCD に表示されないので作業エリアとして使用可能
        v1.setCondition (CKLCD_VP_COND_OPAQUE);  // 非透過モード: Back (背景)面は基本的にこの設定
        v2.setCondition (CKLCD_VP_COND_OPAQUE);  // 非透過モード: 表示されない作業エリアも、コピー用に OPAQUE にしておく
        break;
      case CKLCD_DISP_SCREEN_B2:  // Front面：無し、Back面：仮想画面２(160×128)  ※このモードの時は、仮想画面１は LCD に表示されないので作業エリアとして使用可能
        v1.setCondition (CKLCD_VP_COND_OPAQUE);  // 非透過モード: 表示されない作業エリアも、コピー用に OPAQUE にしておく
        v2.setCondition (CKLCD_VP_COND_OPAQUE);  // 非透過モード: Back (背景)面は基本的にこの設定
        break;
      case CKLCD_DISP_SCREEN_F1:     // Front面：仮想画面１(160×256)、Back面：無し  ※このモードの時は、仮想画面２は LCD に表示されないので作業エリアとして使用可能
      case CKLCD_DISP_SCREEN_F1_B2:  // Front面：仮想画面１(160×256)、Back面：仮想画面２(160×128)  ※ LCD に表示時にアルファブレンド処理されます
        v1.setCondition (CKLCD_VP_COND_TRANS);   // 透過モード  : Front(前景)面は基本的にこの設定
        v2.setCondition (CKLCD_VP_COND_OPAQUE);  // 非透過モード: Back (背景)面は基本的にこの設定
        break;
      case CKLCD_DISP_SCREEN_F2:     // Front面：仮想画面２(160×128)、Back面：無し  ※このモードの時は、仮想画面１は LCD に表示されないので作業エリアとして使用可能
      case CKLCD_DISP_SCREEN_F2_B1:  // Front面：仮想画面２(160×128)、Back面：仮想画面１(160×256)  ※ LCD に表示時にアルファブレンド処理されます
        v1.setCondition (CKLCD_VP_COND_OPAQUE);  // 非透過モード: Back (背景)面は基本的にこの設定
        v2.setCondition (CKLCD_VP_COND_TRANS);   // 透過モード  : Front(前景)面は基本的にこの設定
        break;
      }
    }
  }
  return res;
}


// 仮想画面#1&#2 の全域を 0x0000 の絶対色でクリアし、LCD 表示開始座標とカーソル位置を (0,0) にします。
errCode_t ColorKanjiLCD::clear (void)
{
  errCode_t res = v1.clearEx (CKLCD_CLR_FLG_DISP_START_HOME + CKLCD_CLR_FLG_CURSOR_HOME); // 仮想画面#1: 画面クリア

  if (res == CKLCD_ERR_NONE) {
    res = v2.clearEx (CKLCD_CLR_FLG_DISP_START_HOME + CKLCD_CLR_FLG_CURSOR_HOME); // 仮想画面#2: 画面クリア

    if (res == CKLCD_ERR_NONE) {
      if (getFlipMode() == CKLCD_FLIP_MODE_AUTO) { // フリップモードが AUTO モードの時のみ
        res = flip(); // LCD 全域を再描画
      }
    }
  }

  return res;
}


// 以下は、begin 関数内で取得されている情報

// 漢字 LCD シールドのタイプ: 常に"BS21LAB-006"を意味する 0x06 が返される
uint8_t ColorKanjiLCD::getKlcdType (void)
{
  return _klcdType;
}

// 漢字 LCD シールドのファームウェアの Version 番号 (例: v1.2 の時は 0x12 になる)
uint8_t ColorKanjiLCD::getFirmVer (void)
{
  return _firmVer;
}

