/*
 * Copyright (C) 2013 by I.Kato @ BS21 Lab
 * Color Kanji LCD library for Arduino
 *    ver.1.00 (15/aug/2013)
 *
 * 対象     : カラー漢字 LCD シールド (BS21LAB-006)
 * 接続方法 : TWI(i2c) 通信方式
 *
 * 詳細は http://www.bs21-lab.com/products/p006-ColorKanjiLCD/ を御覧下さい。
 *
 * This file is free software; you can redistribute it and/or modify it
 * under the terms of either the GNU General Public License version 2
 * or the GNU Lesser General Public License version 2.1,
 * both as published by the Free Software Foundation.
 */

#ifndef __COLORKANJILCD_H__
#define __COLORKANJILCD_H__


#include "Arduino.h"
#include "Print.h"

#include <../BS21LabLib/BS21LabLib.h>


//////////////////////////////////////////////
// Wire ライブラリ関係

#ifndef __arm__

// Arduino Uno R3 等の時

#define CKLCD_Wire  Wire

#if F_CPU >= 16000000
  // Arduino の CPU clock が 16MHz 以上の時、TWI 通信速度を 400kHz に高速化します。
  inline void CKLCD_Wire_setHiSpeed (void) {
    TWBR = ((F_CPU / 400000 ) - 16) / 2;
  }
#elif F_CPU >= 8000000
  // Arduino の CPU clock が 8MHz 以上の時、TWI 通信速度を 200kHz に高速化します。
  // ※ Leonardo 等で使用されている ATmega16/32u4 では、TWBR は 10 以上にするべし、とデータシートに書かれている為です。
  //    逆算すると、CPU clock が 14.4MHz以上の時でないと TWBR が 10 以上にならないので、8MHz では TWI 通信速度を 400kHz に設定できない事になります。
  //    なお、168P, 328P, 644P, 1284P のデータシートには TWBR の制限は記載されていないので、もしかすると 16/32u4 のみの制限事項かもしれません。
  inline void CKLCD_Wire_setHiSpeed (void) {
    TWBR = ((F_CPU / 200000 ) - 16) / 2;
  }
#else // F_CPU
  // Arduino の CPU clock が 8MHz 未満の時は、どうすればいいのか判断できないので、TWI 通信速度は標準の 100kHz のままです。
  inline void CKLCD_Wire_setHiSpeed (void) {
    ; // 何もしない
  }
#endif // F_CPU

#else // __arm__

// Arduino Due の時

#define CKLCD_Wire  Wire1

  inline void CKLCD_Wire_setHiSpeed (void) {
    ; // 何もしない ※ TwoWire::begin(void) 内の TWI_ConfigureMaster() を操作すれば、高速化できそうなんだが…
  }

#endif // __arm__


//////////////////////////////////////////////
// カラー漢字 LCD シールドの最大桁数・行数

// ランドスケープモードは液晶画面を横長にした状態を指します。TWIアドレス選択 DIPSW が左側に来る向きにした状態でもあります。
#define CKLCD_MAX_L_WIDTH   160  // ランドスケープモード時：最大グラフィック表示幅 (仮想画面サイズとは関係無く、物理的な画面サイズ)
#define CKLCD_MAX_L_HEIGHT  128  // ランドスケープモード時：最大グラフィック表示高さ (同上)

// ポートレートモードは液晶画面を縦長にした状態を指します。TWIアドレス選択 DIPSW が上側に来る向きにした状態でもあります。
#define CKLCD_MAX_P_WIDTH   128  // ポートレートモード時：最大グラフィック表示幅 (仮想画面サイズとは関係無く、物理的な画面サイズ)
#define CKLCD_MAX_P_HEIGHT  160  // ポートレートモード時：最大グラフィック表示高さ (同上)


//////////////////////////////////////////////
// カラー漢字 LCD シールドの仮想画面

// カラー漢字 LCD シールドは、仮想画面１と仮想画面２の２面を持っており、flip 命令の時にアルファブレンド合成する事が可能です。
// 仮想画面１のサイズは160×256画素サイズ固定であり、この内の160×128画素のみを LCD に表示します（または表示させないモードも有り）。
// 一方、仮想画面２のサイズは160×128画素サイズ固定であり、常に全画素を LCD に表示します（または表示させないモードも有り）。
// どちらの仮想画面も、LCD 表示開始座標を指定できます。リセット後、clear 命令後、または home 命令後、LCD 表示開始座標は (0,0) になります。

#define CKLCD_MAX_VPLANES     2  // 最大仮想画面数

#define CKLCD_VPLANE1         0  // 仮想画面1：160×256画素サイズ固定
#define CKLCD_VPLANE2         1  // 仮想画面2：160×128画素サイズ固定

#define CKLCD_VPLANE1_MAX_L_WIDTH   160  // 仮想画面1：ランドスケープモード時：最大幅
#define CKLCD_VPLANE1_MAX_L_HEIGHT  256  // 仮想画面1：ランドスケープモード時：最大高さ

#define CKLCD_VPLANE1_MAX_P_WIDTH   256  // 仮想画面1：ポートレート　モード時：最大幅
#define CKLCD_VPLANE1_MAX_P_HEIGHT  160  // 仮想画面1：ポートレート　モード時：最大高さ

#define CKLCD_VPLANE2_MAX_L_WIDTH   160  // 仮想画面2：ランドスケープモード時：最大幅
#define CKLCD_VPLANE2_MAX_L_HEIGHT  128  // 仮想画面2：ランドスケープモード時：最大高さ

#define CKLCD_VPLANE2_MAX_P_WIDTH   128  // 仮想画面2：ポートレート　モード時：最大幅
#define CKLCD_VPLANE2_MAX_P_HEIGHT  160  // 仮想画面2：ポートレート　モード時：最大高さ


// ランドスケープモード時(横160画素×縦128画素)の仮想画面１と仮想画面２の模式図は、以下の様になります。
// "☆"が絶対座標の(0,0)位置で、横方向がＸ軸、縦方向がＹ軸になります。("Ｐ"はポートレートモード時の絶対座標の(0,0)相当の座標です)
//
//        0  ←Ｘ→ 159        0  ←Ｘ→ 159
//      ┌──────┐     ┌──────┐
//     0│☆          │    0│☆          │
//      │            │   ↑│            │
//    ↑│            │   Ｙ│    VP#2    │
//    ││            │   ↓│            │
//      │            │  127│Ｐ          │
//    Ｙ│    VP#1    │     └──────┘
//      │            │
//    ││            │
//    ↓│            │
//      │            │
//   255│Ｐ          │
//      └──────┘
//
// ポートレートモード時(横128画素×縦160画素)の仮想画面１と仮想画面２の模式図は、以下の様になります。
// "☆"が絶対座標の(0,0)位置で、横方向がＸ軸、縦方向がＹ軸になります。("Ｌ"はランドスケープモード時の絶対座標の(0,0)相当の座標です)
//
//        0   ←─  Ｘ  ─→  255        0 ←Ｘ→127
//      ┌───────────┐     ┌─────┐
//     0│☆                  Ｌ│    0│☆      Ｌ│
//      │                      │     │          │
//    ↑│                      │   ↑│          │
//    Ｙ│         VP#1         │   Ｙ│   VP#2   │
//    ↓│                      │   ↓│          │
//   159│                      │  159│          │
//      └───────────┘     └─────┘
//
// ランドスケープモードとポートレートモードの違いは、Ｘ軸とＹ軸が入れ替わる事と、絶対座標(0,0) の位置がずれる事だけで、仮想画面のメモリ自体は同一です。
// この為、モードを切り替えても画像クリアしなければ、前のモードの時の画像データがそのまま残り続けることになります。


// 両方の仮想画面をアルファブレンドさせながら LCD に表示させる時は、どちらの仮想画面が Front(前景)面になるかを指示する必要があります。
// Front(前景)面の 16bit絶対色は常に ARGB4444 になります。アルファ成分が有るので、文字や透過画像を配置するのに向いています。
// 一方、Back(背景)面の 16bit絶対色は常に RGB565 になります。R/G/B のビット数が多いので精細な画像を配置するのに向いています。
// （ただし、アルファ成分は無いので、透過画像は強制的に非透過画像に変換されてしまいます）
// 仮想画面１と仮想画面２、別々に LCD 表示開始座標が指定できますので、重ね合わせ位置は自由に変更できます。
// （実際は、仮想画面２のサイズは LCD 画面分ちょうどしかないので、仮想画面１の方をずらす方が便利でしょう）
//
// 以下は、ランドスケープモード時の重ね合わせの例です（ポートレートモード時でも縦横が逆転し、原点位置がずれるだけで、重ね合わせ方は同じです）。
//
//    ┌──────┐       ┌──────┐
//    │Front面     ├┐     │Front面     │
//    │            ││     │            │
//    │    VP#2    ││     │            │
//    │ (ARGB4444) ││     │    VP#1    │
//    │            ││     │ (ARGB4444) │
//    └┬─────┘│     │            │
//      │            │     │            ├┐
//      │Back面      │     │            ││
//      │            │     │            ││← Back面
//      │    VP#1    │     │            ││    VP#2
//      │  (RGB565)  │     │            ││  (RGB565)
//      │            │     └┬─────┘│
//      └──────┘       └──────┘
//
// Front(前景)面無し、にする事も可能です。一方、Back(背景)面は必ず必要です。


//////////////////////////////////////////////
// スクリーンモード

// スクリーンのオリエンテーション
// 液晶画面の表示向きは２種類あり、横長状態をランドスケープ、縦長状態をポートレート、と呼称します。
// ２面ある仮想画面も同時に、縦と横の関係が入れ替わり、絶対座標(0,0)の位置も移動しますので、注意が必要です。
#define CKLCD_ORIENT_LANDSCAPE   0x00  // ランドスケープモード：液晶画面を横長にした状態
#define CKLCD_ORIENT_PORTRAIT    0x40  // ポートレート　モード：液晶画面を縦長にした状態

// 表示させる仮想画面
// Front面無しのモードは描画が高速です。
// 一方、Front面有りのモードは常にアルファブレンド処理が行われるので、その分、描画が遅くなります。
// なお、非表示中の仮想画面は、作業エリアとして使用する事が可能ですので、コピー元の画像とかを置いておくのに最適です。
#define CKLCD_DISP_SCREEN_NONE   B000000  // Front面：無し、Back面：無し  ※常に黒が表示されます。このモードの時は、仮想画面１と２は LCD に表示されないので作業エリアとして使用可能
#define CKLCD_DISP_SCREEN_B1     B000100  // Front面：無し、Back面：仮想画面１(160×256)  ※このモードの時は、仮想画面２は LCD に表示されないので作業エリアとして使用可能
#define CKLCD_DISP_SCREEN_B2     B000101  // Front面：無し、Back面：仮想画面２(160×128)  ※このモードの時は、仮想画面１は LCD に表示されないので作業エリアとして使用可能
#define CKLCD_DISP_SCREEN_F1     B100000  // Front面：仮想画面１(160×256)、Back面：無し  ※このモードの時は、仮想画面２は LCD に表示されないので作業エリアとして使用可能
#define CKLCD_DISP_SCREEN_F2     B101000  // Front面：仮想画面２(160×128)、Back面：無し  ※このモードの時は、仮想画面１は LCD に表示されないので作業エリアとして使用可能
#define CKLCD_DISP_SCREEN_F1_B2  B100101  // Front面：仮想画面１(160×256)、Back面：仮想画面２(160×128)  ※ LCD に表示時にアルファブレンド処理されます
#define CKLCD_DISP_SCREEN_F2_B1  B101100  // Front面：仮想画面２(160×128)、Back面：仮想画面１(160×256)  ※ LCD に表示時にアルファブレンド処理されます


//////////////////////////////////////////////
// フリップモード
// print/write 系関数を呼び出した時、仮想画面上に描画された文字を、自動的に液晶画面に flip させるか・させないか、の設定です。
// 描画途中の仮想画面を表示させず、描画完了してから一気に表示させたい場合は MANU モードにしてください。
// なお、AUTO モードでも、自動的に flip するのは print/write 系関数と clear 関数のみで、それ以外の描画関数は自動的に flip しませんので注意して下さい。
#define CKLCD_FLIP_MODE_AUTO  0x00  // 自動的に flip します。デフォルト設定です。
#define CKLCD_FLIP_MODE_MANU  0x20  // 自動的に flip しません。ユーザーが明示的に flip() を呼ぶ必要があります。


//////////////////////////////////////////////
// カラー漢字 LCD シールドで指定できる色
// カラー漢字 LCD シールドでは、16bit長の絶対色が２種類、32bit長の論理色が１種類、の計３つあります。
// 通常は、32bit長の論理色を使用します。カラー漢字LCD 内で自動的に、どちらかの 16bit長の絶対色に変換されます。
// ユーザーが直接、16bit長の絶対色を操作する事は少ないですが、その挙動は知っておく必要があります。

//                              16bit長                       32bit長
//      絶対色(ARGB4444)    AAAARRRRGGGGBBBB    AAAA---- RRRR---- GGGG---- BBBB----   ※ ARGB4444 は 16bit長のフォーマットの通りですが、32bit長に置き換えると、このような"重み付け"になります
//      絶対色 (RGB565)     RRRRRGGGGGGBBBBB    -------- RRRRR--- GGGGGG-- BBBBB---   ※  RGB565  は 16bit長のフォーマットの通りですが、32bit長に置き換えると、このような"重み付け"になります
//      論理色(ARGB8888)    -----無  し-----    AAAAAAAA RRRRRRRR GGGGGGGG BBBBBBBB   ※ ARGB4444 と RGB565 を包括した 32bit長のみのフォーマットです。16bit長は有りません。

// 絶対色：16bit長
// Front面に選ばれた仮想画面は、ARGB4444 で描画する必要があります。アルファ成分が有る為、代償として R/G/B の各ビット幅が少なくなっています。
#define CKLCD_FRONT_VP_ALPHA_MASK  0xF000  // Front面のα成分：4bit長
#define CKLCD_FRONT_VP_RED_MASK    0x0F00  // Front面の赤成分：4bit長
#define CKLCD_FRONT_VP_GREEN_MASK  0x00F0  // Front面の緑成分：4bit長
#define CKLCD_FRONT_VP_BLUE_MASK   0x000F  // Front面の青成分：4bit長
// Back面に選ばれた仮想画面は、RGB565 で描画する必要があります。アルファ成分は有りません。その代わり、R/G/B の各ビット幅が多くなっています。
#define CKLCD_BACK_VP_RED_MASK     0xF800  // Back面の赤成分：5bit長
#define CKLCD_BACK_VP_GREEN_MASK   0x07E0  // Back面の緑成分：6bit長
#define CKLCD_BACK_VP_BLUE_MASK    0x001F  // Back面の青成分：5bit長

// 論理色：32bit長
// 通常は、この論理色(ARGB8888)を使用します。この論理色は ARGB4444 と RGB565 の２つを包括した論理的な色であり、
// 実際に描画する時には、指定した仮想画面が Front面の時は ARGB4444 に、Back面の時は RGB565 に自動的に変換されます。
#define CKLCD_COL32_ALPHA_MASK  0xFF000000  // 論理色のα成分：8bit長  ※ 実際は先頭の 4bit のみ有効
#define CKLCD_COL32_RED_MASK    0x00FF0000  // 論理色の赤成分：8bit長  ※ 実際は先頭の 5bit 乃至 4bit のみ有効
#define CKLCD_COL32_GREEN_MASK  0x0000FF00  // 論理色の緑成分：8bit長  ※ 実際は先頭の 6bit 乃至 4bit のみ有効
#define CKLCD_COL32_BLUE_MASK   0x000000FF  // 論理色の青成分：8bit長  ※ 実際は先頭の 5bit 乃至 4bit のみ有効

typedef uint16_t cklcdCol16_t; // 絶対色：16bit長
typedef uint32_t cklcdCol32_t; // 論理色：32bit長

#define CKLCD_COL32_MAKE_ALPHA(a)  ((cklcdCol32_t)((a) & 0xFF) << 24)
#define CKLCD_COL32_MAKE_RED(r)    ((cklcdCol32_t)((r) & 0xFF) << 16)
#define CKLCD_COL32_MAKE_GREEN(g)  ((cklcdCol32_t)((g) & 0xFF) <<  8)
#define CKLCD_COL32_MAKE_BLUE(b)   ((cklcdCol32_t)((b) & 0xFF) <<  0)

#define CKLCD_COL32_SPLIT_ALPHA(col32)  ((uint8_t)((col32) >> 24))
#define CKLCD_COL32_SPLIT_RED(col32)    ((uint8_t)((col32) >> 16))
#define CKLCD_COL32_SPLIT_GREEN(col32)  ((uint8_t)((col32) >>  8))
#define CKLCD_COL32_SPLIT_BLUE(col32)   ((uint8_t)((col32) >>  0))

// 透明の論理色：アルファ値==0 の透明色です。仮想画面が透過モードの時は、このままだと全く LCD に表示されないので、アルファ値(1～255)を加算する必要があります。
#define CKLCD_COL32_TRANS               0x00000000  // 完全透明の論理色
#define CKLCD_COL32_TRANS_RED           0x00FF0000
#define CKLCD_COL32_TRANS_GREEN         0x0000FF00
#define CKLCD_COL32_TRANS_YELLOW        0x00FFFF00
#define CKLCD_COL32_TRANS_BLUE          0x000000FF
#define CKLCD_COL32_TRANS_MAGENTA       0x00FF00FF
#define CKLCD_COL32_TRANS_CYAN          0x0000FFFF
#define CKLCD_COL32_TRANS_WHITE         0x00FFFFFF
#define CKLCD_COL32_TRANS_DARK_RED      0x00800000
#define CKLCD_COL32_TRANS_DARK_GREEN    0x00008000
#define CKLCD_COL32_TRANS_DARK_YELLOW   0x00808000
#define CKLCD_COL32_TRANS_DARK_BLUE     0x00000080
#define CKLCD_COL32_TRANS_DARK_MAGENTA  0x00800080
#define CKLCD_COL32_TRANS_DARK_CYAN     0x00008080
#define CKLCD_COL32_TRANS_GRAY          0x00808080

// 不透明の論理色
#define CKLCD_COL32_BLACK         (CKLCD_COL32_ALPHA_MASK | 0)
#define CKLCD_COL32_RED           (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_RED)
#define CKLCD_COL32_GREEN         (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_GREEN)
#define CKLCD_COL32_YELLOW        (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_YELLOW)
#define CKLCD_COL32_BLUE          (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_BLUE)
#define CKLCD_COL32_MAGENTA       (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_MAGENTA)
#define CKLCD_COL32_CYAN          (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_CYAN)
#define CKLCD_COL32_WHITE         (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_WHITE)
#define CKLCD_COL32_DARK_RED      (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_DARK_RED)
#define CKLCD_COL32_DARK_GREEN    (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_DARK_GREEN)
#define CKLCD_COL32_DARK_YELLOW   (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_DARK_YELLOW)
#define CKLCD_COL32_DARK_BLUE     (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_DARK_BLUE)
#define CKLCD_COL32_DARK_MAGENTA  (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_DARK_MAGENTA)
#define CKLCD_COL32_DARK_CYAN     (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_DARK_CYAN)
#define CKLCD_COL32_GRAY          (CKLCD_COL32_ALPHA_MASK | CKLCD_COL32_TRANS_GRAY)


//////////////////////////////////////////////
// 仮想画面の状態設定
#define CKLCD_VP_COND_OPAQUE  0x00  // 非透過モード: Back (背景)面は基本的にこの設定になりますが、flip時に注意する事を条件に setCondition() で透過モードに変更する事も可能です。
#define CKLCD_VP_COND_TRANS   0x04  // 透過モード  : Front(前景)面は基本的にこの設定になりますが、flip時に注意する事を条件に setCondition() で非透過モードに変更する事も可能です。

//////////////////////////////////////////////
// setBlendMode()で指定できる処理
#define CKLCD_BLEND_OVER      0x00  // ブレンドせず、上書きします。処理速度は速いです。
#define CKLCD_BLEND_ADD       0x08  // アルファ値を元に、ブレンド処理します。論理アルファ値==1.0 の時は上書きした時と同じ結果になります。各画素毎に計算するので、処理速度は遅いです。

//////////////////////////////////////////////////////////////
// 描画関係コマンド(pset/line/rect/circle)時のオプション
#define CKLCD_DRAW_COLOR      0x00  // setDrawColor()の論理色で描画する
#define CKLCD_BACK_COLOR      0x10  // setBackColor()の論理色で描画する
#define CKLCD_DRAW_NO_FILL    0x00  // 領域内を塗りつぶさない
#define CKLCD_DRAW_FILL       0x20  // 領域内を塗りつぶす: rect, circle関数でのみ有効


/////////////////////////////////////////////////////
// TWI通信時の画像データ転送時の MAX バッファサイズ
// (Arduino の Wire ライブラリのバッファは、コマンドbyte 含めて 32byte までしかなので、31byte のみ指定可能。それ以外を指定すると必ず CKLCD_ERR_TWI_BUFF_FULL になってしまいます)
#define CKLCD_LOAD_IMG_TWI_BUFF_SIZE_31   0x00  //  31byte
#define CKLCD_LOAD_IMG_TWI_BUFF_SIZE_63   0x10  //  63byte
#define CKLCD_LOAD_IMG_TWI_BUFF_SIZE_95   0x20  //  95byte
#define CKLCD_LOAD_IMG_TWI_BUFF_SIZE_127  0x30  // 127byte

// 画像読み込み時にエラーが発生したら、簡易メッセージを表示させます。
// ※ エラー無く画像読み込めた時は、その画像が表示されるので、特にメッセージは表示されません
// 注意：このフラグを使用する時は、フリップモードは CKLCD_FLIP_MODE_AUTO にしておいた方が良いでしょう。
//       CKLCD_FLIP_MODE_MANU だと仮想画面上に描かれるだけで、flip() が呼ばれるまで LCD に表示されません。
//       画像展開開始座標に負の値を指定している場合、メッセージが見えない可能性があります。
//       また、画像展開先が仮想画面１の時は、setDispStartPos() の設定次第では、LCD の表示範囲外になって、結果的にメッセージが見れない場合があります。
#define CKLCD_LOAD_IMG_DISP_ERR_MSG  0x80  // エラー発生時に (xPos,yPos) の位置に簡易メッセージを表示


//////////////////////////////////////////////
// copyWithRotate 関係

// カラー漢字 LCD ライブラリの逆倍率: 例えば 1.2倍にしたい時は、(2^10)/1.2 を指定します。実際は CKLCD_FLOAT_TO_INVRATIO10 マクロを使用して下さい。
// 固定小数点値(Q6.10)：整数部6bit+小数部10bit: 表現できる範囲は 0.0 ～ +63.99

typedef uint16_t cklcdInvRatio10_t;  // 逆倍率: 16bit長固定小数点値(Q6.10)

#define CKLCD_FLOAT_TO_INVRATIO10(f)  ((cklcdInvRatio10_t)(1024.0 / (f) + 0.5))  // float から 逆倍率(Q6.10) へ変換します: 指定できる f の範囲は 0.1～+63.99 ※0.0だと0除算エラーになります
#define CKLCD_INVRATIO10_TO_FLOAT(q)  (1024.0 / (float)(q))  // 逆倍率(Q6.10) から float へ変換します


// カラー漢字 LCD ライブラリの角度：単位"BAM10" (Binary Angular Measurement 10) : 1024 BAM10 で 360度になります。よって、1 BAM10 ≒ 0.35度です。
// 一種の固定小数点値(Q10)です。有効10bit長(上位6bitは無視)で、0x0000(0度)～0x0200(180度)～0x03FF(約359.6度) の間を循環します
#define CKLCD_BAM10_TWO_PI  0x0400              // 2π ラジアン == 360度 ※上位6bitを無視すれば 0x0000(0度) と等しくなります。
#define CKLCD_BAM10_PI      (CKLCD_TWO_PI / 2)  //  π ラジアン == 180度
#define CKLCD_BAM10_HALF_PI (CKLCD_TWO_PI / 4)  // π/2ラジアン ==  90度

typedef uint16_t cklcdBAM10_t;  // "BAM10"単位: 16bit長だが実際は下位10bitのみ有効。上位6bitは無視される

#define CKLCD_RAD_TO_BAM10(rad)  ((cklcdBAM10_t)((rad) / TWO_PI * CKLCD_BAM10_TWO_PI)) // ラジアンから、"BAM10"単位へ変換します
#define CKLCD_DEG_TO_BAM10(deg)  ((cklcdBAM10_t)((deg) / 360.0  * CKLCD_BAM10_TWO_PI)) // 度から、"BAM10"単位へ変換します


//////////////////////////////////////////////
// 仮想画面のクリア・フラグ
#define CKLCD_CLR_FLG_FLIP             0x01  // フリップモードが AUTO モードの時、現在の LCD 表示開始座標に従って、全表示領域を flip させます
#define CKLCD_CLR_FLG_DISP_START_HOME  0x02  // LCD 表示開始位置を (0,0) にする
#define CKLCD_CLR_FLG_CURSOR_HOME      0x04  // カーソル位置を (0,0) にする

//////////////////////////////////////////////
// フォントのオプション

// 全て、M+フォント＋IPAフォントです。
// モノスペース    (固定幅)：全角文字は同じ幅になるので setCursor関数で指定しやすくなります。
//                           その代わり、文字の間隔があいてしまうので、欧文などの表示が美しくありません。
// プロポーショナル(可変幅)：漢字以外の文字が可変幅になり、読みやすくなります。
//                           その代わり、setCursor 関数で指定しづらくなってしまいます。setCursorMin 関数を使用するようにして下さい。

// ルビ用フォント：幅は最大８画素で、高さは10画素です。ただし、ASCII、ISO8859-1、ひらがなとカタカナ、一部の記号のみです。
#define CKLCD_FONT_TYPE_RUBY_MS       0x00  // モノスペース　　(固定幅)：８画素固定
#define CKLCD_FONT_TYPE_RUBY_P        0x01  // プロポーショナル(可変幅)：最大８画素

// 12pxフォント：幅は最大12画素で、高さは15画素です。ただし、ASCII、ISO8859-1、ひらがなとカタカナ、一部の記号のみです。
#define CKLCD_FONT_TYPE_12px_MS       0x02  // モノスペース　　(固定幅)：12画素固定
#define CKLCD_FONT_TYPE_12px_P        0x03  // プロポーショナル(可変幅)：最大12画素

// 標準フォント：幅は最大16画素で、高さは20画素です。JIS 第一水準～第三水準の全ての文字をサポートしています。
#define CKLCD_FONT_TYPE_MONOSPACED    0x04  // モノスペース　　(固定幅)：16画素固定
#define CKLCD_FONT_TYPE_PROPORTIONAL  0x05  // プロポーショナル(可変幅)：最大16画素

// 20pxフォント：幅は最大20画素で、高さは25画素です。ただし、ASCII、ISO8859-1、極一部の記号のみです。
#define CKLCD_FONT_TYPE_20px_MS       0x06  // モノスペース　　(固定幅)：20画素固定
#define CKLCD_FONT_TYPE_20px_P        0x07  // プロポーショナル(可変幅)：最大20画素

// 24pxフォント：幅は最大24画素で、高さは30画素です。ただし、ASCII、ISO8859-1、極一部の記号のみです。
#define CKLCD_FONT_TYPE_24px_MS       0x08  // モノスペース　　(固定幅)：24画素固定
#define CKLCD_FONT_TYPE_24px_P        0x09  // プロポーショナル(可変幅)：最大24画素

// 28pxフォント：幅は最大28画素で、高さは35画素です。ただし、ASCII、ISO8859-1、極一部の記号のみです。
#define CKLCD_FONT_TYPE_28px_MS       0x0A  // モノスペース　　(固定幅)：28画素固定
#define CKLCD_FONT_TYPE_28px_P        0x0B  // プロポーショナル(可変幅)：最大28画素


// フォントの属性
#define CKLCD_FONT_ATTR_UNDERLINE   0x10  // アンダーラインを表示
#define CKLCD_FONT_ATTR_STRIKE      0x20  // 取消し線を表示
#define CKLCD_FONT_ATTR_INVERTED    0x40  // 文字色を反転


// 現在のフォント情報：この値を直接指定するのではなく、ColorKanjiLCDVPlane クラスの getFontType(), getFontWidth(), getFontHeight() を使用して下さい。
#define CKLCD_CURR_FONT_INFO_TYPE    0
#define CKLCD_CURR_FONT_INFO_WIDTH   1
#define CKLCD_CURR_FONT_INFO_HEIGHT  2


//////////////////////////////////////////////
// サポートしている符号化方式：setEncoding()で指定

#define CKLCD_ENC_UTF8      0  // UTF-8 方式 : default
#define CKLCD_ENC_UTF16     1  // UTF-16 方式 (BOM 自動判別)
#define CKLCD_ENC_UTF16BE   2  // UTF-16 方式 (Big Endian 固定)
#define CKLCD_ENC_UTF16LE   3  // UTF-16 方式 (Little Endian 固定)
#define CKLCD_ENC_SJIS      4  // シフトJIS 方式
#define CKLCD_ENC_EUC_JP    5  // EUC-JP 方式
#define CKLCD_MAX_ENCODING  6  // 最大符号化方式

#define CKLCD_UTF16_BOM  (L'\uFEFF') // Byte Order Mark



////////////////////////////////////////////////////
// print/write 系関数で使用できる Escape Sequence
// ※末尾が'm'のものは、メモリを節約したい時は、"\x1b[0;4;31;47m" という風にも書けます。この場合、描画=赤, 背景=白, アンダーライン有り、になります。

#define CKLCD_CSI_DEFAULT       "\x1b[m"  // デフォルト：描画=白, 背景=黒/透明, アンダーライン/色反転/取消し線無し

#define CKLCD_CSI_BLACK         "\x1b[30m"  // 0xFF000000: 描画色：setDrawColor()と同じ
#define CKLCD_CSI_RED           "\x1b[31m"  // 0xFFFF0000
#define CKLCD_CSI_GREEN         "\x1b[32m"  // 0xFF00FF00
#define CKLCD_CSI_YELLOW        "\x1b[33m"  // 0xFFFFFF00
#define CKLCD_CSI_BLUE          "\x1b[34m"  // 0xFF0000FF
#define CKLCD_CSI_MAGENTA       "\x1b[35m"  // 0xFFFF00FF
#define CKLCD_CSI_CYAN          "\x1b[36m"  // 0xFF00FFFF
#define CKLCD_CSI_WHITE         "\x1b[37m"  // 0xFFFFFFFF
#define CKLCD_CSI_TRANS         "\x1b[38m"  // 0x00000000: ARGB4444では透明、RGB565では黒

#define CKLCD_CSI_BACK_BLACK    "\x1b[40m"  // 0xFF000000: 背景色：setBackColor()と同じ
#define CKLCD_CSI_BACK_RED      "\x1b[41m"  // 0xFFFF0000
#define CKLCD_CSI_BACK_GREEN    "\x1b[42m"  // 0xFF00FF00
#define CKLCD_CSI_BACK_YELLOW   "\x1b[43m"  // 0xFFFFFF00
#define CKLCD_CSI_BACK_BLUE     "\x1b[44m"  // 0xFF0000FF
#define CKLCD_CSI_BACK_MAGENTA  "\x1b[45m"  // 0xFFFF00FF
#define CKLCD_CSI_BACK_CYAN     "\x1b[46m"  // 0xFF00FFFF
#define CKLCD_CSI_BACK_WHITE    "\x1b[47m"  // 0xFFFFFFFF
#define CKLCD_CSI_BACK_TRANS    "\x1b[48m"  // 0x00000000: ARGB4444では透明、RGB565では黒

#define CKLCD_CSI_UNDERLINE     "\x1b[4m"  // アンダーライン有り  CKLCD_FONT_ATTR_UNDERLINE
#define CKLCD_CSI_INVERTED      "\x1b[7m"  // 色反転させる        CKLCD_FONT_ATTR_INVERTED
#define CKLCD_CSI_STRIKE        "\x1b[9m"  // 取消し線有り        CKLCD_FONT_ATTR_STRIKE

#define CKLCD_CSI_NO_UNDERLINE  "\x1b[24m"  // アンダーライン無し  ~CKLCD_FONT_ATTR_UNDERLINE
#define CKLCD_CSI_NO_INVERTED   "\x1b[27m"  // 色反転させない      ~CKLCD_FONT_ATTR_INVERTED
#define CKLCD_CSI_NO_STRIKE     "\x1b[29m"  // 取消し線無し        ~CKLCD_FONT_ATTR_STRIKE


#define CKLCD_CSI_CLR_SCREEN      "\x1b[2J"  // 仮想画面全体を現在の背景色で消去

#define CKLCD_CSI_ERASE_LINE      "\x1b[K"   // カーソル行全体を、現在の背景色で消去
#define CKLCD_CSI_ERASE_LINE_H2C  "\x1b[1K"  // 行頭からカーソルまで、現在の背景色で消去
#define CKLCD_CSI_ERASE_LINE_C2T  "\x1b[2K"  // カーソルから行末まで、現在の背景色で消去


#define CKLCD_CSI_FNT_RUBY_MS  "\x1b[0 D"   // モノスペース　　(固定幅)：８画素固定
#define CKLCD_CSI_FNT_RUBY_P   "\x1b[1 D"   // プロポーショナル(可変幅)：最大８画素
#define CKLCD_CSI_FNT_12px_MS  "\x1b[2 D"   // モノスペース　　(固定幅)：12画素固定
#define CKLCD_CSI_FNT_12px_P   "\x1b[3 D"   // プロポーショナル(可変幅)：最大12画素
#define CKLCD_CSI_FNT_STD_MS   "\x1b[4 D"   // モノスペース　　(固定幅)：16画素固定
#define CKLCD_CSI_FNT_STD_P    "\x1b[5 D"   // プロポーショナル(可変幅)：最大16画素
#define CKLCD_CSI_FNT_20px_MS  "\x1b[6 D"   // モノスペース　　(固定幅)：20画素固定
#define CKLCD_CSI_FNT_20px_P   "\x1b[7 D"   // プロポーショナル(可変幅)：最大20画素
#define CKLCD_CSI_FNT_24px_MS  "\x1b[8 D"   // モノスペース　　(固定幅)：24画素固定
#define CKLCD_CSI_FNT_24px_P   "\x1b[9 D"   // プロポーショナル(可変幅)：最大24画素
#define CKLCD_CSI_FNT_28px_MS  "\x1b[10 D"  // モノスペース　　(固定幅)：28画素固定
#define CKLCD_CSI_FNT_28px_P   "\x1b[11 D"  // プロポーショナル(可変幅)：最大28画素


//////////////////////////////////////////////
// カラー漢字 LCD ライブラリのエラーコード

#define CKLCD_ERR_NONE                   0  // 正常

#define CKLCD_ERR_TWI_BUFF_FULL          1  // Wire ライブラリの送信バッファが一杯   ※Arduino の Wire ライブラリのエラーコードと同一
#define CKLCD_ERR_TWI_ADDR_NACK          2  // TWI アドレス送信で NACK が返ってきた  ※同上
#define CKLCD_ERR_TWI_DATA_NACK          3  // TWI データ送信で NACK が返ってきた    ※同上
#define CKLCD_ERR_TWI_OTHER_ERR          4  // Wire ライブラリのその他のエラー       ※同上

#define CKLCD_ERR_LOAD_IMG_END           0x80  // LoadImg の callback 終了
#define CKLCD_ERR_IMG_LOADING            0x81  // 画像読み込み中：実際はエラーではなく、単なる通知情報
#define CKLCD_ERR_IMG_LOADING_FRAME_END  0x82  // 画像読み込み中(フレームの終端)：実際はエラーではなく、単なる通知情報

#define CKLCD_ERR_IMG_NO_JPG             0xA0  // JPG: JFIFフォーマットでないJPGデータ
#define CKLCD_ERR_JPG_TOO_BIG            0xA1  // JPG: 1024×1024 画素以上の寸法。メモリ容量の関係で、この寸法以下しか読み込めません。
#define CKLCD_ERR_JPG_INSUF_MEM          0xA2  // JPG: 処理するためのメモリが足らない(通常発生しないはず)
#define CKLCD_ERR_JPG_DATA_ERR           0xA3  // JPG: 異常なJFIFフォーマット(壊れている？)
#define CKLCD_ERR_JPG_INSUF_DATA         0xA4  // JPG: 画像の大きさに必要な画像データに達する前にデータ終了を検知した：多分、中途半端な画像で途切れているハズ
#define CKLCD_ERR_JPG_UNSUP_FMT          0xA5  // JPG: JFIFフォーマットとしては正しいが、本機ではサポートしていない
#define CKLCD_ERR_JPG_UNSUP_PRGRS        0xA6  // JPG: Progressive JPG は、未サポート

#define CKLCD_ERR_IMG_NO_PNG             0xB0  // PNG: PNGフォーマットでないPNGデータ
#define CKLCD_ERR_PNG_TOO_BIG            0xB1  // PNG: 256×1024 画素以上の寸法。メモリ容量の関係で、この寸法以下しか読み込めません。
#define CKLCD_ERR_PNG_INSUF_MEM          0xB2  // PNG: 処理するためのメモリが足らない(通常発生しないはず)
#define CKLCD_ERR_PNG_DATA_ERR           0xB3  // PNG: 異常なPNGフォーマット(壊れている？)
#define CKLCD_ERR_PNG_INSUF_DATA         0xB4  // PNG: 画像の大きさに必要な画像データに達する前にデータ終了を検知した：多分、中途半端な画像で途切れているハズ
#define CKLCD_ERR_PNG_CHUNK_ERR          0xB5  // PNG: 異常なPNGフォーマット(壊れている？)：チャンクエラー：異常なチャンクや、チャンクの順番がおかしい
#define CKLCD_ERR_PNG_CRC_ERR            0xB6  // PNG: 異常なPNGフォーマット(壊れている？)：CRCエラー
#define CKLCD_ERR_PNG_IHDR_ERR           0xB7  // PNG: 異常なPNGフォーマット(壊れている？)：IHDR異常
#define CKLCD_ERR_PNG_UNSUP_DEPTH        0xB8  // PNG: 対応していない色深度。対応しているのは 1/2/4/8 bit深度のみです。16 bit深度の png には非対応です。
#define CKLCD_ERR_PNG_UNSUP_INTERLACE    0xB9  // PNG: interlace PNG には対応していない
#define CKLCD_ERR_PNG_INV_CHAN           0xBA  // PNG: チャンネル数が3でも4でもない：グレースケール画像からカラー画像への変換(といってもRGB同じ値にしてるだけ)に失敗した時に発生する
#define CKLCD_ERR_PNG_ZLIB_ERR           0xBB  // PNG: 異常なPNGフォーマット(壊れている？)：zlibでエラー
#define CKLCD_ERR_PNG_IMG_ERR            0xBC  // PNG: 異常なPNGフォーマット(壊れている？)：画像部(IDAT)が異常
#define CKLCD_ERR_PNG_UNKNOWN            0xBF  // PNG: 上記の分類に当てはまらないエラー

#define CKLCD_ERR_IMG_NO_GIF             0xC0  // GIF: GIF87/89フォーマットでないGIFデータ
#define CKLCD_ERR_GIF_TOO_BIG            0xC1  // GIF: 640×1024 画素以上の寸法。メモリ容量の関係で、この寸法以下しか読み込めません。
#define CKLCD_ERR_GIF_INSUF_MEM          0xC2  // GIF: 処理するためのメモリが足らない(通常発生しないはず)
#define CKLCD_ERR_GIF_DATA_ERR           0xC3  // GIF: 異常なGIF87/89フォーマット(壊れている？)
#define CKLCD_ERR_GIF_INSUF_DATA         0xC4  // GIF: 画像の大きさに必要な画像データに達する前にデータ終了を検知した：多分、中途半端な画像で途切れているハズ
#define CKLCD_ERR_GIF_NO_COL_TBL         0xC5  // GIF: カラーテーブルが無い：異常なGIF87/89フォーマット

#define CKLCD_ERR_IMG_NO_BMP             0xD0  // BMP: 'BM'から開始していないBMPデータ
#define CKLCD_ERR_BMP_TOO_BIG            0xD1  // BMP: 640×1024 画素以上の寸法。メモリ容量の関係で、この寸法以下しか読み込めません。
#define CKLCD_ERR_BMP_INSUF_MEM          0xD2  // BMP: 処理するためのメモリが足らない(通常発生しないはず)
#define CKLCD_ERR_BMP_DATA_ERR           0xD3  // BMP: 異常なBMPフォーマット(壊れている？)
#define CKLCD_ERR_BMP_INSUF_DATA         0xD4  // BMP: 画像の大きさに必要な画像データに達する前にデータ終了を検知した：多分、中途半端な画像で途切れているハズ
#define CKLCD_ERR_BMP_UNSUP_HEAD         0xD5  // BMP: 対応していないヘッダー。対応しているのは BITMAPINFOHEADER, BITMAPV4HEADER, BITMAPV5HEADER のみです。BITMAPCOREHEADER には非対応です。
#define CKLCD_ERR_BMP_UNSUP_PLANES       0xD6  // BMP: 対応していないプレーン数。対応しているのは 1 のみです。
#define CKLCD_ERR_BMP_UNSUP_COMP         0xD7  // BMP: 対応していない圧縮タイプ。対応しているのは BI_RGB, BI_BITFIELDS, BI_ALPHABITFIELDS のみです。RLE圧縮(BI_RLE4/BI_RLE8)には非対応です。
#define CKLCD_ERR_BMP_UNSUP_BITCNT       0xD8  // BMP: 対応していないBitCount。対応しているのは 1/2/4/8/16/24/32 bpp のみです。

#define CKLCD_ERR_BUSY                   0xF0  // カラー漢字 LCD シールドが Busy なので、要求を処理できない
#define CKLCD_ERR_TIME_OUT               0xF1  // カラー漢字 LCD シールドとの通信がタイムアウトした
#define CKLCD_ERR_ABORTED                0xF2  // ユーザーによって中断された
#define CKLCD_ERR_INVALID_STATE          0xF7  // 不正なステート：例えば、loadXxx() 中で無いのに、CKLCD_CMD_LOAD_IMG_XLAT/CANCEL コマンドを発行した場合
#define CKLCD_ERR_CONFLICT               0xF8  // 衝突していて解決できない：例えば、copyWithRotate() でコピー元とコピー先が部分的に重なっている場合
#define CKLCD_ERR_SAME                   0xF9  // 同じものを指定した：例えば、copy() でコピー元とコピー先が同一の時
#define CKLCD_ERR_OUT_OF_RANGE           0xFA  // 指定できる範囲から逸脱している
#define CKLCD_ERR_INVALID_PARAM          0xFB  // 引数が不正
#define CKLCD_ERR_UNK_CMD                0xFC  // BS21LAB-006 でサポートしていないコマンド
#define CKLCD_ERR_NOT_INIT               0xFD  // 初期化されていない
#define CKLCD_ERR_UNDEFINED              0xFE  // 未定義エラー
#define CKLCD_ERR_UNKNOWN                0xFF  // 原因不明のエラーが発生した


//////////////////////////////////////////////
// コールバック関数の定義

// LoadImg した画像のサイズ構造体
// アニメーション GIF の時のみ、画像のサイズではなく、各フレームのサイズになります。
// 例えば 100x100 の GIF 画像でも、あるフレームでは 10x10 しか無い場合は、サイズは 10,10 になります。
struct ColorKanjiLCDLoadImgSize {
  uint16_t  x_offset;  // アニメーション GIF の時のみ有効。それ以外では常に0
  uint16_t  y_offset;  // 同上
  uint16_t  width;     // アニメーション GIF の時は各フレームの大きさ。それ以外では画像の大きさ
  uint16_t  height;    // 同上
};

// loadBmp, loadJpg, loadPng, loadGif 関数から呼び出されるユーザー定義 callback 関数
// buff に buff_size 分の画像データを転送してください。正常に転送できた時は CKLCD_ERR_NONE を返してください。
// もし転送できなかったら CKLCD_ERR_UNKNOWN を返してください。その時は loadXxx 関数の処理を中断します。
// user_def_data は ColorKanjiLCDLoadImgSize の派生構造体のポインタで、ユーザーが自由に使えます。
typedef errCode_t (*cklcdLoadImgCallback_t) (uint8_t * buff, uint8_t buff_size, ColorKanjiLCDLoadImgSize * user_def_data);

// loadGif 関数から呼び出されるユーザー定義 callback 関数
// アニメーション GIF の時のみ、各フレーム毎に呼び出されます。flip 関数で LCD に表示させると、アニメーションしている様に演出できます。
// ただし、最終フレームの時は、この callback 関数は呼ばれずに、loadGif 関数自体が正常終了しますので、最終フレームだけは、他の loadXxx 関数と同じ手順で表示させてください。
// また、アニメーション GIF でない単なる GIF 画像の場合、この callback 関数が呼び出される事はありません。他の loadXxx 関数と同じ手順で表示させてください。
// なお、この callback 関数内では、以下の '★' 印の付いた関数だけしか呼び出せませんので、注意して下さい。これ以外の関数を呼び出すと、loadGif 関数が中断される事になります。
// 特に print/write 系関数は一切、呼び出せない（CKLCD 内の作業メモリが足らないので不可能）ので、アニメーションの最中に、文字を表示させる事はできません。
// また、透過色有りのアニメーションGIF を前景面に描画している場合、背景面が GIF 画像データ展開の作業メモリになっているので、重ね合わせはできません(ゴミが表示されます)。
// 一方、タクトSW の現在の状態は読み出せますので、タクトSW が押されたらアニメーション GIF をコマ送りする事は可能です。
typedef errCode_t (*cklcdImgFrameCallback_t) (ColorKanjiLCDLoadImgSize * user_def_data);


//////////////////////////////////////////////
// カラー漢字 LCD コアクラスの宣言
// ※ 後述の ColorKanjiLCD クラスで派生されるので、ユーザーがこのクラスを生成する必要はありません。
// ※ 主に TWI(i2c)通信を行います。後述の ColorKanjiLCDVPlane クラスがこの機能を必要とする為、ColorKanjiLCD クラスから分離させました。
class ColorKanjiLCDCore {
public:
  ColorKanjiLCDCore (uint8_t twiAddr); // コンストラクタ：カラー漢字 LCD シールドの TWI(i2c)アドレス値(0x3C～0x3Fの範囲)を指定する
  ~ColorKanjiLCDCore (); // デストラクタ

public:
  // TWI 送信
  errCode_t sendBegin (void);         // 送信開始：実際は TWI アドレスをセットするだけ
  size_t    sendData  (uint8_t dat);  // データ送信：実際は内部バッファ(最大32バイト)にデータをセットするだけ
  errCode_t sendEnd   (void);         // 送信終了：実際はここで初めて TWI 通信する。

public:
  // TWI 受信：一度に受信できるのは Wire ライブラリの制限で、最大32バイトまでです
  errCode_t recvData (uint8_t * dat, uint8_t siz); // データ受信：siz バイト分を受信し、dat 配列に入れる。dat 配列の大きさは siz 以上ある事。

public:
  errCode_t waitWhileBusy (void); // 漢字 LCD シールドがコマンド受付け可能になるまで待つ  ★

  // コマンド(nバイト)転送
  errCode_t command  (uint8_t cmd);
  errCode_t command2 (uint8_t cmd, uint8_t dat1);
  errCode_t command3 (uint8_t cmd, uint8_t dat1, uint8_t dat2);
  errCode_t command4 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3);
  errCode_t command5 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4);
  errCode_t command6 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5);
  errCode_t command7 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5, uint8_t dat6);
  errCode_t command8 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5, uint8_t dat6, uint8_t dat7);
  errCode_t command9 (uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5, uint8_t dat6, uint8_t dat7, uint8_t dat8);
  errCode_t command10(uint8_t cmd, uint8_t dat1, uint8_t dat2, uint8_t dat3, uint8_t dat4, uint8_t dat5, uint8_t dat6, uint8_t dat7, uint8_t dat8, uint8_t dat9);

public:
  // LCD 物理画面の情報
  uint8_t getPhyWidth  (void) const {return _phyWidth;}  // LCD 物理画面の幅  ：ランドスケープモード時は 160 が、ポートレートモード時は 128 が返されます。★
  uint8_t getPhyHeight (void) const {return _phyHeight;} // LCD 物理画面の高さ：ランドスケープモード時は 128 が、ポートレートモード時は 160 が返されます。★

protected:
  errCode_t getScreenMode (void); // 現在のスクリーンモード

public:
  // LCD 表示の ON/OFF
  errCode_t noDisplay (void); // LCD を表示させない
  errCode_t display   (void); // LCD を表示させる

public:
  // フリップモードの設定
  // CKLCD_FLIP_MODE_xxx のいずれかを指定します
  errCode_t setFlipMode (uint8_t flipMode);

public:
  // カラー漢字 LCD シールドのステータス
  errCode_t getStatus (void); // カラー漢字 LCD シールドの現在のステータスを取得します。カラー漢字 LCD シールドが Busy の場合、取得せずに CKLCD_ERR_BUSY エラーを返します ★

  // 以下の関数は、直前に getStatus 関数を呼び出し、CKLCD_ERR_NONE が返ってきたのを確認してから、呼び出してください。
  uint8_t getError    (void); // 現在のカラー漢字 LCD シールドのエラーコード : CKLCD_ERR_xxx のいずれかの値  ★
  uint8_t getDisplay  (void); // 現在の LCD 表示の状態 : 0=OFF, 1=ON  ★
  uint8_t getFlipMode (void); // 現在のフリップモード  ★
  uint8_t getLeftSW   (void); // 現在の左スイッチの状態 : 0=OFF, 1=ON  ★ ※ 約10ms間隔で更新されるので、それよりも短い間隔で読み出しても、同じ値しか読み出せません
  uint8_t getRightSW  (void); // 現在の右スイッチの状態 : 0=OFF, 1=ON  ★ ※ 同上
  uint8_t getUpSW     (void); // 現在の上スイッチの状態 : 0=OFF, 1=ON  ★ ※ 同上
  uint8_t getDownSW   (void); // 現在の下スイッチの状態 : 0=OFF, 1=ON  ★ ※ 同上

public:
  // グラフィック描画

  // 仮想画面#1/#2 をアルファ合成しながら LCD へ転送する事で、初めて LCD に表示されます
  errCode_t flip (void);  // 全表示領域  ★
  errCode_t flip (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1); // 指定領域のみ  ★

public:
  errCode_t res;     // レスポンス変数
  uint8_t   _count;  // ただのカウンター

protected:
  uint8_t _twiAddr; // カラー漢字 LCD シールドの TWI アドレス値: 0x3C～0x3F のいずれかの値

  uint8_t _phyWidth;  // LCD 物理画面の幅  ：ランドスケープモード時は 160 が、ポートレートモード時は 128
  uint8_t _phyHeight; // LCD 物理画面の高さ：ランドスケープモード時は 128 が、ポートレートモード時は 160

  uint8_t _currScreenMode; // 現在のスクリーンモード

  uint8_t _currStatus[2]; // カラー漢字 LCD シールドの現在のステータス
};


///////////////////////////////////////////////////////////////////
// カラー漢字 LCD 仮想画面クラスの宣言 ※Printクラスの派生クラス
// ※ 後述の ColorKanjiLCD クラス内で生成されるので、ユーザーがこのクラスを生成する必要はありません。
class ColorKanjiLCDVPlane : public Print {
public:
  ColorKanjiLCDVPlane (uint8_t vplane_no, ColorKanjiLCDCore& core); // コンストラクタ
  ~ColorKanjiLCDVPlane (); // デストラクタ

public:
  // 仮想画面の基本設定

  // 仮想画面の状態設定：CKLCD_VP_COND_xxx のいずれかを設定します
  // ※ ColorKanjiLCD::setScreenMode() から呼び出されます。ユーザーが変更する事も可能ですが、間違うとおかしな色で描画される事になるので注意して下さい。
  errCode_t setCondition (uint8_t cond);
  uint8_t   getCondition (void); // 仮想画面の状態取得

  // 仮想画面のオリエンテーションモード設定：CKLCD_VP_ORIENT_xxx のいずれかを設定します
  // ※ ColorKanjiLCD::setScreenMode()から呼び出されます。通常、ユーザーが設定する必要はありません。
  errCode_t setOrientMode (uint8_t orient);

  int16_t getWidth  (void) const {return _width;}  // 仮想画面の最大幅  ：ランドスケープモード時は 160 が、ポートレートモード時は 128 か 256 が返されます。 ★
  int16_t getHeight (void) const {return _height;} // 仮想画面の最大高さ：ランドスケープモード時は 128 か 256 が、ポートレートモード時は 160 が返されます。 ★

public:
  // 符号化方式の設定
  errCode_t setEncoding (uint8_t enc); // カラー漢字 LCD シールドの符号化方式を変更する: CKLCD_ENC_xxx のいずれかを指定する。範囲外の値を指定した時は CKLCD_ERR_INVALID_PARAM エラーを返します

public:
  // フォントタイプの設定
  // モノスペース(固定幅)か、プロポーショナル(可変幅)か、を指定します。
  errCode_t setFontType (uint8_t type);  // CKLCD_FONT_TYPE_xxx のいずれかを設定します
  // フォントの属性を設定します
  errCode_t setFontAttr (uint8_t attr);  // CKLCD_FONT_ATTR_xxx の組み合わせを設定します

  // フォント情報の取得
  errCode_t getFontInfo (void);  // 現在のフォント情報を取得する

  uint8_t getFontType   (void) const {return _currFontInfo[CKLCD_CURR_FONT_INFO_TYPE];}    // 現在のフォントタイプ：CKLCD_FONT_TYPE_xxx のいずれか
  uint8_t getFontWidth  (void) const {return _currFontInfo[CKLCD_CURR_FONT_INFO_WIDTH];}   // 現在のフォントの幅
  uint8_t getFontHeight (void) const {return _currFontInfo[CKLCD_CURR_FONT_INFO_HEIGHT];}  // 現在のフォントの高さ

public:
  // カラー漢字 LCD シールドの仮想画面のステータス
  errCode_t getStatus (void); // 現在のステータスを取得する ※カラー漢字 LCD シールドが Busy の場合、取得せずに CKLCD_ERR_BUSY エラーを返します

  // 以下の関数は、直前に getStatus 関数を呼び出し、CKLCD_ERR_NONE が返ってきたのを確認してから、呼び出す事
  uint8_t getEncoding      (void);  // 現在の符号化方式の状態 : CKLCD_ENC_xxx のいずれかの値
  uint8_t getLackedCode    (void);  // 文字コード欠乏状態：0=充足中, 1=欠乏中
  uint8_t getFontUnderline (void);  // 現在のフォント属性の設定：アンダーライン ON/OFF
  uint8_t getFontStrike    (void);  // 現在のフォント属性の設定：取り消し線 ON/OFF
  uint8_t getFontInverted  (void);  // 現在のフォント属性の設定：文字色反転 ON/OFF

public:
  // LCD 表示開始座標：この座標が LCD 画面の左上になります。
  errCode_t setDispStartPos (uint8_t   x, uint8_t   y);  // LCD 表示開始座標の指定：あくまで表示位置の設定だけであり、現在のカーソル位置には影響しません。 ★
  errCode_t getDispStartPos (uint8_t * x, uint8_t * y);  // 現在の LCD 表示開始座標の取得  ★

public:
  // 論理色の設定：
  //   描画色と背景色の２つがあります。基本的に、文字本体やグラフィック関係(pset,line,rect,circle)は描画色で描画されます。
  //   背景色は、文字や画像の背景に使われます。描画色と違う背景色を指定しておかないと、文字が見づらかったり、四角形が表示されるだけ、になってしまいます。

  //   ややこしいですが、「背景面」と「背景色」は、言葉は似ていますが、別物なので注意して下さい。
  //   前景面(ARGB4444)と背景面(RGB565)、それぞれ別々に「描画色」と「背景色」が設定できますので、カラー漢字LCDシールド内では、同時に４つの論理色を管理している事になります。

  // 描画色: 基本的にこの色が使用されます。
  errCode_t setDrawColor (uint8_t red, uint8_t green, uint8_t blue); // 論理色：ARGB8888 の内のR/G/B 指定のみ。Alphaは255になる。背景面(RGB565)用の論理色指定に最適
  errCode_t setDrawColor (uint8_t red, uint8_t green, uint8_t blue, uint8_t alpha); // 論理色：ARGB8888 を個別に指定。前景面(ARGB4444)用の論理色指定に最適
  errCode_t setDrawColor (cklcdCol32_t col32); // 論理色：ARGB8888 をダイレクトに指定

  // 背景色: write(), print(), clearRect(), roll(), copyWithRotate(), loadBmp(), loadPng(), loadGif() で、背景の色として使用されます。
  errCode_t setBackColor (uint8_t red, uint8_t green, uint8_t blue); // 論理色：ARGB8888 の内のR/G/B 指定のみ。Alphaは255になる。背景面(RGB565)用の論理色指定に最適
  errCode_t setBackColor (uint8_t red, uint8_t green, uint8_t blue, uint8_t alpha); // 論理色：ARGB8888 を個別に指定。前景面(ARGB4444)用の論理色指定に最適
  errCode_t setBackColor (cklcdCol32_t col32); // 論理色：ARGB8888 をダイレクトに指定

public:
  // 仮想画面に描画する時のアルファブレンド方法を指定します。
  //   CKLCD_BLEND_OVER：ブレンドしないで上書きします。処理速度は速いです。
  //   CKLCD_BLEND_ADD ：アルファ値を元に、ブレンド処理します。論理アルファ値==1.0 の時は上書きした時と同じ結果になります。各画素毎に計算するので、処理速度は遅いです。

  errCode_t setBlendMode (uint8_t blend);  // アルファブレンドの設定：CKLCD_BLEND_xxx のいずれかを設定
  uint8_t   getBlendMode (void);           // 現在のアルファブレンドの設定の取得

public:
  // 仮想画面のクリア

  // 仮想画面の全域を 0x0000 の絶対色でクリアし、LCD 表示開始座標とカーソル位置を (0,0) にします。
  errCode_t clear   (void);
  // 仮想画面の全域を 0x0000 の絶対色でクリアし、LCD 表示開始座標とカーソル位置を (0,0) にするか・しないかを選べます。CKLCD_CLR_FLG_XXX を組み合わせます。
  errCode_t clearEx (uint8_t flags);

  // 仮想画面の矩形領域(x0,y0)-(x1,y1)を setBackColor() で設定した論理色でクリアします。
  errCode_t clearRect (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1);

public:
  // 仮想画面をＹ方向に ySizeライン分スクロールUP/DOWN します。空いた領域は setBackColor() で設定した論理色になります。
  // 現在のカーソル位置と LCD 表示開始座標は変化しませんので、注意して下さい。
  // 実際は setSrcRect() と copy() と clearRect() の３つを呼び出しているので、遅いです。
  errCode_t scrollUp   (uint8_t ySize);  // 全領域
  errCode_t scrollUp   (uint8_t y0, uint8_t y1, uint8_t ySize);  // y0～y1 の範囲内
  errCode_t scrollDown (uint8_t ySize);  // 全領域
  errCode_t scrollDown (uint8_t y0, uint8_t y1, uint8_t ySize);  // y0～y1 の範囲内

public:
  // カーソル位置：
  // フォントがモノスペース(固定幅)の時は、文字の幅が固定なので setCursor() で指定できます（もちろん setCursorMin() でもOKです）。
  // プロポーショナル(可変幅)の時は、文字の幅が一意ではありませんので、より詳細な位置を指定できる setCursorMin() でカーソル位置を指定して下さい。

  errCode_t home (void); // 画面はそのままで、LCD 表示開始座標とカーソル位置の両方を (0,0) にします。

  // カーソル位置(キャラクタ座標)を設定: フォントがモノスペース(固定幅)の時に使用できます。
  // ※ 実際は内部で setCursorMin(col*fontWidth, row*fontHeight) を呼び出しているだけです。
  errCode_t setCursor (uint8_t col, uint8_t row);

  // 現在のカーソル位置(キャラクタ座標)を取得します。フォントがモノスペース(固定幅)の時に使用できます。
  errCode_t getCursor (uint8_t * col, uint8_t * row);

  // カーソル位置(グラフィック座標)を設定: x=0～CKLCD_VPLANEn_MAX_o_WIDTH-1, y=0～CKLCD_VPLANEn_MAX_o_HEIGHT-1 の範囲で指定。nは1か2、oはLかPです。
  // フォントがプロポーショナル(可変幅)の時は、こちらを使用して下さい。
  errCode_t setCursorMin (uint8_t x, uint8_t y);

  // 現在のカーソル位置(グラフィック座標)を取得します。
  // 型が uint16_t * なのは、カーソル移動の結果、仮想画面の範囲外に出ている可能性が有るためです。
  errCode_t getCursorMin (uint16_t * x, uint16_t * y);

public:
  // 文字コード(UTF-8/SJIS/EUC-JP)転送：
  //   フリップモードが CKLCD_FLIP_MODE_AUTO の時は、仮想画面に文字を描画後、自動的に仮想画面から LCD 画面に転送されます。
  //   一方、フリップモードが CKLCD_FLIP_MODE_MANU の時は、仮想画面に文字が描画されるだけで、
  //   自動的に LCD 画面には転送されませんので、ユーザーが明示的に flip() を呼び出す必要があります。

  //   モノスペース(固定幅)の時、カーソル位置は、半角文字表示後に +8、全角文字表示後に +16 されます。
  //   プロポーショナル(可変幅)の時は、漢字表示後に +16、それ以外の文字の時は +5～+16 のいずれかの値が加算されます。

  //   カーソル位置が仮想画面の右端に来ても自動的に改行はされず、カーソル位置は仮想画面の範囲外のままです。
  //   0x0D('\r')の復帰コードで、行頭に移動します。つまり (0,y) のカーソル位置に移動します。
  //   また、0x0A('\n')の改行コードで改行されます。つまり (0,y+20) のカーソル位置に移動します。
  //   カーソル位置が仮想画面の下端を越えても自動的にスクロールとかはされず、カーソル位置は仮想画面の範囲外のままです。

  //   カーソル位置が仮想画面の範囲外の状態で write/print しても、LCD 画面上はおろか、仮想画面上にすら、文字は描画されません。
  //   ユーザーが明示的に、仮想画面上の描画可能域に、カーソル位置を移動させる必要があります。

  using Print::write;
  size_t write (uint8_t dat); // 文字コードを１バイト転送する：基底クラスの Print から呼び出される
  size_t write (const uint8_t * buffer, size_t size); // 文字コードを指定バイト分まとめて転送する：基底クラスの Print の関数をオーバーライド

  size_t print_P (const prog_char * pstr); // プログラムメモリ上の文字列を転送する
  size_t write_P (const prog_char * buffer, size_t size); // プログラムメモリ上の文字コードを指定バイト分まとめて転送する

public:
  // ワイド文字(UTF-16)対応：
  //   ASCII 文字が少なく、和文が多い場合は、こちらの方がメモリ効率が良くなります。
  //   ただし、Arduino IDE は、ワイド文字に対応した文字列操作関数が存在しないので、文字列の操作に不自由します。
  //   なお、以下の関数を使用する時は、先に setEncoding 関数で UTF-16 方式に切り替えておいてください。

  size_t wprint   (wchar_t wch); // UTF-16 文字を１文字転送する
  size_t wprint   (const wchar_t * wstr); // UTF-16 文字列を転送する
  size_t wprint_P (const prog_wchar * pwstr); // プログラムメモリ上の UTF-16 文字列を転送する

public:
  // 特殊効果付きの文字列の描画
  // 一旦、仮想画面上で処理してから、一気に液晶画面に表示するようにします。この為、内部で ColorKanjiLCDCore::setFlipMode() が呼び出されます。
  // 注意：現在の setDrawColor() の設定が変更されてしまいますので、再度、setDrawColor() で論理色を指定しなおす必要があります。なお、setBackColor() の方は変更されません。

  // 影付き文字列の描画
  // X/Y ともに offset 分ずれた位置に colShadow 色の影がある colBody 色の文字列を描画します。
  // 注意：余りにも離れた位置に影があると、影付き文字とは感じられなくなってしまいますので、offset 値は程々にしておくべきでしょう。
  // 内部では、str を colShadow 色で描画、blur 処理、再度 str を colBody 色でアルファブレンド描画しています。
  errCode_t shadowPrint_M (uint8_t x, uint8_t y, uint8_t offset, cklcdCol32_t colBody, cklcdCol32_t colShadow, const void * str, uint8_t memType);
#define shadowPrint(x,y,offset,colBody,colShadow,str)    shadowPrint_M(x,y,offset,colBody,colShadow,str,0)   // SRAM
#define shadowPrint_P(x,y,offset,colBody,colShadow,str)  shadowPrint_M(x,y,offset,colBody,colShadow,str,1)   // PROGMEM

  // 縁取り付き文字列の描画
  // colStroke 色の縁取りが付いた colBody 色の文字列を描画します。遅いです。
  // 注意：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate 処理がかかり、おかしくなってしまいます。
  // 内部では、str を colStroke 色で描画、dilate 処理、再度 str を colBody 色でアルファブレンド描画しています。
  errCode_t strokePrint_M (uint8_t x, uint8_t y, cklcdCol32_t colBody, cklcdCol32_t colStroke, const void * str, uint8_t memType);
#define strokePrint(x,y,colBody,colStroke,str)    strokePrint_M(x,y,colBody,colStroke,str,0)   // SRAM
#define strokePrint_P(x,y,colBody,colStroke,str)  strokePrint_M(x,y,colBody,colStroke,str,1)   // PROGMEM

  // グロー効果の文字列の描画
  // colGlow 色のグロー効果が付いた colBody 色の文字列を描画します。すごく遅いです。
  // 注意：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate & blur 処理がかかり、おかしくなってしまいます。
  // 内部では、str を colGlow 色で描画、dilate 処理、blur 処理×２回、再度 str を colBody 色でアルファブレンド描画しています。
  errCode_t glowPrint_M (uint8_t x, uint8_t y, cklcdCol32_t colBody, cklcdCol32_t colGlow, const void * str, uint8_t memType);
#define glowPrint(x,y,colBody,colGlow,str)    glowPrint_M(x,y,colBody,colGlow,str,0)   // SRAM
#define glowPrint_P(x,y,colBody,colGlow,str)  glowPrint_M(x,y,colBody,colGlow,str,1)   // PROGMEM

  // ネオン効果の文字列の描画
  // colGlow2 色のグロー効果に、さらに colGlow1 色のグロー効果が付いた colBody 色の文字列を描画します。激しく遅いです。
  // 注意1：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate & blur 処理がかかり、おかしくなってしまいます。
  // 注意2：colGlow1 色と colGlow2 色の組み合わせによっては、２回目の dilate 処理の時に破綻して、変なネオン効果になる場合があります。
  // 内部では、str を colGlow2 色で描画、dilate 処理、blur 処理、str を colGlow1 色で再描画、再 dilate 処理、再 blur 処理、str を colBody 色で再々描画しています。
  errCode_t neonPrint_M (uint8_t x, uint8_t y, cklcdCol32_t colBody, cklcdCol32_t colGlow1, cklcdCol32_t colGlow2, const void * str, uint8_t memType);
#define neonPrint(x,y,colBody,colGlow1,colGlow2,str)    neonPrint_M(x,y,colBody,colGlow1,colGlow2,str,0)   // SRAM
#define neonPrint_P(x,y,colBody,colGlow1,colGlow2,str)  neonPrint_M(x,y,colBody,colGlow1,colGlow2,str,1)   // PROGMEM

public:
  // グラフィック描画
  // 一旦、仮想画面#1/#2 に描いておいてから、ColorKanjiLCD::flip 関数で一括して LCD へ転送する事で、初めて表示されます。
  // なお、line 関数, circle 関数, circleFill 関数では、スムージング処理は行ってませんので、ジャギーが発生します。

  // 仮想画面に点(x,y)を描画：setBlendMode()の手法で setDrawColor() の論理色で描画します。
  errCode_t pset (uint8_t x, uint8_t y);

  // 仮想画面に線(x0,y0)-(x1,y1)を描画：setBlendMode()の手法で setDrawColor() の論理色で描画します。
  errCode_t line (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1);

  // 仮想画面に矩形(x0,y0)-(x1,y1)を描画：setBlendMode()の手法で setDrawColor() の論理色で描画します。
  errCode_t rect     (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1); // 枠線のみ
  errCode_t rectFill (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1); // 内部も塗りつぶす

  // 仮想画面に丸(x,y)を描画：setBlendMode()の手法で setDrawColor() の論理色で描画します。
  errCode_t circle     (uint8_t x, uint8_t y, uint8_t radius); // 円周線のみ
  errCode_t circleFill (uint8_t x, uint8_t y, uint8_t radius); // 内部も塗りつぶす

public:
  // 仮想画面内でのコピー
  // 注意：基本的には、同一の仮想画面内でコピーする事が望ましいです。これらの関数は、２種類ある16bit絶対色の変換は行わず、単なる16bitの値として、そのままコピーするだけです。
  //       仮想画面#1と#2にまたがってのコピーもできますが、その場合、16bit絶対色が同じではない可能性に気をつける必要があります。
  //       片方が ARGB4444 で、もう片方が RGB565 の場合、コピー処理そのものは正しく行われますが、コピー後の画像は滅茶苦茶な色になっているでしょう。
  // copy() の場合、同一の仮想画面内でコピー元とコピー先の領域が部分的に重なってる場合でも、正しくコピーされます。
  // 一方、copyWithRotate() の方は、コピー元とコピー先の領域は部分的にも重なっていてはいけません（暴走こそしませんが、おかしな結果になります）。

  // コピー元の矩形領域(x0,y0)-(x1,y1)を指定します。copy()の前に指示します。この関数単独では仮想画面を変化させる事はありません。
  // なお、仮想画面#1と#2にまたがってコピー出来るようにした為、ある瞬間に setSrcRect() で指定できるコピー元矩形領域は、仮想画面#1・#2のどちらか片方だけです。
  // 仮想画面#1のコピー元矩形領域と、仮想画面#2のコピー元矩形領域を、同時に別々に設定させる事はできませんので注意して下さい。最後に設定した方だけが有効です。
  errCode_t setSrcRect (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1);

  // setSrcRect()で指定されている矩形領域を(x,y)～(x+w-1,y+h-1)にコピーします。w と h はコピー元矩形領域の幅と高さが、そのまま使用されます。
  // 1:1 のコピーだけで、拡大/縮小とかはしません。また、仮想画面の領域からはみ出した部分は、単純にクリッピングされます。
  // 同じ仮想画面で、コピー元とコピー先が同一の場合、CKLCD_ERR_SAME エラーになります（違う仮想画面の時はエラーにならない）。
  // 仮想画面#1と#2にまたがってのコピーもできますが、上記の 16bit絶対色の注意事項に注意して下さい。
  errCode_t copy (uint8_t x, uint8_t y);

  // 回転＆拡大/縮小させながらコピー：劇遅!!
  // setSrcRect()で指定されている矩形領域を回転("BAM10"単位)させながら、(x0,y0)-(x1,y1)の範囲内にコピーします。
  // invRatio10 倍の拡大/縮小もできますが、単なる近傍点補間しかしてないので、結果画像は汚いです。大昔の家庭用ゲーム機みたいな効果しかありません。
  // 回転の中心はそれぞれ、コピー元の矩形領域の中心座標と、コピー先の矩形領域の中心座標、になります。
  // 回転＆拡大/縮小の時、コピー元の矩形領域に存在し無い部分は setBackColor()で設定している色になります。
  // コピー元の矩形領域と、コピー先の矩形領域は、部分的にも重なっていてはいけません。重なってた場合、おかしな結果画像になってしまいます。
  errCode_t copyWithRotate (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, cklcdInvRatio10_t invRatio10, cklcdBAM10_t bam10);

public:
  // 画像処理
  // ３つしかありません。基本的に特殊効果付きの文字列に使用する処理だけです。

  // 膨張処理：(x0,y0)-(x1,y1)の領域内に 3x3 の膨張処理を行います。
  // 文字のグロー/ネオン効果の為に追加したので、かなりインチキくさい処理です。一般的な画像に対して膨張処理させると、見るも無残な結果になるだけ、と思われます。
  errCode_t dilate (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1);

  // 収縮処理：(x0,y0)-(x1,y1)の領域内に 3x3 の収縮処理を行います。
  // こちらも、かなりインチキくさい処理です。一般的な画像に対して収縮処理させると、見るも無残な結果になるだけ、と思われます。
  errCode_t erode (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1);

  // ブラー処理：(x0,y0)-(x1,y1)の領域内に 3x3 のガウシアンフィルター処理を行います。とても遅いです。
  // 文字列だけでなく、一般的な画像に対しても効果があります。
  errCode_t blur (uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1);

public:
  // bmp/jpg/png/gif 画像の読込み
  // 画像データを仮想画面上に展開します。当たり前ですが、バイト数が多い画像データほど、展開に時間が掛かります。
  // なお、サポートしているのは読込み(Load)機能のみで、保存(Save)機能はありません。

  // 注意１：png 画像と gif 画像の２つは、画像展開に際し、大量の作業用メモリが必要になります。この為、もう片方の仮想画面を作業用メモリとして使用しますので、
  //         そちらの仮想画面の現在の内容が破壊されてしまいます。スクリーンモードで２画面重ね合わせのモードを選んでいる時は、注意して下さい。

  // 注意２：各画像フォーマット毎に色々な制限事項があります。制限事項に逸脱した画像データの場合、読み込みエラーになる可能性が有りますので注意して下さい。

  // 背景面に展開した場合、RGB565 として展開されます。ディザ拡散などは行わず、単に下位bitデータを削っているだけですので、画像によっては、擬似輪郭などが発生する場合があります。
  // 前景面に展開した場合、ARGB4444 として展開されます。こちらも単純に下位bitデータを削っているだけですので、より一層、擬似輪郭などが発生する場合があります。
  // アルファ値や透過色を持たない画像を前景面に展開する事も可能ですが、α=1.0 固定の、実質 RGB444 の画像になるだけですので、お奨めしません。背景面に展開しましょう。
  // 逆に、アルファ値や透過色を持つ画像を背景面に展開すると、せっかくのアルファ値や透過色が無くなってしまいますので、こちらの画像は前景面に展開しましょう。

  // アルファ値有り(PNG/BMP)or透過色有り(GIF)の画像の場合、仮想画面の状態設定と、ブレンド処理の設定から、以下の様に処理されます。
  //     非透過モード ＋ 上書き        :  現在の setBackColor()で設定されている色に対してαブレンドしながら、RGB565 として上書きされる。仮想画面に現在描画されている画像は無視される。
  //     非透過モード ＋ αブレンドON  :  仮想画面に現在描画されている画像に対してαブレンドされ、RGB565 として書き込まれる。アルファ値は保持されない。
  //     透過モード   ＋ 上書き        :  アルファ値を保持したまま、ARGB4444 として上書きされる。仮想画面に現在描画されている画像は無視される。
  //     透過モード   ＋ αブレンドON  :  仮想画面に現在描画されている画像に対してαブレンドされ、ARGB4444 として書き込まれる。アルファ値は保持される。
  // アルファ値なし(JPG/PNG/BMP)or透過色なし(GIF)の画像の場合は、ブレンド処理の現在の設定は無視されて、以下の様に処理されます。
  //     非透過モード  :  完全不透明の  RGB565  として上書きされる。仮想画面に現在描画されている画像は無視される。
  //     透過モード    :  完全不透明の ARGB4444 として上書きされる。仮想画面に現在描画されている画像は無視される。

  // 搭載メモリー容量の関係上、全画像データは一気に読込めないため、cklcdLoadImgCallback_t 型のユーザー定義 callback 関数を使って、ちまちまと読込みます。
  // カラー漢字LCDシールド側から、buff_size バイト分の画像データが欲しいと要求してきますので、callback 関数内で buff ポインタが示す配列に画像データを転送して下さい。
  // 同じくメモリー容量の関係上、読み込める画像の寸法にも制限があります（特に幅の大きさ）。１画素でもオーバーすると、問答無用で寸法エラーになりますので、注意して下さい。

  // jpg 画像の読込み
  // 読み込める jpg 画像は、最大 1024×1024 画素までです。これよりも大きい場合は、画像寸法エラーになります。
  // Chan 氏作成の TJpgDec R0.01b を使用しています。
  // 読み込める jpg 画像は、一般的な JFIFフォーマットのみです。ガンマ値の JFIFタグや、テキスト等の画像データに関与しない JFIFタグは、在っても無視します。
  // また、プログレッシブ(Progressive)タイプの jpg は扱えません。必ず CKLCD_ERR_JPG_UNSUP_PRGRS エラーになりますので、お絵かきソフト等で非プログレッシブに変換しておいてください。
  // それと、YCbCr 表色系でない JFIF フォーマットや、一般的には使われないサンプリングファクターが使用された JFIF フォーマットも、エラーになります（TJpgDec が対応してない為）。
  // jpg 画像のサイズが(w,h)ならば、仮想画面の(xPos,yPos)～(xPos+w-1,yPos+h-1)の範囲に描画します。ただし、仮想画面の領域からはみ出した部分はクリッピングされます。
  // jpg 画像にアルファ値は有りませんので、setBlend() の設定に関わらず、常に上書きされます(アルファブレンドされません)。
  // なお、仮想画面が透過モード(ARGB4444)の時は、論理アルファ値==1.0(不透明)として描画します。
  errCode_t loadJpg      (int16_t xPos, int16_t yPos, uint8_t effect, uint16_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data); // ファイルサイズが 64kB未満の時
  errCode_t loadJpgLarge (int16_t xPos, int16_t yPos, uint8_t effect, uint32_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data); // ファイルサイズが 64kB以上の時

  // png 画像の読込み
  // 注意!!：読込む仮想画面では無い方の仮想画面を作業エリアとして破壊します。例えば、仮想画面#1 に読み込む時は、仮想画面#2 の内容が破壊されます。逆の時も同じです。
  // 読み込める png 画像は、作業用メモリの関係で、最大 256×1024 画素までです。これよりも大きい場合は、画像寸法エラーになります。
  // libpng v1.6.1 と zlib 1.2.8 を使用しています。ただし、かなり改変しています（オリジナル版だとメモリ消費量が膨大な為、必要最小限まで削りました）。
  // 読み込める png 画像は、0x89+"PNG" から開始する一般的なフォーマットのみです。
  // 処理できるタグは、表示に必要な最低限レベルである IDAT, IEND, IHDR, PLTE, sBIT, tRNS タグのみです。これ以外のタグは、在っても無視されます。
  // この為、ガンマ補正や色空間設定などは読み飛ばして無視されます。oFFs タグは今一、仕様が理解できなかったのと、サンプル画像が見つからなかったので無視する事にしました。
  // また、対応しているチャンネル毎のbit深度(depth)は 1/2/4/8bit までです。16bit深度には対応していません(エラーになります)。
  // なお、ここで言う16bit深度とは、16bitグレー、16bitグレー＋16bitアルファ、48bitR/G/B、48bitR/G/B＋16bitアルファ、の４種類を指します。
  // また、インターレース形式の png は、処理が余りにも複雑な為、非サポートです。必ずエラーになりますので注意して下さい。(libpng が余りにもややこしいのでバッサリと削っちゃった)
  // png 画像のサイズが(w,h)ならば、仮想画面の(xPos,yPos)～(xPos+w-1,yPos+h-1)の範囲に描画します。ただし、仮想画面の領域からはみ出した部分はクリッピングされます。
  // アルファ値を持つ png 画像の場合、CKLCD_BLEND_xxx に従ってブレンド処理されます。アルファ値無しの時は常に上書き、です。
  errCode_t loadPng      (int16_t xPos, int16_t yPos, uint8_t effect, uint16_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data); // ファイルサイズが 64kB未満の時
  errCode_t loadPngLarge (int16_t xPos, int16_t yPos, uint8_t effect, uint32_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data); // ファイルサイズが 64kB以上の時

  // gif 画像の読込み
  // 注意!!：読込む仮想画面では無い方の仮想画面を作業エリアとして破壊します。例えば、仮想画面#1 に読み込む時は、仮想画面#2 の内容が破壊されます。逆の時も同じです。
  // 読み込める gif 画像は、作業用メモリの関係で、最大 640×1024 画素までです。これよりも大きい場合は、画像寸法エラーになります。
  // giflib 5.0.4 を使用しています。ただし、若干改変しています（オリジナル版だとmalloc/freeしまくって、アニメgif時に支障をきたすので、最大640画素決め打ちでmalloc１回のみにした）
  // 読み込める gif 画像は、"GIF87A" もしくは "GIF89A" から開始する一般的なフォーマットのみです。インターレース gif にも対応しています(インターレース png よりは展開方法が簡単な為)。
  // gif 画像のサイズが(w,h)ならば、仮想画面の(xPos,yPos)～(xPos+w-1,yPos+h-1)の範囲に描画します。ただし、仮想画面の領域からはみ出した部分はクリッピングされます。
  // アニメーション gif は、各フレームごとに、cklcdImgFrameCallback_t 型のユーザー定義関数の funcFrame が呼び出されますので、アニメーションさせる事が可能です。
  // ただし、大きな画像になると、読み込み＆描画に時間が掛かるので、スムーズなアニメーションにはならなくなります。また、アニメーション中は呼び出せる関数に制限があります。
  // なお、仮想画面が透過モードの時は、透過ピクセル部分は論理アルファ値==0.0(透明)、それ以外の画素は論理アルファ値==1.0(不透明)として描画します。
  // 透過ピクセルが有る gif 画像を Back面に上書き(CKLCD_BLEND_OVER)する時は、setBackColor()で設定している色になります。
  errCode_t loadGif      (int16_t xPos, int16_t yPos, uint8_t effect, uint16_t file_size, cklcdLoadImgCallback_t funcLoad, cklcdImgFrameCallback_t funcFrame, ColorKanjiLCDLoadImgSize * user_def_data); // ファイルサイズが 64kB未満の時
  errCode_t loadGifLarge (int16_t xPos, int16_t yPos, uint8_t effect, uint32_t file_size, cklcdLoadImgCallback_t funcLoad, cklcdImgFrameCallback_t funcFrame, ColorKanjiLCDLoadImgSize * user_def_data); // ファイルサイズが 64kB以上の時

  // bmp 画像の読込み
  // 読み込める bmp 画像は、"BM" から開始する一般的なフォーマットのみです。
  // 自作の展開エンジンにて展開しています。
  // 読み込める bmp 画像は、作業用メモリの関係で、最大 640×1024 画素までです。これよりも大きい場合は、画像寸法エラーになります。
  // bmp 画像のサイズが(w,h)ならば、仮想画面の(xPos,yPos)～(xPos+w-1,yPos+h-1)の範囲に描画します。ただし、仮想画面の領域からはみ出した部分はクリッピングされます。
  // なお、無圧縮(BI_RGBかBI_BITFIELDS)の bmp 画像にのみ対応しており、RLE 圧縮(BI_RLE4/BI_RLE8)の bmp 画像には対応していません。
  // また、OS/2フォーマット(BITMAPCOREHEADER: サイズが12byteのヤツ)にも対応していません(今時、そんな bmp 画像ファイルは無いでしょうが)。
  // BITMAPINFOHEADER(40byte), BITMAPV4HEADER(108byte), BITMAPV5HEADER(124byte) に対応してます。ただし、色空間、ガンマ値、プロファイルデータは在っても無視します。
  // パレット情報(RGBQUAD)を持つ bmp 画像は、そのまま読み込めます。高さ情報(biHeight)が負の bmp 画像も、そのまま読み込めます。
  // アルファ値を持つ bmp 画像の場合、CKLCD_BLEND_xxx に従ってブレンド処理されます。アルファ値無しの時は常に上書き、です。
  errCode_t loadBmp      (int16_t xPos, int16_t yPos, uint8_t effect, uint16_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data); // ファイルサイズが 64kB未満の時
  errCode_t loadBmpLarge (int16_t xPos, int16_t yPos, uint8_t effect, uint32_t file_size, cklcdLoadImgCallback_t funcLoad, ColorKanjiLCDLoadImgSize * user_def_data); // ファイルサイズが 64kB以上の時

private:
  // 画像読み込みループ関数：内部使用
  errCode_t loopLoadImg (cklcdLoadImgCallback_t funcLoad, cklcdImgFrameCallback_t funcFrame, ColorKanjiLCDLoadImgSize * user_def_data);
  // 最後に LoadImg した画像のサイズを取得する関数：内部使用
  errCode_t getLoadImgSize (ColorKanjiLCDLoadImgSize * user_def_data);

public:
  // 16bit絶対色の直接操作
  // 仮想画面の設定が Front面か Back面により、ARGB4444 か RGB565 のどちらかになります。
  // 仮想画面を LCD 表示に使用せず、単なる cklcdCol16_t (uint16_tと同じ) の SRAM 二次元配列として使用する時も、この関数を使用して下さい。
  errCode_t setPixel (uint8_t x, uint8_t y, cklcdCol16_t   col16); // 仮想画面上の絶対座標(x,y)に、16bit絶対色を上書きします。setBlend()の設定は無視されます。
  errCode_t getPixel (uint8_t x, uint8_t y, cklcdCol16_t * col16); // 仮想画面上の絶対座標(x,y)の、現在の16bit絶対色を取得します。

private:
  // メンバー変数
  uint8_t _vplane_no; // 仮想画面番号: CKLCD_VPLANE1 か CKLCD_VPLANE2 のいずれか
  uint8_t _cond;      // 仮想画面の状態設定：CKLCD_VP_COND_xxx のいずれか
  uint8_t _blend;     // アルファブレンドの設定：CKLCD_BLEND_xxx のいずれか
  uint8_t _merge;     // 上記設定をORしたもの

  uint8_t _orient;    // 仮想画面のオリエンテーション設定：CKLCD_VP_ORIENT_xxx のいずれか

  uint8_t _currFontInfo[3];  // 現在のフォント情報

  uint8_t _currVPStatus; // カラー漢字 LCD シールドの仮想画面の現在の状態

  int16_t _width;  // 仮想画面の最大幅  ：ランドスケープモード時は 160、ポートレートモード時は 128 か 256
  int16_t _height; // 仮想画面の最大高さ：ランドスケープモード時は 128 か 256、ポートレートモード時は 160

  ColorKanjiLCDCore& _core; // TWI(i2c)通信用
};


//////////////////////////////////////////////
// カラー漢字 LCD クラス宣言
// 主に TWI(i2c)通信を行う ColorKanjiLCDCore クラスの派生クラスです。
// ※ インスタンスを生成すると、ColorKanjiLCDVPlane クラスも２つ生成されます。public の v1, v2 メンバー変数として使用できます。
class ColorKanjiLCD : public ColorKanjiLCDCore {
public:
  ColorKanjiLCD (uint8_t twiAddr); // コンストラクタ：カラー漢字 LCD シールドの TWI(i2c) アドレス値(0x3C～0x3Fの範囲)を指定する
  ~ColorKanjiLCD (); // デストラクタ

public:
  ColorKanjiLCDVPlane v1; // 仮想画面#1：160×256画素サイズ固定
  ColorKanjiLCDVPlane v2; // 仮想画面#2：160×128画素サイズ固定

public:
  // カラー漢字 LCD シールドを初期化します。
  // スクリーンモードと符号化方式を指定します。これらは、後で個別に変更することも可能です。
  // 正常に実行された場合は、画面クリア、カーソル位置 (0,0)、LCD 表示ON、フリップモード=AUTO、になります
  // 注意：カラー漢字 LCD シールドは電源ONした直後、200ms ほど待たないと TWI 通信が開始できません。waitTime にはその待ち時間(ミリ秒)を指定します。
  //       デフォルト値は 200 ですが、この関数を呼び出す前に、別の処理で例えば 20ms 待っていれば、その分を差っ引いた 180 にしても構いません。
  errCode_t begin (uint8_t screenMode, uint8_t enc = CKLCD_ENC_UTF8, uint8_t waitTime = 200);

  // 終了処理
  errCode_t end (void);

public:
  // スクリーンモードの設定
  // CKLCD_ORIENT_xxx のいずれかと、CKLCD_DISP_SCREEN_xxx のいずれかを OR 指定します。
  // 注意1：CKLCD_ORIENT_xxx を変更した場合、x軸/y軸の入れ替わりと、原点位置が移動する為、現在の LCD 表示開始座標とカーソル位置は、意図しないものになります。
  //        この関数の直後に、仮想画面#1/#2 それぞれの LCD 表示開始座標とカーソル位置を設定し直した方が良いでしょう。
  // 注意2：この関数内で、各仮想画面の setCondition() をセットしています。
  errCode_t setScreenMode (uint8_t screenMode);

public:
  // 仮想画面#1&#2 のクリア

  // 仮想画面#1&#2 の全域を 0x0000 の絶対色でクリアし、LCD 表示開始座標とカーソル位置を (0,0) にします。
  errCode_t clear (void);

public:
  // カラー漢字 LCD シールドの状態

  // 以下は、begin 関数内で取得されている情報
  uint8_t getKlcdType (void); // カラー漢字 LCD シールドのタイプ: 常に "BS21LAB-006" を意味する 0x06 が返される
  uint8_t getFirmVer  (void); // カラー漢字 LCD シールドのファームウェアの Version 番号 (例: v1.2 の時は 0x12 になる)

private:
  uint8_t _klcdType; // カラー漢字 LCD シールドのタイプ : 現ライブラリは 0x06 のみに対応
  uint8_t _firmVer;  // カラー漢字 LCD シールドのファームウェアの Version 番号 (例: v1.2 の時は 0x12 になる)
};

#endif // __COLORKANJILCD_H__

