/*
  Color Kanji LCD Shield - EUC-JP Demo #1

  本スケッチは "カラー漢字 LCD シールド" で、EUC-JIS-2004 の一覧を表示させるデモスケッチ例です。
  左右スイッチで、EUC-JIS-2004 符号化コードを増減させます。
  上下スイッチで、フォントタイプを切り替えられます。

  Circuit:
    pins A4(SDA), A5(SCL) for TWI communication in Arduino standard.
    pins SDA, SCL for TWI communication in Arduino UNO R3, Leonardo.

  created 14/aug/2013 by I.Kato @ BS21 Lab

  This code is in the public domain.
*/

// TWI 通信用ライブラリ：必須です
#include <Wire.h>

// カラー漢字 LCD 用ライブラリ
#include <ColorKanjiLCD.h>


// カラー漢字 LCD シールドの TWI アドレスを 0x3C 以外に設定した場合は、以下の値を書きかえてください。
// TWI 通信できていない時は、カラー漢字 LCD シールド自身に現在設定されている TWI(i2c) アドレスが
// 表示されているはずなので、御確認ください。
#define CKLCD_TWI_ADDR  0x3C


// カラー漢字 LCD のインスタンス生成
ColorKanjiLCD cklcd (CKLCD_TWI_ADDR); // カラー漢字 LCD シールドの TWI アドレスを指定する


#define MAX_DISP_NO  (94*3+6)

int16_t disp_no; // 表示させたい画面の番号
int16_t curr_no; // 現在、表示している画面の番号

uint8_t font_type_no; // 現在のフォントタイプ番号

// フォントタイプのリスト
const uint8_t FontTypeList[] = {
  CKLCD_FONT_TYPE_RUBY_MS,      // ルビ文字のモノスペース(固定幅)
  CKLCD_FONT_TYPE_RUBY_P,       // ルビ文字のプロポーショナル(可変幅)
  CKLCD_FONT_TYPE_12px_MS,      // 12px文字のモノスペース(固定幅)
  CKLCD_FONT_TYPE_12px_P,       // 12px文字のプロポーショナル(可変幅)
  CKLCD_FONT_TYPE_MONOSPACED,   // モノスペース(固定幅)
  CKLCD_FONT_TYPE_PROPORTIONAL, // プロポーショナル(可変幅)
  CKLCD_FONT_TYPE_20px_MS,      // 20px文字のモノスペース(固定幅)
  CKLCD_FONT_TYPE_20px_P,       // 20px文字のプロポーショナル(可変幅)
  CKLCD_FONT_TYPE_24px_MS,      // 24px文字のモノスペース(固定幅)
  CKLCD_FONT_TYPE_24px_P,       // 24px文字のプロポーショナル(可変幅)
  CKLCD_FONT_TYPE_28px_MS,      // 28px文字のモノスペース(固定幅)
  CKLCD_FONT_TYPE_28px_P,       // 28px文字のプロポーショナル(可変幅)
};
#define MAX_FONT_TYPE_LIST (sizeof(FontTypeList) / sizeof(*FontTypeList))

// フォント名のリスト
const char FontNameList[][8] = {
  "ruby:MS",
  "ruby:P",
  "12px:MS",
  "12px:P",
  "std:MS",
  "std:P",
  "20px:MS",
  "20px:P",
  "24px:MS",
  "24px:P",
  "28px:MS",
  "28px:P",
};

uint8_t left_sw, right_sw, up_sw, down_sw; // カラー漢字 LCD シールドの４つのスイッチの状態


// 初期化関数
void setup()
{
  // 必ず、最初に Wire ライブラリを初期化させておいてください。
  // 注意："CKLCD_Wire" は Arduino Due では "Wire1" を、それ以外では "Wire" になります。
  CKLCD_Wire.begin();

  // 続いて、カラー漢字 LCD ライブラリを初期化します。
  // 符号化方式は EUC-JP 方式に設定します。
  // 初期化に成功すると、画面はクリアされ、LCD 表示方向はランドスケープ(横向き) に、カーソル位置は (0,0) になります。
  // カラー漢字 LCD は、電源投入直後は自動的に Version と TWI アドレスを表示していますので、
  // 初期化に失敗した場合は、TWI アドレスが、上記の CKLCD_TWI_ADDR と一致しているか確認してください。
  // 仮想画面１(160×256) を背景面に設定します。仮想画面２(160×128) は使用しません。
  cklcd.begin (CKLCD_ORIENT_LANDSCAPE + CKLCD_DISP_SCREEN_B1, CKLCD_ENC_EUC_JP);


  // フォントタイプをモノスペース(固定幅)にします。
  font_type_no = 4;

  disp_no =  0; // 表示させたい画面の番号：とりあえず1番目を表示させる
  curr_no = -1; // 現在、表示している画面の番号：無しを意味する -1に設定しておく

  // スイッチは押されていない状態にする
  left_sw = right_sw = up_sw = down_sw = false;
}


// ループ関数
void loop()
{
  // 最初にカラー漢字 LCD の現在のステータスを取得します。
  // これを呼び出しておかないと、以下の getLeft/Up/Down/RightSW 関数が
  // 何時までも更新されない事になってしまいます。
  // カラー漢字 LCD がコマンド処理中で忙しい時は、ステータス取得に失敗して CKLCD_ERR_BUSY が返って来ます。
  // ステータス取得成功を意味する CKLCD_ERR_NONE が返ってきた時だけ、処理を進めるようにして下さい。
  if (cklcd.getStatus() == CKLCD_ERR_NONE) {
    // ステータス取得成功した

    // 左スイッチの処理
    if (cklcd.getLeftSW()) { // 左スイッチが押されている
      // 押され続けている間はフラグを立てるだけの処理
      left_sw = true;
    } else { // 左スイッチが離された
      if (left_sw == true) { // さっきまで左スイッチが押されていた
        if (disp_no == 0) {
          disp_no = MAX_DISP_NO;
        }
        disp_no--;
        // 左スイッチ・フラグをクリア
        left_sw = false;
      }
    }

    // 上スイッチの処理
    if (cklcd.getUpSW()) { // 上スイッチが押されている
      // 押され続けている間はフラグを立てるだけの処理
      up_sw = true;
    } else { // 上スイッチが離された
      if (up_sw == true) { // さっきまで上スイッチが押されていた
        // フォントタイプ番号を +1します。
        font_type_no++;
        if (font_type_no >= MAX_FONT_TYPE_LIST) {
          font_type_no = 0;
        }
        curr_no = -1; // 強制的に再描画させる
        // 上スイッチ・フラグをクリア
        up_sw = false;
      }
    }

    // 下スイッチの処理
    if (cklcd.getDownSW()) { // 下スイッチが押されている
      // 押され続けている間はフラグを立てるだけの処理
      down_sw = true;
    } else { // 下スイッチが離された
      if (down_sw == true) { // さっきまで下スイッチが押されていた
        // フォントタイプ番号を -1します。
        if (font_type_no == 0) {
          font_type_no = MAX_FONT_TYPE_LIST;
        }
        font_type_no--;
        curr_no = -1; // 強制的に再描画させる
        // 下スイッチ・フラグをクリア
        down_sw = false;
      }
    }

    // 右スイッチの処理
    if (cklcd.getRightSW()) { // 右スイッチが押されている
      // 押され続けている間はフラグを立てるだけの処理
      right_sw = true;
    } else { // 右スイッチが離された
      if (right_sw == true) { // さっきまで右スイッチが押されていた
        disp_no++;
        if (disp_no >= MAX_DISP_NO) {
          disp_no = 0;
        }
        // 右スイッチ・フラグをクリア
        right_sw = false;
      }
    }
  }

  if (curr_no != disp_no) { // 現在表示している画面と表示させたい画面の番号が一致していない
    cklcd.v1.clear(); // 一旦、仮想画面をクリアする
    uint8_t x, y, euc_hi, euc_lo;
    // 表示させる EUC-JP コードの計算
    euc_hi = 0xA0 + (disp_no / 3);        // EUC-JP の上位バイト
    euc_lo = 0xA0 + (disp_no % 3) * 0x20; // EUC-JP の下位バイト
    if (euc_hi == 0xA0) {
      euc_hi = 0x00; // ASCII
      euc_lo &= 0x7F;
    } else if (euc_hi == 0xFF) {
      euc_hi = 0x8E; // 半角カナ
    }

    cklcd.v1.setFontType (CKLCD_FONT_TYPE_12px_MS);  // フォントタイプを12px:モノスペース(固定幅)にします。
    if (euc_hi) {
      // ２バイト文字
      cklcd.v1.print (euc_hi, HEX);
      cklcd.v1.print (euc_lo, HEX);
    } else {
      // ASCII：１バイト文字
      cklcd.v1.print ("00");
      cklcd.v1.print (euc_lo, HEX);
    }
    cklcd.v1.print (": ");
    cklcd.v1.print (FontNameList[font_type_no]); // フォント名
    cklcd.v1.print ("\n");

    // 縦４行分
    cklcd.v1.setFontType (FontTypeList[font_type_no]);  // フォントタイプをモノスペース(固定幅)/プロポーショナル(可変幅)にします。
    for (y = 0; y < 4; y++) {
      // 横８文字分
      for (x = 0; x < 8; x++) {
        if (euc_hi) {
          cklcd.v1.write (euc_hi); // EUC-JP の上位バイト
        }
        cklcd.v1.write (euc_lo); // EUC-JP の下位バイト
        euc_lo++;
      }
      cklcd.v1.print ("<\n");
    }
    curr_no = disp_no; // 現在、表示している画面の番号を更新
  }
}

