/*
  Color Kanji LCD Shield - Print Effect Demonstration #2 (Portrait)

  本スケッチは "カラー漢字 LCD シールド" でデモ表示するスケッチ例です。
  影/縁取り/グロー/ネオン効果が確認できます。背景色は黒/緑です。

  Circuit:
    pins A4(SDA), A5(SCL) for TWI communication in Arduino standard.
    pins SDA, SCL for TWI communication in Arduino UNO R3, Leonardo.

  created 31/july/2013 by I.Kato @ BS21 Lab

  This code is in the public domain.
*/

// TWI 通信用ライブラリ：必須です
#include <Wire.h>

// カラー漢字 LCD 用ライブラリ
#include <ColorKanjiLCD.h>


// カラー漢字 LCD シールドの TWI アドレスを 0x3C 以外に設定した場合は、以下の値を書きかえてください。
// TWI 通信できていない時は、カラー漢字 LCD シールド自身に現在設定されている TWI(i2c) アドレスが
// 表示されているはずなので、御確認ください。
#define CKLCD_TWI_ADDR  0x3C


// カラー漢字 LCD のインスタンス生成
ColorKanjiLCD cklcd (CKLCD_TWI_ADDR); // カラー漢字 LCD シールドの TWI アドレスを指定する


uint8_t xDispPos = 0;  // LCD 表示開始x座標


// 初期化関数
void setup()
{
  // 必ず、最初に Wire ライブラリを初期化させておいてください。
  // 注意："CKLCD_Wire" は Arduino Due では "Wire1" を、それ以外では "Wire" になります。
  CKLCD_Wire.begin();

#if 1
  // 注意：ここで、強制的に TWI 通信を高速化させています。カラー漢字 LCD シールドは 400kHz の速度に対応していますが、
  //       標準 100kHz の速度にしか対応していない別の TWI 機器を同時に使用する場合、そちらの TWI 機器が正常に TWI 通信できなくなる可能性があります。
  //       その場合は、ここの「#if 1」の "1" を "0" に書き換えて、TWI 通信の高速化を行わない(つまり標準 100kHz のまま)ようにしてください。
  //       当然ですが、そうした場合は、カラー漢字 LCD シールドとの TWI 通信速度も遅くなってしまいます。
  CKLCD_Wire_setHiSpeed();
#endif

  // 続いて、カラー漢字 LCD ライブラリを初期化します。
  // 符号化方式は UTF-8 方式がデフォルトなので、特に指定しなくても OK です。
  // 初期化に成功すると、画面はクリアされ、LCD 表示方向はポートレート(縦向き) に、カーソル位置は (0,0) になります。
  // カラー漢字 LCD は、電源投入直後は自動的に Version と TWI アドレスを表示していますので、
  // 初期化に失敗した場合は、TWI アドレスが、上記の CKLCD_TWI_ADDR と一致しているか確認してください。
  // 仮想画面１(256×160) を背景面に、仮想画面２(128×160) を前景面に設定します。
  cklcd.begin (CKLCD_ORIENT_PORTRAIT + CKLCD_DISP_SCREEN_F2_B1);

  // 自動フリップモードを OFF にします。flip() で一括して表示させます。
  cklcd.setFlipMode (CKLCD_FLIP_MODE_MANU);

  // 仮想画面１の右半分を緑で塗りつぶす
  cklcd.v1.setDrawColor (CKLCD_COL32_GREEN);
  cklcd.v1.rectFill (CKLCD_VPLANE1_MAX_P_WIDTH / 2, 0, CKLCD_VPLANE1_MAX_P_WIDTH - 1, CKLCD_VPLANE1_MAX_P_HEIGHT - 1); // 矩形描画(塗りつぶし)


  // フォントタイプをプロポーショナル(可変幅)にします。
  cklcd.v2.setFontType (CKLCD_FONT_TYPE_PROPORTIONAL);

  // 普通の文字
  cklcd.v2.setDrawColor (CKLCD_COL32_GREEN);
  cklcd.v2.setCursorMin (5, 10);
  cklcd.v2.print ("普通の文字");

  // 影付き文字列の描画
  // X/Y ともに offset分ずれた位置に colShadow 色の影がある colBody 色の文字列を描画します。
  // 注意：余りにも離れた位置に影があると、とても変な影付き文字に感じますので、offset値は程々にしておくべきでしょう。
  // 内部では、str を colShadow 色で描画、blur 処理、再度 str を colBody 色でアルファブレンド描画しています。
  cklcd.v2.shadowPrint (5, 40, 2, CKLCD_COL32_GREEN, CKLCD_COL32_BLACK, "影付き文字");

  // 縁取り付き文字列の描画
  // colStroke 色の縁取りが付いた colBody 色の文字列を描画します。遅いです。
  // 注意：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate 処理がかかり、おかしくなってしまいます。
  // 内部では、str を colStroke 色で描画、dilate 処理、再度 str を colBody 色でアルファブレンド描画しています。
  cklcd.v2.strokePrint (5, 70, CKLCD_COL32_GREEN, CKLCD_COL32_BLACK, "縁取り付き文字");

  // グロー効果の文字列の描画
  // colGlow 色のグロー効果が付いた colBody 色の文字列を描画します。すごく遅いです。
  // 注意：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate & blur 処理がかかり、おかしくなってしまいます。
  // 内部では、str を colGlow 色で描画、dilate 処理、blur 処理x2、再度 str を colBody 色でアルファブレンド描画しています。
  cklcd.v2.glowPrint (5, 100, CKLCD_COL32_GREEN, CKLCD_COL32_BLACK, "ｸﾞﾛｰ効果の文字");

  // ネオン効果の文字列の描画
  // colGlow2 色のグロー効果に、さらに colGlow1 色のグロー効果が付いた colBody 色の文字列を描画します。激しく遅いです。
  // 注意1：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate & blur 処理がかかり、おかしくなってしまいます。
  // 注意2：colGlow1 色と colGlow2 色の組み合わせによっては、２回目の dilate 処理の時に破綻して、変なネオン効果になる場合があります。
  // 内部では、str を colGlow2 色で描画、dilate 処理、blur 処理、str を colGlow1 色で再描画、再 dilate 処理、再 blur 処理、str を colBody 色で再々描画しています。
  cklcd.v2.neonPrint (5, 130, CKLCD_COL32_BLACK, CKLCD_COL32_GREEN, CKLCD_COL32_BLACK, "ﾈｵﾝ効果の文字");
}


// ループ関数
void loop()
{
  // 仮想画面１の LCD 表示開始座標をずらします
  cklcd.v1.setDispStartPos (xDispPos++, 0);

  // LCD 全領域を描画させます：25ms 弱かかります
  cklcd.flip();
}

