/*
  Color Kanji LCD Shield - Rotation Sphere GIF from SD Demo #1

  本スケッチは "カラー漢字LCDシールド" で、SD から読み込んだRotation Sphere アニメーション GIF をデモ表示するスケッチ例です。
  SD カードが読み込めるシールド(例：Ethernet R3シールド)が、別途、必要になります。
  添付の R_Sphere.gif を、SD カードのルートフォンダにコピーしてから、本スケッチを実行して下さい。
  また、右のタクトSW が押されている間は、アニメーションを停止します。離すとアニメーションを再開します。

  コンパイルすると 21kB 強になるので、Arduino Uno (ATmega328P) 等のフラッシュメモリが 32kB 以上でないと動作しません。

  Circuit:
    pins A4(SDA), A5(SCL) for TWI communication in Arduino standard.
    pins SDA, SCL for TWI communication in Arduino UNO R3, Leonardo.

  created 16/aug/2013 by I.Kato @ BS21 Lab

  This code is in the public domain.
*/

// TWI 通信用ライブラリ：必須です
#include <Wire.h>

// SD ライブラリ
#include <SD.h>

// カラー漢字 LCD 用ライブラリ
#include <ColorKanjiLCD.h>


// カラー漢字 LCD シールドの TWI アドレスを 0x3C 以外に設定した場合は、以下の値を書きかえてください。
// TWI 通信できていない時は、カラー漢字 LCD シールド自身に現在設定されている TWI(i2c) アドレスが
// 表示されているはずなので、御確認ください。
#define CKLCD_TWI_ADDR  0x3C


// Ether シールドの SD 選択ピン番号です。
// Ether シールド以外で SD を読み込ませる場合、このピン番号とは違っているかもしれません。適宜、変更して下さい。
#define SD_CHIP_SEL_PIN  4


// カラー漢字 LCD のインスタンス生成
ColorKanjiLCD cklcd (CKLCD_TWI_ADDR); // カラー漢字 LCD シールドの TWI アドレスを指定する


const int16_t imgLeft = 30; // GIF 画像開始 x 座標
const int16_t imgTop  =  4; // GIF 画像開始 y 座標

// Rotation Sphere GIF パターン : 100×121 画素, 22kB, 6 frames
//   Made by BorisFromStockdale, from http://en.wikipedia.org/wiki/File:Rotating_Sphere.gif
//     La bildo estas kopiita de wikipedia:en. La originala priskribo estas:
//     made with mathematica
//   This file is licensed under the Creative Commons Attribution-ShareAlike 3.0 Unported license.
//   このファイルはクリエイティブ・コモンズ 表示-継承 3.0 (CC BY-SA 3.0) 非移植ライセンスのもとに利用を許諾されています。
const char fname[] = "R_Sphere.gif";


// ユーザー定義データ構造体：ColorKanjiLCDLoadImgSize 構造体の派生構造体にしてください
// ユーザーが自由に使えるデータです。通常は画像データに関する構造体にします。
struct UserDefData : public ColorKanjiLCDLoadImgSize {
  File imgFile;  // SD カード内の GIF ファイルへの File オブジェクト
};


uint8_t left_sw, right_sw, up_sw, down_sw; // カラー漢字 LCD シールドの４つのスイッチの状態：以下で実際に使用しているのは右SW だけ


// 初期化関数
void setup()
{
  // 必ず、最初に Wire ライブラリを初期化させておいてください。
  // 注意："CKLCD_Wire" は Arduino Due では "Wire1" を、それ以外では "Wire" になります。
  CKLCD_Wire.begin();

#if 1
  // 注意：ここで、強制的に TWI 通信を高速化させています。カラー漢字 LCD シールドは 400kHz の速度に対応していますが、
  //       標準 100kHz の速度にしか対応していない別の TWI 機器を同時に使用する場合、そちらの TWI 機器が正常に TWI 通信できなくなる可能性があります。
  //       その場合は、ここの「#if 1」の "1" を "0" に書き換えて、TWI 通信の高速化を行わない(つまり標準 100kHz のまま)ようにしてください。
  //       当然ですが、そうした場合は、カラー漢字 LCD シールドとの TWI 通信速度も遅くなってしまいます。
  CKLCD_Wire_setHiSpeed();
#endif


  // 続いて、カラー漢字 LCD ライブラリを初期化します。
  // 符号化方式は UTF-8 方式がデフォルトなので、特に指定しなくても OK です。
  // 初期化に成功すると、画面はクリアされ、LCD 表示方向はランドスケープ(横向き) に、カーソル位置は (0,0) になります。
  // 漢字 LCD は、電源投入直後は自動的に Version と TWI アドレスを表示していますので、
  // 初期化に失敗した場合は、TWI アドレスが CKLCD_TWI_ADDR と一致しているか確認してください。
  cklcd.begin (CKLCD_ORIENT_LANDSCAPE + CKLCD_DISP_SCREEN_B1); // Back面：仮想画面１(160×256) を使用


  // SPI 通信を有効にする為、SS ピンを出力に切り替えます。
  pinMode (SS, OUTPUT);

  // SD ライブラリを初期化します。
  if (SD.begin (SD_CHIP_SEL_PIN) == false) {
    // 失敗した場合はメッセージ表示して終了
    cklcd.v1.print ("SD Lib 初期化失敗");
  }


  // スイッチは押されていない状態にする
  left_sw = right_sw = up_sw = down_sw = false;
}


// loadGif 関数から呼び出されるユーザー定義 callback 関数(Load)
// buff に buff_size 分の画像データを転送してください。正常に転送できた時は CKLCD_ERR_NONE を返してください。
// もし転送できなかったら CKLCD_ERR_NONE 以外（例えば CKLCD_ERR_UNKNOWN）を返してください。その時は loadGif 関数の処理を中断します。
errCode_t UserDefLoadGifCallback (uint8_t * buff, uint8_t buff_size, ColorKanjiLCDLoadImgSize * user_def_data)
{
  UserDefData * udd = (UserDefData *)user_def_data; // ユーザー定義データ構造体にキャストする：ダウンキャストなので要注意！

  // buff ポインタが指すメモリに buff_size バイト分の Rotation Sphere GIF パターン画像データを転送します。
  while (buff_size) { // 0バイトになるまで繰り返す
    int dat = udd->imgFile.read(); // SD から1バイト読み出す
    if (dat < 0) {
      // SD から読み出せなかったら、転送を中断して、終了
      return CKLCD_ERR_UNKNOWN;
    }
    // SD から1バイト読み出せた
    *buff++ = (uint8_t)dat; // バッファに書く
    --buff_size;
  }

  // 正常に転送できた
  return CKLCD_ERR_NONE;
}


// loadGif 関数から呼び出されるユーザー定義 callback 関数(Frame)
// アニメーション GIF の時のみ、各フレーム毎に呼び出されます。flip 関数で LCD に表示させると、アニメーションしている様に演出できます。
// ただし、最終フレームの時は、この callback 関数は呼ばれずに、loadGif 関数自体が正常終了しますので、最終フレームだけは、他の loadXxx 関数と同じ手順で表示させてください。
// また、アニメーション GIF でない単なる GIF 画像の場合、この callback 関数が呼び出される事はありません。他の loadXxx 関数と同じ手順で表示させてください。
// なお、この callback 関数内では、以下の '★' 印の付いた関数だけしか呼び出せませんので、注意して下さい。これ以外の関数を呼び出すと、loadGif 関数が中断される事になります。
// 特に print/write 系関数は一切、呼び出せない（CKLCD 内の作業メモリが足らないので不可能）ので、アニメーションの最中に、文字を表示させる事はできません。
// 一方、タクトSW の現在の状態は読み出せますので、タクトSW が押されたらアニメーション GIF をコマ送りする事は可能です。
errCode_t UserDefGifFrameCallback (ColorKanjiLCDLoadImgSize * user_def_data)
{
  UserDefData * udd = (UserDefData *)user_def_data; // ユーザー定義データ構造体にキャストする：ダウンキャストなので要注意！

  // LCD 全領域を描画させます
//  cklcd.flip();  ← これだとより時間が掛かる

  // アニメーション GIF 描画範囲だけを LCD に描画させます
  // x_offset, y_offset も加算しているのは、アニメーション GIF の場合、フレームによって描画位置を矩形範囲が変化する為です。
  int16_t flipLeft   = imgLeft  + udd->x_offset;
  int16_t flipTop    = imgTop   + udd->y_offset;
  int16_t flipRight  = flipLeft + udd->width  - 1;
  int16_t flipBottom = flipTop  + udd->height - 1;
  cklcd.flip (flipLeft, flipTop, flipRight, flipBottom);

  // このユーザー定義 callback 関数(Frame)で、少なくとも一度は、タクト SW のチェックを行います。
  // ただし、フレームとフレームにまたがって、タクトSW の ON/OFF の変化を検出させる場合は、
  // フレーム描画に時間が掛かる場合（通常は寸法の大きなアニメーション GIF）に、検出の反応が悪くなる場合があります。
  while(1) {
    // 最初にカラー漢字 LCD の現在のステータスを取得します。
    // これを呼び出しておかないと、以下の getRightSW 関数, getLeftSW 関数が
    // 何時までも更新されない事になってしまいます。
    // カラー漢字 LCD がコマンド処理中で忙しい時は、ステータス取得に失敗して CKLCD_ERR_BUSY が返って来ます。
    // ステータス取得成功を意味する CKLCD_ERR_NONE が返ってきた時だけ、処理を進めるようにして下さい。
    if (cklcd.getStatus() == CKLCD_ERR_NONE) {
      // ステータス取得成功した

      // 右スイッチの処理：押されている間はアニメーションを停止する
      if (cklcd.getRightSW()) { // 右スイッチが押されている
        // 押され続けている間はフラグを立てるだけの処理
        right_sw = true;
      } else { // 右スイッチが離された
        if (right_sw == true) { // さっきまで右スイッチが押されていた
          // 右スイッチ・フラグをクリア
          right_sw = false;
        }
      }
    }

    // 右スイッチが離されていたら、while 文を break させます。
    if (right_sw == false) {
      break;
    }

    // 10ms 待ちを入れる。
    // カラー漢字 LCD シールドのスイッチのチェック間隔が約 10ms (変更不可) なので、
    // これよりも短い間隔で getStatus 関数を呼び出しても、前と同じスイッチ状態が読み出されるだけです。
    // なので、10ms 以上の間をおいてから、getStatus 関数を呼び出すようにしてください。
    delay(10);
  }

  // 正常に転送できた
  return CKLCD_ERR_NONE;
}


// ループ関数
void loop()
{
  // SD からアニメーション GIF ファイルを読み出し、カラー漢字 LCD シールドに転送します

  // ユーザー定義データ構造体
  UserDefData udd;

  // GIF ファイルを、読み込みモードにてオープンします。
  udd.imgFile = SD.open (fname, FILE_READ);

  if (udd.imgFile == false) {
    // オープンできなかったら、メッセージ表示して終了
    cklcd.v1.print (fname);
    cklcd.v1.print ("\nのOpenに失敗");
    return;
  }

  // "読込み中"メッセージを表示させます（仮想画面１を使用）。
//  cklcd.v1.print ("GIF画像読込み中...");

  // GIF 読み込み予定位置を暗い緑で塗りつぶしておきます。こうする事で GIF の端が分かりやすくなります。
//  cklcd.v1.setBackColor (CKLCD_COL32_DARK_GREEN); // 暗い緑
//  cklcd.v1.clearRect (0, 0, 159, 127);

  // 仮想画面１の(30,4)に、Rotation Sphere GIF 画像データを読み込ませます。エラー発生時はメッセージを表示させます。
  errCode_t res = cklcd.v1.loadGif (imgLeft, imgTop, CKLCD_LOAD_IMG_DISP_ERR_MSG, udd.imgFile.size(), UserDefLoadGifCallback, UserDefGifFrameCallback, &udd);

  // オープンした GIF ファイルをクローズします。
  udd.imgFile.close();

//  cklcd.v1.setBackColor (CKLCD_COL32_BLACK); // 黒に戻す

  if (res != CKLCD_ERR_NONE) { // エラーの時はエラーメッセージを表示
    cklcd.v1.print ("\nErrCode=0x");
    cklcd.v1.print (res, HEX);
  }

  // LCD 全領域を描画させます
//  cklcd.flip();  ← これだとより時間が掛かる

  // アニメーション GIF 描画範囲だけを LCD に描画させます
  // x_offset, y_offset も加算しているのは、アニメーション GIF の場合、フレームによって描画位置を矩形範囲が変化する為です。
  int16_t flipLeft   = imgLeft  + udd.x_offset;
  int16_t flipTop    = imgTop   + udd.y_offset;
  int16_t flipRight  = flipLeft + udd.width  - 1;
  int16_t flipBottom = flipTop  + udd.height - 1;
  cklcd.flip (flipLeft, flipTop, flipRight, flipBottom);
}

