/*
  Color Kanji LCD Shield - Rotation Demo #1

  本スケッチは "カラー漢字 LCD シールド" の回転＆拡大/縮小のスケッチ例です。

  Circuit:
    pins A4(SDA), A5(SCL) for TWI communication in Arduino standard.
    pins SDA, SCL for TWI communication in Arduino UNO R3, Leonardo.

  created 31/july/2013 by I.Kato @ BS21 Lab

  This code is in the public domain.
*/

// TWI 通信用ライブラリ：必須です
#include <Wire.h>

// カラー漢字 LCD 用ライブラリ
#include <ColorKanjiLCD.h>


// カラー漢字 LCD シールドの TWI アドレスを 0x3C 以外に設定した場合は、以下の値を書きかえてください。
// TWI 通信できていない時は、カラー漢字 LCD シールド自身に現在設定されている TWI(i2c) アドレスが
// 表示されているはずなので、御確認ください。
#define CKLCD_TWI_ADDR  0x3C


// カラー漢字 LCD のインスタンス生成
ColorKanjiLCD cklcd (CKLCD_TWI_ADDR); // カラー漢字 LCD シールドの TWI アドレスを指定する


// 初期化関数
void setup()
{
  // 必ず、最初に Wire ライブラリを初期化させておいてください。
  // 注意："CKLCD_Wire" は Arduino Due では "Wire1" を、それ以外では "Wire" になります。
  CKLCD_Wire.begin();

#if 1
  // 注意：ここで、強制的に TWI 通信を高速化させています。カラー漢字 LCD シールドは 400kHz の速度に対応していますが、
  //       標準 100kHz の速度にしか対応していない別の TWI 機器を同時に使用する場合、そちらの TWI 機器が正常に TWI 通信できなくなる可能性があります。
  //       その場合は、ここの「#if 1」の "1" を "0" に書き換えて、TWI 通信の高速化を行わない(つまり標準 100kHz のまま)ようにしてください。
  //       当然ですが、そうした場合は、カラー漢字 LCD シールドとの TWI 通信速度も遅くなってしまいます。
  CKLCD_Wire_setHiSpeed();
#endif

  // 続いて、カラー漢字 LCD ライブラリを初期化します。
  // 符号化方式は UTF-8 方式がデフォルトなので、特に指定しなくても OK です。
  // 初期化に成功すると、画面はクリアされ、LCD 表示方向はランドスケープ(横向き) に、カーソル位置は (0,0) になります。
  // カラー漢字 LCD は、電源投入直後は自動的に Version と TWI アドレスを表示していますので、
  // 初期化に失敗した場合は、TWI アドレスが、上記の CKLCD_TWI_ADDR と一致しているか確認してください。
  // 仮想画面１(160×256) を背景面(RGB565)に設定します。仮想画面２(160×128) は表示されませんが、作業用(RGB565)として使用します。
  cklcd.begin (CKLCD_ORIENT_LANDSCAPE + CKLCD_DISP_SCREEN_B1);

  // フォントタイプをプロポーショナル(可変幅)にします。
  cklcd.v2.setFontType (CKLCD_FONT_TYPE_PROPORTIONAL);

  // ネオン効果の文字列の描画
  // colGlow2 色のグロー効果に、さらに colGlow1 色のグロー効果が付いた colBody 色の文字列を描画します。激しく遅いです。
  // 注意1：文字列は１行以内に収めてください。複数行にまたがると、既存の描画部分も dilate & blur 処理がかかり、おかしくなってしまいます。
  // 注意2：colGlow1 色と colGlow2 色の組み合わせによっては、２回目の dilate 処理の時に破綻して、変なネオン効果になる場合があります。
  // 内部では、str を colGlow2 色で描画、dilate 処理、blur 処理、str を colGlow1 色で再描画、再 dilate 処理、再 blur 処理、str を colBody 色で再々描画しています。
  cklcd.v2.neonPrint (18, 53, CKLCD_COL32_BLACK, CKLCD_COL32_WHITE, CKLCD_COL32_BLUE, "回転・拡大・縮小");

  // コピー元の矩形領域を指定
  cklcd.v2.setSrcRect (0, 0, CKLCD_VPLANE2_MAX_L_WIDTH - 1, CKLCD_VPLANE2_MAX_L_HEIGHT - 1);

  // 文字色：白
  cklcd.v2.setDrawColor (CKLCD_COL32_WHITE);

  // 背景色：暗い緑
  // 回転＆拡大/縮小の時に、コピー元に無い部分は、この背景色で代用されます。
  cklcd.v2.setBackColor (CKLCD_COL32_DARK_GREEN);
}


cklcdInvRatio10_t ratio = CKLCD_FLOAT_TO_INVRATIO10(1.2); // 1.2倍でコピー
cklcdBAM10_t      bam10 = 0; // 回転角度 ("BAM10"単位)


// ループ関数
void loop()
{
  // 現在の回転角度の値を描画
  cklcd.v2.setCursorMin (60, 80);
  cklcd.v2.print (bam10, DEC);

  // 回転＆拡大/縮小させながらコピー：劇遅!!
  // setSrcRect()で指定されている矩形領域を回転("BAM10"単位)させながら、(x0,y0)-(x1,y1)の範囲内にコピーします。
  // invRatio10 倍の拡大/縮小もできますが、単なる近傍点補間しかしてないので、結果画像は汚いです。大昔の家庭用ゲーム機みたいな効果しかありません。
  // 回転の中心はそれぞれ、コピー元の矩形領域の中心座標と、コピー先の矩形領域の中心座標、になります。
  // 回転＆拡大/縮小の時、コピー元の矩形領域に存在し無い部分は setBackColor()で設定している色になります。
  // コピー元の矩形領域と、コピー先の矩形領域は、部分的にも重なっていてはいけません。重なってた場合、おかしな結果画像になってしまいます。
  errCode_t res = cklcd.v1.copyWithRotate (0, 0, CKLCD_MAX_L_WIDTH - 1, CKLCD_MAX_L_HEIGHT - 1, ratio, bam10);

  // LCD 全領域を描画させます
  cklcd.flip();

  bam10++; // 次の回転角度
}

