/*
  Color Kanji LCD Shield - Set Pixel Test #1 (Portrait)

  本スケッチは "カラー漢字LCDシールド" で、Set Pixel を test するスケッチ例です。

  Circuit:
    pins A4(SDA), A5(SCL) for TWI communication in Arduino standard.
    pins SDA, SCL for TWI communication in Arduino UNO R3, Leonardo.

  created 14/aug/2013 by I.Kato @ BS21 Lab

  This code is in the public domain.
*/

// TWI 通信用ライブラリ：必須です
#include <Wire.h>

// カラー漢字 LCD 用ライブラリ
#include <ColorKanjiLCD.h>


// カラー漢字 LCD シールドの TWI アドレスを 0x3C 以外に設定した場合は、以下の値を書きかえてください。
// TWI 通信できていない時は、カラー漢字 LCD シールド自身に現在設定されている TWI(i2c) アドレスが
// 表示されているはずなので、御確認ください。
#define CKLCD_TWI_ADDR  0x3C


// カラー漢字 LCD のインスタンス生成
ColorKanjiLCD cklcd (CKLCD_TWI_ADDR); // カラー漢字 LCD シールドの TWI アドレスを指定する


// カラー漢字 LCD シールドの４つのスイッチの状態
uint8_t left_sw, up_sw, down_sw, right_sw;



// 初期化関数
void setup()
{
  // 必ず、最初に Wire ライブラリを初期化させておいてください。
  // 注意："CKLCD_Wire" は Arduino Due では "Wire1" を、それ以外では "Wire" になります。
  CKLCD_Wire.begin();

#if 1
  // 注意：ここで、強制的に TWI 通信を高速化させています。カラー漢字 LCD シールドは 400kHz の速度に対応していますが、
  //       標準 100kHz の速度にしか対応していない別の TWI 機器を同時に使用する場合、そちらの TWI 機器が正常に TWI 通信できなくなる可能性があります。
  //       その場合は、ここの「#if 1」の "1" を "0" に書き換えて、TWI 通信の高速化を行わない(つまり標準 100kHz のまま)ようにしてください。
  //       当然ですが、そうした場合は、カラー漢字 LCD シールドとの TWI 通信速度も遅くなってしまいます。
  CKLCD_Wire_setHiSpeed();
#endif

  // 続いて、カラー漢字 LCD ライブラリを初期化します。
  // 符号化方式は UTF-8 方式がデフォルトなので、特に指定しなくても OK です。
  // 初期化に成功すると、画面はクリアされ、LCD 表示方向はポートレート(縦向き) に、カーソル位置は (0,0) になります。
  // 漢字 LCD は、電源投入直後は自動的に Version と TWI アドレスを表示していますので、
  // 初期化に失敗した場合は、TWI アドレスが CKLCD_TWI_ADDR と一致しているか確認してください。
  cklcd.begin (CKLCD_ORIENT_PORTRAIT + CKLCD_DISP_SCREEN_B1); // Back面：仮想画面１(256×160) を使用

  // メッセージ表示
  cklcd.v1.print ("全面setPixelに\n時間がかかります");

  // 仮想画面１の全領域に cklcdCol16_t 値を書き込む
  uint16_t x, y;
#if 1
  // 仮想画面#1 にフルに書き込むので、400kHz でも、約12秒かかります。
  for (x = 0; x <= 255; x++) {
    for (y = 0; y <= 159; y++) {
      cklcdCol16_t col16 = (x << 8) + y;
      cklcd.v1.setPixel (x, y, col16);
    }
  }
#else
  // 書き込む量を減らして、早くしたもの
  for (y = 0; y <= 255; y++) {
    cklcd.v1.setPixel (y, y, (cklcdCol16_t)y << 8);
  }
#endif

  // スイッチは押されていない状態にする
  left_sw = up_sw = down_sw = right_sw = false;
}


uint8_t xDispPos   = 0;  // LCD 表示開始x座標
uint8_t scrollFlag = 1;  // スクロールflag


// ループ関数
void loop()
{
  // 最初にカラー漢字 LCD の現在のステータスを取得します。
  // これを呼び出しておかないと、以下の getLeft/Up/Down/RightSW 関数が
  // 何時までも更新されない事になってしまいます。
  // カラー漢字 LCD がコマンド処理中で忙しい時は、ステータス取得に失敗して CKLCD_ERR_BUSY が返って来ます。
  // ステータス取得成功を意味する CKLCD_ERR_NONE が返ってきた時だけ、処理を進めるようにして下さい。
  if (cklcd.getStatus() == CKLCD_ERR_NONE) {
    // ステータス取得成功した

    // 左スイッチの処理
    if (cklcd.getLeftSW()) { // 左スイッチが押されている
      // 押され続けている間はフラグを立てるだけの処理
      left_sw = true;
    } else { // 左スイッチが離された
      if (left_sw == true) { // さっきまで左スイッチが押されていた
        // 左スイッチ・フラグをクリア
        left_sw = false;
      }
    }

    // 上スイッチの処理
    if (cklcd.getUpSW()) { // 上スイッチが押されている
      // 押され続けている間はフラグを立てるだけの処理
      up_sw = true;
    } else { // 上スイッチが離された
      if (up_sw == true) { // さっきまで上スイッチが押されていた
        // 上スイッチ・フラグをクリア
        up_sw = false;
      }
    }

    // 下スイッチの処理
    if (cklcd.getDownSW()) { // 下スイッチが押されている
      // 押され続けている間はフラグを立てるだけの処理
      down_sw = true;
    } else { // 下スイッチが離された
      if (down_sw == true) { // さっきまで下スイッチが押されていた
        // LCD 表示開始座標をずらします
        cklcd.v1.setDispStartPos (xDispPos = 0, 0);
        // 下スイッチ・フラグをクリア
        down_sw = false;
      }
    }

    // 右スイッチの処理
    if (cklcd.getRightSW()) { // 右スイッチが押されている
      // 押され続けている間はフラグを立てるだけの処理
      right_sw = true;
    } else { // 右スイッチが離された
      if (right_sw == true) { // さっきまで右スイッチが押されていた
        scrollFlag ^= 1; // スクロールflag をトグル
        // 右スイッチ・フラグをクリア
        right_sw = false;
      }
    }
  }

  if (scrollFlag) {
    // LCD 表示開始座標をずらします
    cklcd.v1.setDispStartPos (xDispPos++, 0);
  }

  // LCD 全領域を描画させます
  cklcd.flip();
}

